/* Copyright (c) 2004-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Wizard to Build/Update the CIDB (Code Intelligence Database) with data
 * about a language library for completion.
 *
 * An object may be passed in at window.arguments[0]. The following optional
 * attributes on that object are supported:
 *      .launchedAutomatically
 *          a boolean indicating that this wizard was launched automatically.
 *          Default is false.
 *      .mode
 *          One of the batch update request types: "language", "directory",
 *          or "upgrade". This controls the wizard panel path (see below).
 *      .activePrefs
 *          null or a koIPreferenceSet instance on which the wizard should
 *          *look up* data (not change it). Typically this is passed in
 *          by Komodo's Preferences dialog if it is the caller, in case the
 *          user has changed some relevant prefs but not yet saved them.
 *
 * Wizard Paths:
 * - Scanning language installation(s):
 *    Panel1: Select Languages
 *    Panel2: Ready to Build
 *    Panel3: Building Database (page with progress meter)
 *    Panel4: Build Database Completed/Errors/Aborted
 * - Scanning directories:
 *    Panel0: Select Directories
 *    Panel2: Ready to Build
 *    Panel3: Building Database (page with progress meter)
 *    Panel4: Build Database Completed/Errors/Aborted
 * - Upgrading the CIDB:
 *    Panel3: Upgrading Database (page with progress meter)
 *    Panel4: Build Database Errors (only if there is an error)
 */


//---- globals

var log = getLoggingMgr().getLogger("buildcidb");
//log.setLevel(LOG_DEBUG);

var gWidgets = new Object();
var gLangsToSetup = [];
var gPanelsInitialized = {};
var gLang2Path = {};
var gLang2PathExtra = {};  // values for .batchUpdateRequest's "extra" arg
var gCompletionReason = null;
var gErrorStr = null;
var gCancelOnCommand = null; // a replacement Cancel button oncommand handler
var gLaunchedAutomatically = false;
var gMode = "language";
var gDirsToScan = [];

var gGlobalPrefs = null;
var gActivePrefs = null;
var gCodeIntelSvc = null;



//--- internal support routines

function _setTitleBar(title, subtitle /*=null*/) {
    if (typeof(subtitle) == "undefined") subtitle = null;
    gWidgets.title.value = title;
    var subtitleWidget = gWidgets.subtitle;
    while (subtitleWidget.hasChildNodes())
        subtitleWidget.removeChild(subtitleWidget.firstChild);
    if (subtitle) {
        var text = document.createTextNode(subtitle);
        subtitleWidget.appendChild(text);
    }
}

function _setBackButton(label, enabled, accesskey /*=null*/) {
    if (typeof(accesskey) == "undefined") accesskey = null;
    gWidgets.back.setAttribute("label", "< "+label);
    if (enabled) {
        gWidgets.back.removeAttribute("disabled");
    } else {
        gWidgets.back.setAttribute("disabled", "true");
    }
    if (accesskey) {
        gWidgets.back.setAttribute("accesskey", accesskey);
    } else {
        gWidgets.back.removeAttribute("accesskey");
    }
}

function _setNextButton(label, enabled, accesskey /*=null*/) {
    if (typeof(accesskey) == "undefined") accesskey = null;
    gWidgets.next.setAttribute("label", label+" >");
    if (enabled) {
        gWidgets.next.removeAttribute("disabled");
    } else {
        gWidgets.next.setAttribute("disabled", "true");
    }
    if (accesskey) {
        gWidgets.next.setAttribute("accesskey", accesskey);
    } else {
        gWidgets.next.removeAttribute("accesskey");
    }
}

function _setCancelButton(label, enabled, accesskey /*=null*/,
                          oncommand /*=null*/) {
    if (typeof(accesskey) == "undefined") accesskey = null;
    if (typeof(oncommand) == "undefined") oncommand = null;
    gWidgets.cancel.setAttribute("label", label);
    if (enabled) {
        gWidgets.cancel.removeAttribute("disabled");
    } else {
        gWidgets.cancel.setAttribute("disabled", "true");
    }
    if (accesskey) {
        gWidgets.cancel.setAttribute("accesskey", accesskey);
    } else {
        gWidgets.cancel.removeAttribute("accesskey");
    }
    if (oncommand) {
        gCancelOnCommand = oncommand;
    } else {
        gCancelOnCommand = null;
    }
}

function _englishJoinList(list, conjunction /* ="and" */) {
    if (typeof(conjunction) == "undefined" || conjunction == null)
        conjunction = "and";
    var s = "";
    if (list.length > 0) {
        s += list[0];
        for (i = 1; i < list.length-1; ++i) {
            s += ", "+list[i];
        }
        if (list.length > 1) {
            s += " "+conjunction+" "+list[list.length-1];
        }
    }
    return s;
}


// This dialog can be given an active pref set (i.e. from the main
// Preferences dialog) that *may* override the global prefs. Therefore,
// we read from the active prefs, if we have them, and write to *both*.
function _hasStringPref(name) {
    if (gActivePrefs) {
        return gActivePrefs.hasStringPref(name);
    } else {
        return gGlobalPrefs.hasStringPref(name);
    }
}
function _getStringPref(name) {
    if (gActivePrefs) {
        return gActivePrefs.getStringPref(name);
    } else {
        return gGlobalPrefs.getStringPref(name);
    }
}
function _setStringPref(name, value) {
    if (gActivePrefs) {
        gActivePrefs.setStringPref(name, value);
    }
    gGlobalPrefs.setStringPref(name, value);
}


// Return all interpreter path information for the given language that is
// required to call koICodeIntelService.batchUpdateRequest().
function _getInterpreterPathInfo(language) {
    var pathInfo = {"path": null, "extra": null};
    var sysUtils = Components.classes["@activestate.com/koSysUtils;1"].
                   getService(Components.interfaces.koISysUtils);
    switch(language) {
    case "Python":
        if (_hasStringPref("pythonDefaultInterpreter")) {
            pathInfo.path = _getStringPref("pythonDefaultInterpreter");
        }
        if (!pathInfo.path) {
            //XXX Should this use koIPythonInfoEx?
            pathInfo.path = sysUtils.Which("python");
        }
        break;
    case "Perl":
        if (_hasStringPref("perlDefaultInterpreter")) {
            pathInfo.path = _getStringPref("perlDefaultInterpreter");
        }
        if (!pathInfo.path) {
            //XXX Should this use koIPerlInfoEx?
            pathInfo.path = sysUtils.Which("perl");
        }
        break;
    case "PHP":
        var phpInfo = Components.classes["@activestate.com/koPHPInfoEx;1"].
                      createInstance(Components.interfaces.koIPHPInfoEx);
        pathInfo.path = phpInfo.executablePath;
        pathInfo.extra = phpInfo.cfg_file_path;
        break;
    case "Ruby":
        if (_hasStringPref("rubyDefaultInterpreter")) {
            pathInfo.path = _getStringPref("rubyDefaultInterpreter");
        }
        if (!pathInfo.path) {
            //XXX Should this use koIRubyInfoEx?
            pathInfo.path = sysUtils.Which("ruby");
        }
        break;
    case "Tcl":
        if (_hasStringPref("tclshDefaultInterpreter")) {
            pathInfo.path = _getStringPref("tclshDefaultInterpreter");
        }
        if (!pathInfo.path) {
            var tclInfo = Components.classes["@activestate.com/koTclInfoEx;1"].
                          createInstance(Components.interfaces.koITclInfoEx);
            pathInfo.path = tclInfo.tclsh_path;
        }
        break;
    default:
        throw("unexpected language to setup: '"+language+"'\n");
    }
    return pathInfo;
}



//---- Progress UI handler component

function ProgressUIHandler()
{
    log.debug("ProgressUIHandler()");
    try {
        this.widgets = new Object();
        this.widgets.status = document.getElementById("panel3-status");
        this.widgets.pm = document.getElementById("panel3-progressmeter");
        this.widgets.numFiles = document.getElementById("panel3-num-files");
        this.widgets.eta = document.getElementById("panel3-eta");
    } catch(ex) {
        log.exception(ex);
    }
}
ProgressUIHandler.prototype.constructor = ProgressUIHandler;

ProgressUIHandler.prototype.QueryInterface = function(iid) {
    if (!iid.equals(Components.interfaces.koICodeIntelBatchUpdateProgressUIHandler) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
}

ProgressUIHandler.prototype.setStatusMessage = function(message)
{
    log.debug("ProgressUIHandler.setStatusMessage('"+message+"')");
    try {
        this.widgets.status.setAttribute("value", message);
    } catch(ex) {
        log.exception(ex);
    }
}

ProgressUIHandler.prototype.setProgressMeterValue = function(percent)
{
    log.debug("ProgressUIHandler.setProgressMeterValue('"+percent+"')");
    try {
        this.widgets.pm.setAttribute("value", percent);
    } catch(ex) {
        log.exception(ex);
    }
}

ProgressUIHandler.prototype.setRemainingFiles = function(remaining)
{
    log.debug("ProgressUIHandler.setRemainingFiles("+remaining+")");
    try {
        if (remaining == -1) {
            this.widgets.numFiles.removeAttribute("value");
        } else {
            var msg = "Number of files to process: "+remaining;
            this.widgets.numFiles.setAttribute("value", msg);
        }
    } catch(ex) {
        log.exception(ex);
    }
}

ProgressUIHandler.prototype.setETA = function(eta)
{
    log.debug("ProgressUIHandler.setETA('"+eta+"')");
    try {
        if (eta == null) {
            this.widgets.eta.removeAttribute("value");
        } else {
            var msg = "Estimated time to completion: "+eta;
            this.widgets.eta.setAttribute("value", msg);
        }
    } catch(ex) {
        log.exception(ex);
    }
}

ProgressUIHandler.prototype.disableCancel = function(why)
{
    log.debug("ProgressUIHandler.disableCancel()");
    try {
        why = why.replace("'", "\\'");
        _setCancelButton("Cancel", false, null,
                         "alert('Cannot cancel. "+why+"');");
    } catch(ex) {
        log.exception(ex);
    }
}

ProgressUIHandler.prototype.completed = function()
{
    log.debug("ProgressUIHandler.completed()");
    try {
        if (gMode == "upgrade") {
            window.close(); // Just immediately end if upgrade succeeds.
        } else {
            gCompletionReason = "completed";
            SwitchToPanel(4);
        }
    } catch(ex) {
        log.exception(ex);
    }
}

ProgressUIHandler.prototype.cancelled = function()
{
    log.debug("ProgressUIHandler.cancelled()");
    try {
        gCompletionReason = "cancelled";
        SwitchToPanel(4);
    } catch(ex) {
        log.exception(ex);
    }
}

ProgressUIHandler.prototype.erroredout = function(errstr)
{
    log.debug("ProgressUIHandler.erroredout()");
    try {
        gCompletionReason = "erroredout";
        gErrorStr = errstr;
        SwitchToPanel(4);
    } catch(ex) {
        log.exception(ex);
    }
}



//---- interface routines for XUL

function OnLoad()
{
    log.debug("OnLoad()");
    try {
        gWidgets.title = document.getElementById("title");
        gWidgets.subtitle = document.getElementById("subtitle");
        gWidgets.back = document.getElementById("back");
        gWidgets.next = document.getElementById("next");
        gWidgets.cancel = document.getElementById("cancel");
        gWidgets.deck = document.getElementById("panel-deck");
        gWidgets.aborted = document.getElementById("panel4-aborted");
        gWidgets.completed = document.getElementById("panel4-completed");
        gWidgets.erroredout = document.getElementById("panel4-erroredout");
        gWidgets.buildErrorNote = document.getElementById("panel4-build-error-note");
        gWidgets.upgradeErrorNote = document.getElementById("panel4-upgrade-error-note");
        gWidgets.errors = document.getElementById("panel4-errors");

        // Process window arguments, if any.
        var arg;
        if (typeof(window.arguments) != "undefined"
            && window.arguments.length >= 1) {
            if (typeof window.arguments[0].launchedAutomatically != "undefined"
                && window.arguments[0].launchedAutomatically)
            {
                gLaunchedAutomatically = true;
            }
            if (typeof window.arguments[0].mode != "undefined")
            {
                gMode = window.arguments[0].mode;
            }
            if (typeof window.arguments[0].activePrefs != "undefined"
                && window.arguments[0].activePrefs)
            {
                gActivePrefs = window.arguments[0].activePrefs;
            }
        }

        if (gMode == "upgrade") {
            window.title = "Upgrade Code Intelligence Database";
        } else {
            window.title = "Build Code Intelligence Database";
        }

        gCodeIntelSvc = Components.classes["@activestate.com/koCodeIntelService;1"].getService();
        gGlobalPrefs = Components.classes["@activestate.com/koPrefService;1"].
                      getService(Components.interfaces.koIPrefService).prefs;
        _setStringPref("codeintel_have_run_build_wizard_for_cidb_version",
                       gCodeIntelSvc.cidbVersion);

        if (gMode == "directory") {
            SwitchToPanel(0);
        } else if (gMode == "language") {
            SwitchToPanel(1);
        } else if (gMode == "upgrade") {
            SwitchToPanel(3);
        } else {
            throw("unexpected batch update mode: '"+gMode+"'");
        }
    } catch (ex) {
        log.exception(ex);
    }
}


function OnUnload()
{
    log.debug("OnUnload()");
    try {
    } catch (ex) {
        log.exception(ex);
    }
}


function SwitchToPanel(num)
{
    log.debug("SwitchToPanel(num="+num+")");
    try {
        gWidgets.deck.selectedIndex = num;

        if (!(num in gPanelsInitialized)) {
            eval("Panel"+num+"_Initialize();");
            gPanelsInitialized[num] = true;
        }
        eval("Panel"+num+"_SwitchTo();");
    } catch (ex) {
        log.exception(ex);
    }
}

function Next() {
    log.debug("Next()");
    try {
        //XXX <deck>.selectedIndex returns the empty string (!?) after
        //    having been set to 0.
        var index = gWidgets.deck.selectedIndex || 0;
        eval("Panel"+index+"_Next();");
    } catch (ex) {
        log.exception(ex);
    }
}
function Back() {
    log.debug("Back()");
    try {
        //XXX <deck>.selectedIndex returns the empty string (!?) after
        //    having been set to 0.
        var index = gWidgets.deck.selectedIndex || 0;
        eval("Panel"+index+"_Back();");
    } catch (ex) {
        log.exception(ex);
    }
}


function Panel0_Initialize() {
    log.debug("Panel0_Initialize()");
    var origWindow = windowManager_getMainWindow();
    var cwd = origWindow.View_GetCwd();
    var pathlist = document.getElementById("codeIntelDirsToScan");
    pathlist.init() // must happen after onpageload
    pathlist.setCwd(cwd);
}
function Panel0_SwitchTo() {
    _setTitleBar("Select Directories",
                 "Select which directories to scan");
    _setBackButton("Back", false, "B");
    _setNextButton("Next", false, "N");
    _setCancelButton("Cancel", true, null, "window.close();");
}
function Panel0_UpdateUI()
{
    log.debug("Panel0_UpdateUI()");
    try {
        var os = Components.classes["@activestate.com/koOs;1"].getService(Components.interfaces.koIOs);
        trace_onload();
        var pathlist = document.getElementById("codeIntelDirsToScan")
        var dirsToScanData = document.getElementById("codeIntelDirsToScan").getData();
        gDirsToScan = dirsToScanData.split(os.pathsep);
        var next = document.getElementById("next");
        if (dirsToScanData.length > 0) {
            if (next.hasAttribute("disabled"))
                next.removeAttribute("disabled");
        } else {
            next.setAttribute("disabled", "true");
        }
    } catch (ex) {
        log.exception(ex);
    }
}

function Panel0_Next()
{
    log.debug("Panel0_Next()");
    try {
        SwitchToPanel(2);
    } catch (ex) {
        log.exception(ex);
    }
}


function Panel1_Initialize() {
    log.debug("Panel1_Initialize()");
    try {
        var i, widget, path, lang;
        // By default, disable languages for which an interpreter cannot
        // be found.
        allLangs = ["Python", "Perl", "PHP", "Ruby", "Tcl"];
        var anyDisabled = false;
        for (i = 0; i < allLangs.length; ++i) {
            lang = allLangs[i];
            widget = document.getElementById("panel1-"+lang);
            var pathInfo = null;
            try {
                pathInfo = _getInterpreterPathInfo(lang);
                path = pathInfo.path;
            } catch(ex2) {
                if (ex2.indexOf("unexpected language to setup") == -1) {
                    log.exception(ex2);
                }
                path = null;
            }
            if (path) {
                if (widget.hasAttribute("disabled"))
                    widget.removeAttribute("disabled");
            } else {
                widget.setAttribute("disabled", "true");
                anyDisabled = true;
            }
        }
        // If any are disabled then add an explanation for why.
        widget = document.getElementById("panel1-disabled-langs-warning");
        if (anyDisabled) {
            widget.removeAttribute("collapsed");
        } else {
            widget.setAttribute("collapsed", "true");
        }
    } catch (ex) {
        log.exception(ex);
    }
}
function Panel1_SwitchTo() {
    _setTitleBar("Select Languages",
                 "Select which languages to setup for Code Intelligence");
    _setBackButton("Back", false, "B");
    _setNextButton("Next", true, "N");
    _setCancelButton("Cancel", true);
    if (gLaunchedAutomatically) {
        document.getElementById("panel1-howto-launch")
                .removeAttribute("collapsed");
    }
    Panel1_UpdateUI();
}
function Panel1_UpdateUI()
{
    log.debug("Panel1_UpdateUI()");
    try {
        var python = document.getElementById("panel1-Python");
        var perl = document.getElementById("panel1-Perl");
        var php = document.getElementById("panel1-PHP");
        var ruby = document.getElementById("panel1-Ruby");
        var tcl = document.getElementById("panel1-Tcl");
        var next = document.getElementById("next");
        if (python.checked || perl.checked || php.checked || ruby.checked
            || tcl.checked)
        {
            if (next.hasAttribute("disabled"))
                next.removeAttribute("disabled");
        } else {
            next.setAttribute("disabled", true);
        }
    } catch (ex) {
        log.exception(ex);
    }
}
function Panel1_Next()
{
    log.debug("Panel1_Next()");
    try {
        var i;
        var python = document.getElementById("panel1-Python");
        var perl = document.getElementById("panel1-Perl");
        var php = document.getElementById("panel1-PHP");
        var ruby = document.getElementById("panel1-Ruby");
        var tcl = document.getElementById("panel1-Tcl");
        gLangsToSetup = [];
        if (python.checked) gLangsToSetup.push("Python");
        if (perl.checked) gLangsToSetup.push("Perl");
        if (php.checked) gLangsToSetup.push("PHP");
        if (ruby.checked) gLangsToSetup.push("Ruby");
        if (tcl.checked) gLangsToSetup.push("Tcl");

        // Find the interpreter path for each language to be setup.
        gLang2Path = {};
        gLang2PathExtra = {};
        langsWithNoPath = [];
        var sysUtils = Components.classes["@activestate.com/koSysUtils;1"].
                       getService(Components.interfaces.koISysUtils);
        for (i = 0; i < gLangsToSetup.length; ++i) {
            var language = gLangsToSetup[i];
            var pathInfo = _getInterpreterPathInfo(language);
            if (pathInfo.path) {
                gLang2Path[language] = pathInfo.path;
                gLang2PathExtra[language] = pathInfo.extra;
            } else {
                langsWithNoPath.push(language);
            }
        }

        // If could not find the language's interpreter path then notify the
        // user and offer to continue with the remaining languages, if any.
        var switchTo = 2;
        if (langsWithNoPath.length > 0) {
            var msg = "Could not find an installation for ";
            msg += _englishJoinList(langsWithNoPath, "or");
            msg += ".";
            if (langsWithNoPath.length == gLangsToSetup.length) {
                msg += ' You can setup language interpreters in Komodo\'s '+
                       '"Languages" Preferences panel. Press OK to abort.';
                dialog_alert(msg);
                switchTo = 4; // The "finished" page.
            } else {
                msg += ' You can setup these languages in Komodo\'s '+
                       '"Languages" Preferences panel. Would you like to '+
                       'continue building Komodo\'s database for the '+
                       'remaining languages?';
                var response = dialog_okCancel(msg);
                if (response == "Cancel") {
                    switchTo = 4; // The "finished" page.
                }
            }
        }

        SwitchToPanel(switchTo);
    } catch (ex) {
        log.exception(ex);
    }
}



function Panel2_Initialize() {
    log.debug("Panel2_Initialize()");
    try {
    } catch (ex) {
        log.exception(ex);
    }
}
function Panel2_SwitchTo() {
    _setTitleBar("Ready to Build",
                 "Komodo is ready to begin building its database");
    _setBackButton("Back", true, "B");
    _setNextButton("Build Now", true, "N");
    _setCancelButton("Cancel", true);

    var textbox = document.getElementById("panel2-textbox");
    textbox.value = '';
    if (gMode == "directory") {
        textbox.value = gDirsToScan.join('\n');
        document.getElementById("interpreterChangeNote").setAttribute('collapsed', 'true');
    } else {
        // Doing standard distros
        for (var lang in gLang2Path) {
            textbox.value += lang+" ("+gLang2Path[lang]+")\n";
        }
        document.getElementById("interpreterChangeNote").removeAttribute('collapsed');
    }
}
function Panel2_Back()
{
    log.debug("Panel2_Back()");
    try {
        if (gMode == "directory") {
            SwitchToPanel(0);
        } else if (gMode == "language") {
            SwitchToPanel(1);
        } else {
            throw("unexpected mode: '"+gMode+"'");
        }
    } catch (ex) {
        log.exception(ex);
    }
}
function Panel2_Next()
{
    log.debug("Panel2_Next()");
    try {
        SwitchToPanel(3);
    } catch (ex) {
        log.exception(ex);
    }
}


function Panel3_Initialize() {
    log.debug("Panel3_Initialize()");
    try {
    } catch (ex) {
        log.exception(ex);
    }
}
function Panel3_SwitchTo() {
    if (gMode == "upgrade") {
        _setTitleBar("Upgrading Database");
    } else {
        _setTitleBar("Building Database");
    }

    _setBackButton("Back", false, "B");
    //XXX Have to do this in a timeout or the button refuses to disable.
    window.setTimeout("_setNextButton('Next', false, 'N');", 0);
    if (gMode == "upgrade") {
        _setCancelButton("Cancel", false, null, null);
    } else {
        _setCancelButton("Cancel", true, null, "Panel3_Cancel();");
    }

    if (!gCodeIntelSvc.isBackEndActive
        && (gMode == "language" || gMode == "directory"))
    {
        try {
            gCodeIntelSvc.activateBackEnd();
        } catch(ex) {
            var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"].
                                getService(Components.interfaces.koILastErrorService);
            var err = lastErrorSvc.getLastErrorMessage();
            dialog_internalError(err, err);
            SwitchToPanel(4);
            return;
        }
    }

    var useKomodoNote = document.getElementById("useKomodoNote");
    if (gLaunchedAutomatically) {
        useKomodoNote.removeAttribute("collapsed");
    } else {
        useKomodoNote.setAttribute("collapsed", "true");
    }
    var upgradeNote = document.getElementById("upgradeNote");
    var whileWaitingNode = document.getElementById("whileWaitingNode");
    if (gMode == "directory") {
        upgradeNote.setAttribute("collapsed", "true");
        for (var i in gDirsToScan) {
            gCodeIntelSvc.batchUpdateRequest("directory", gDirsToScan[i],
                                             '*', null);
        }
    } else if (gMode == "language") {
        upgradeNote.setAttribute("collapsed", "true");
        for (var lang in gLang2Path) {
            gCodeIntelSvc.batchUpdateRequest("language", gLang2Path[lang],
                                             lang, gLang2PathExtra[lang]);
        }
    } else if (gMode == "upgrade") {
        upgradeNote.removeAttribute("collapsed");
        gCodeIntelSvc.batchUpdateRequest("upgrade", gCodeIntelSvc.cidbPath,
                                         null, "replace");
    }
    var progessUIHandler = new ProgressUIHandler();
    gCodeIntelSvc.batchUpdateStart(progessUIHandler);
}
function Panel3_Next()
{
    log.debug("Panel3_Next()");
    try {
    } catch (ex) {
        log.exception(ex);
    }
}
function Panel3_Cancel()
{
    log.debug("Panel3_Cancel()");
    try {
        // Ask the user if they really want to cancel.
        var response = dialog_yesNo(
            "Are you sure you want to cancel building Komodo's Code "+
            "Intelligence database? If you do, Komodo's Object Browser, "+
            "AutoComplete and CallTips features will not have "+
            "information on unscanned files.",
            "Yes");
        if (response == "Yes") {
            try {
                gCodeIntelSvc.batchUpdateStop();
            } catch(ex) {
                var err = "Unexpected error stopping batch update.";
                dialog_internalError(err, err+"\n"+ex);
            }
            SwitchToPanel(4);
        }
    } catch (ex) {
        log.exception(ex);
    }
}



function Panel4_Initialize() {
    log.debug("Panel4_Initialize()");
    try {
    } catch (ex) {
        log.exception(ex);
    }
}
function Panel4_SwitchTo() {
    var prefix = "Database Build";
    if (gMode == "upgrade") {
        prefix = "Database Upgrade"
    }
    if (gCompletionReason == "completed") {
        _setTitleBar(prefix+" Completed");
        gWidgets.aborted.setAttribute("collapsed", "true");
        gWidgets.erroredout.setAttribute("collapsed", "true");
        gWidgets.completed.removeAttribute("collapsed");
        while (gWidgets.completed.hasChildNodes())
            gWidgets.completed.removeChild(gWidgets.completed.firstChild);
        var msg = "Komodo's Code Intelligence database has been "+
                  "successfully updated";
        if (gMode == "language") {
            msg += " for "+_englishJoinList(gLangsToSetup, "and");
        }
        msg += ". Click Finish to exit the wizard."
        var text = document.createTextNode(msg);
        gWidgets.completed.appendChild(text);
    } else if (gCompletionReason == "erroredout") {
        _setTitleBar(prefix+" Errors");
        gWidgets.erroredout.removeAttribute("collapsed");
        if (gMode == "upgrade") {
            gWidgets.buildErrorNote.setAttribute("collapsed", "true");
            gWidgets.upgradeErrorNote.removeAttribute("collapsed");
        } else {
            gWidgets.buildErrorNote.removeAttribute("collapsed");
            gWidgets.upgradeErrorNote.setAttribute("collapsed", "true");
        }
        gWidgets.aborted.setAttribute("collapsed", "true");
        gWidgets.completed.setAttribute("collapsed", "true");
        gWidgets.errors.value = gErrorStr;
    } else { // gCompletionReason == "cancelled"
        _setTitleBar(prefix+" Aborted");
        gWidgets.aborted.removeAttribute("collapsed");
        gWidgets.erroredout.setAttribute("collapsed", "true");
        gWidgets.completed.setAttribute("collapsed", "true");
    }
    _setBackButton("Back", false, "B");
    //XXX Have to do this in a timeout or the button refuses to disable.
    window.setTimeout("_setNextButton('Finish', true, 'F');", 0);
    _setCancelButton("Cancel", false, null, "window.close();");
    gWidgets.next.focus();
}
function Panel4_Next() {
    log.debug("Panel4_Next()");
    try {
        window.close();
    } catch (ex) {
        log.exception(ex);
    }
}


function OpenRelNotes()
{
    log.debug("OpenRelNotes()");
    try {
        launch_MainHelp("relnotes.html");
    } catch (ex) {
        log.exception(ex);
    }
}

function OpenUserGuide()
{
    log.debug("OpenUserGuide()");
    try {
        launch_MainHelp("index.html");
    } catch (ex) {
        log.exception(ex);
    }
}


function Cancel()
{
    log.debug("Cancel()");
    try {
        if (gCancelOnCommand != null) {
            eval(gCancelOnCommand);
        } else {
            var response = dialog_yesNo(
                "Are you sure you want to cancel building Komodo's Code "+
                "Intelligence database? If you do, Komodo's "+
                "AutoComplete and CallTips features will not have "+
                "information on unscanned files.",
                "Yes");
            if (response == "Yes") {
                SwitchToPanel(4);
            }
        }
    } catch (ex) {
        log.exception(ex);
    }
}
