/* Copyright (c) 2000-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Komodo's Find and Replace general functions
 *
 * Presumably these are to used by the editor for findNext shortcuts and
 * by the Find and Replace dialog.
 *
 * Usage:
 *      Find_FindNext(...)
 *      Find_FindAll(...)
 *      Find_MarkAll(...)
 *      Find_Replace(...)
 *      Find_ReplaceAll(...)
 *      Find_FindAllInFiles(...)
 * where:
 *      "editor" is generally the editor object on which getFocusedView() et al
 *          can be called;
 *      "pattern" is the pattern string to search for;
 *      "replacement" is the replacement string to use; and
 *      "context" is a koIFindContext component.
 */


//---- globals

var findLog = getLoggingMgr().getLogger("find_functions");
//findLog.setLevel(LOG_DEBUG);

var findSvc = null; // the find XPCOM service that does all the grunt work

// The find session is used to hold find results and to determine when to stop
// searching (i.e. when we have come full circle).
var gFindSession = Components.classes["@activestate.com/koFindSession;1"]
                   .getService(Components.interfaces.koIFindSession);


//--- internal support routines that Komodo's view system should have

// Return the view after this one appropriate for searching. If there is no
// such view (e.g. only one view), then return null.
//
// Dev Note: This in very inefficient for a lot of open files.
//
//XXX Once all view types have viewIds we can do this more properly. I.e.
//    don't default to 'editor' view 0 as the "next" view for any non-'editor'
//    view.
function _GetNextView(editor, currView)
{
    // Filter out view types that cannot be searched.
    var views = editor.gViewMgr.topView.getViews(true);
    var searchableViews = [];
    for (var i = 0; i < views.length; ++i) {
        //XXX Should the view grow a getType() method instead of having to do
        //    this? Does _this_ even work?
        if (views[i].getAttribute("type") == "editor") {
            searchableViews.push(views[i]);
        } else {
            findLog.debug("_GetNextView: '"+views[i].getAttribute("type")+
                          "' view is not searchable");
        }
    }

    // Determine the index of the current one in this list.
    var currIndex = -1;
    if (currView.document) {
        var currDisplayPath = currView.document.displayPath;
        for (currIndex = 0; currIndex < searchableViews.length; ++currIndex) {
            if (searchableViews[currIndex].document.displayPath == currDisplayPath)
                break;
        }
    }

    // Find the next one.
    if (currIndex == -1) {
        if (searchableViews.length)
            return searchableViews[0];
        else
            return null;
    } else {
        if (searchableViews.length == 1) {
            return null;
        } else {
            var nextIndex = (currIndex + 1) % searchableViews.length;
            return searchableViews[nextIndex];
        }
    }
}

// Return the view before this one appropriate for searching. If there is no
// such view (e.g. only one view), then return null.
function _GetPreviousView(editor, currView)
{
    // Filter out view types that cannot be searched.
    var views = editor.gViewMgr.topView.getViews(true);
    var searchableViews = [];
    for (var i = 0; i < views.length; ++i) {
        //XXX Should the view grow a getType() method instead of having to do
        //    this? Does _this_ even work?
        if (views[i].getAttribute("type") == "editor") {
            searchableViews.push(views[i]);
        } else {
            findLog.debug("_GetPreviewView: '"+views[i].type+
                          "' view is not searchable");
        }
    }

    // Determine the index of the current one in this list.
    var currIndex = -1;
    if (currView.document) {
        var currDisplayPath = currView.document.displayPath;
        for (currIndex = 0; currIndex < searchableViews.length; ++currIndex) {
            if (searchableViews[currIndex].document.displayPath == currDisplayPath)
                break;
        }
    }

    // Find the next one.
    if (currIndex == -1) {
        if (searchableViews.length)
            return searchableViews[searchableViews.length-1];
        else
            return null;
    } else {
        if (searchableViews.length == 1) {
            return null;
        } else {
            // XXX JavaScript get's modulus wrong for negative values, so add the
            //     .length to ensure the value is not negative.
            var prevIndex = (currIndex - 1 + searchableViews.length)
                            % searchableViews.length;
            return searchableViews[prevIndex];
        }
    }
}

function _GetViewFromViewId(editor, displayPath)
{
    // Filter out view types that cannot be searched.
    var views = editor.gViewMgr.topView.getViews(true);
    for (var i = 0; i < views.length; ++i) {
        if (views[i].document && views[i].document.displayPath == displayPath) {
            return views[i];
        }
    }
    return null;
}



//--- internal support routines

function _SetupAndFindNext(editor, context, pattern, mode)
{
    // Return a find result from using the Find Service. Return null if the
    // find session is complete (i.e. if the pattern could not be found or if
    // repeated finds fully wrapped around the find context.)
    // "mode" is either 'find' or 'replace'. The maintenance of a find session
    //      needs to know if a find-next operation is in the context of
    //      finding or replacing.

    // abort if there is no current view
    var view = editor.gViewMgr.currentView;
    if (view == null) {
        return null;
    }

    // Get the necessary data (context may be other than the whole file).
    var scimoz = view.scintilla.scimoz;
    // url: The unique view identifier. This is not ideal. Eventually the view
    //      system should grow unique view ids that can be used here.
    var url = view.document.displayPath;
    var text; // the text to search (this is a subset of whole buffer if
              // searching within a selection)
    var contextOffset; // "text"s offset into the whole scimoz buffer
    var startOffset; // offset into "text" at which to begin searching
    if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC
        || context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS)
    {
        text = scimoz.text;
        contextOffset = 0;
        startOffset = scimoz.charPosAtPosition(-1);
    }
    else if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION)
    {
        text = scimoz.getTextRange(
                    scimoz.positionAtChar(0, context.startIndex),
                    scimoz.positionAtChar(0, context.endIndex));
        contextOffset = context.startIndex;
        startOffset = scimoz.charPosAtPosition(-1) - contextOffset;
    }
    else
    {
        throw("unexpected context: name='" + context.name + "' type=" + context.type);
    }

    // Repeated searching forward works fine because a forward search sets up
    // for the next search when it selects the find result, which, as a
    // side-effect, sets the cursor position to the end of the find result.
    // This does not work in reverse because you keep getting the same result.
    // If there is a current find result (i.e. a selection) then set the
    // startOffset to the *start* of the selection.
    if (findSvc.options.searchBackward && scimoz.selText != "") {
        startOffset -= scimoz.selText.length;
    }

    // Special case: avoid duplicate matches for special patterns that match no
    // characters.
    // If there was a previous match in this URL that matched no characters
    // (e.g. "^", "$", "(...)*"), then advance the startOffset by one character
    // to ensure that this match does not happen again (because duplicate
    // matches end a find session).
    // XXX I *think* this is a valid fix, i.e. the result is what the user
    //     expects.
    var lastFindResult = gFindSession.GetLastFindResult();
    var unadjustedStartOffset = startOffset;
    if (lastFindResult && url == lastFindResult.url
        && lastFindResult.value.length == 0)
    {
        if (findSvc.options.searchBackward) {
            startOffset -= 1;
            if (startOffset < 0) {
                startOffset = stringutils_bytelength(text.slice(0, text.length-1));
            }
        } else {
            startOffset += 1;
            if (startOffset > stringutils_bytelength(text.slice(0, text.length-1))) {
                startOffset = 0;
            }
        }
    }

    // find it
    gFindSession.StartFind(pattern, url, findSvc.options,
                           // Lie to the find session manager about the
                           // possible startOffset adjustment, else the find
                           // session will reset itself.
                           // Note: these three are *char* offsets.
                           unadjustedStartOffset + contextOffset,
                           scimoz.charPosAtPosition(scimoz.selectionStart),
                           scimoz.charPosAtPosition(scimoz.selectionEnd),
                           mode);
    var findResult = null;
    while (1) {  // while not done searching all file(s)
        findResult = findSvc.find(url, text, pattern, startOffset);
        // fix up the find result for context
        if (findResult != null) {
            findResult.start += contextOffset;
            findResult.end += contextOffset;
        }

        // if was not found then wrap (if have not already) and try again
        if (findResult == null && !gFindSession.wrapped) {
            gFindSession.wrapped = true;
            if (!findSvc.options.searchBackward) {
                startOffset = 0;
            } else {
                startOffset = stringutils_bytelength(text.slice(0, text.length-1));
            }
            continue;
        }

        // If finished searching the current document then:
        // - restore the cursor/selection state
        // - move to the next document (if we are searching more than one)
        if (findResult == null || gFindSession.WasAlreadyFound(findResult)
            || (mode == "replace" && gFindSession.IsRecursiveHit(findResult)))
        {
            // Switch to the next document to search, if there are any.
            if (context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS) {
                var newView;
                if (findSvc.options.searchBackward) {
                    newView = _GetPreviousView(editor, view);
                } else {
                    newView = _GetNextView(editor, view);
                }
                if (newView == null)
                    break;  // Either there is no view or only the one.
                view = newView;
                url = view.document.displayPath;
                // restore the cursor/selection state in current document
                scimoz.setSel(gFindSession.fileSelectionStartPos, gFindSession.fileSelectionEndPos);
                //XXX Is this bad if the working view is not visible???
                scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(scimoz.currentPos));
                // we are done if this document has already been searched
                if (gFindSession.HaveSearchedThisUrlAlready(url)) {
                    break;
                }
                // make the switch
                scimoz = view.scintilla.scimoz;
            } else {
                break;
            }

            // Save the start information in this doc.
            gFindSession.fileStartPos = scimoz.currentPos;
            gFindSession.fileSelectionStartPos = scimoz.selectionStart;
            gFindSession.fileSelectionEndPos = scimoz.selectionEnd;

            // Setup for a search in this document. (Currently only "all open
            // documents" should reach this code.)
            text = scimoz.text;
            if (findSvc.options.searchBackward) {
                startOffset = stringutils_bytelength(text.slice(0, text.length - 1));
            } else {
                startOffset = 0;
            }
            contextOffset = 0;

            continue;
        }

        // fix up the find result for context and return it
        if (findResult != null) {
            gFindSession.NoteFind(findResult);
            return findResult;
        }
    }

    // Restore the editor state to that when this find session started.
    scimoz.hideSelection(true);
    if (context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS) {
        // switch to the original url if it is still open
        var firstView = _GetViewFromViewId(editor, gFindSession.firstUrl);
        if (editor.gViewMgr.currentView.document.displayPath
            != firstView.document.displayPath) {
            firstView.makeCurrent(false);
        }
        scimoz.setSel(
            scimoz.positionAtChar(0, gFindSession.firstFileSelectionStartPos),
            scimoz.positionAtChar(0, gFindSession.firstFileSelectionEndPos));
    } else if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC) {
        scimoz.setSel(
            scimoz.positionAtChar(0, gFindSession.firstFileSelectionStartPos),
            scimoz.positionAtChar(0, gFindSession.firstFileSelectionEndPos));
    } else if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
        // We have to investigate this but find session termination is broken and so we can't
        // elicit this code path.
        scimoz.setSel(
            scimoz.positionAtChar(0, context.startIndex),
            scimoz.positionAtChar(0, context.endIndex)); // XXXX
    } else {
        throw("unexpected context: name='" + context.name + "' type=" + context.type);
    }
    scimoz.hideSelection(false);
    scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(scimoz.currentPos));
    return null;
}


function _ReplaceLastFindResult(editor, context, pattern, replacement)
{
    // Replace the last find result in the current find session, if there is
    // one and if it is genuine and current. No return value.
    var view = editor.gViewMgr.currentView;
    var scimoz = view.scintilla.scimoz;
    var url = view.document.displayPath;
    var findResult = gFindSession.GetLastFindResult();
    var replaceResult = null;
    if (findResult != null) {
        var startByte = scimoz.positionAtChar(0, findResult.start);
        var endByte = scimoz.positionAtChar(startByte,
                                            findResult.end-findResult.start);
        var findText = scimoz.getTextRange(startByte, endByte);
        var startOffset;
        var contextOffset = findResult.start;
        if (findSvc.options.searchBackward) {
            startOffset = stringutils_bytelength(findText);
        } else {
            startOffset = 0;
        }
        gFindSession.StartReplace(pattern, replacement, url, findSvc.options);
        replaceResult = findSvc.replace(url, findText, pattern,
                                        replacement, startOffset);
        if (replaceResult) {
            // fix it up
            replaceResult.start += contextOffset;
            replaceResult.end += contextOffset;

            findLog.info("replace result (from FindSvc): s/"+replaceResult.value+
                         "/"+replaceResult.replacement+"/ at "+
                         replaceResult.start+"-"+replaceResult.end);
        }


        // - make the replacement (if the findResult is genuine)
        if (replaceResult &&
            findResult.start == replaceResult.start &&
            findResult.end == replaceResult.end &&
            findText == replaceResult.value)
        {
            startByte = scimoz.positionAtChar(0, replaceResult.start);
            endByte = scimoz.positionAtChar(startByte,
                replaceResult.end-replaceResult.start);
            var byteLength = stringutils_bytelength(replaceResult.replacement);

            scimoz.targetStart = startByte;
            scimoz.targetEnd = endByte;
            scimoz.replaceTarget(byteLength, replaceResult.replacement);
            scimoz.anchor = startByte;
            scimoz.currentPos = startByte + byteLength;

            // fix up the search context as appropriate
            if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC
                || context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS)
            {
                // nothing for now
            } else if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
                // adjust the context range for the change in length
                //XXX I think perhaps context.endIndex should be measured in
                //    chars and not in bytes (as this stringutils_bytelength
                //    is converting it to).
                context.endIndex += stringutils_bytelength(replaceResult.replacement) -
                                    stringutils_bytelength(replaceResult.value);
            } else {
                throw("unexpected context: name='" + context.name + "' type=" + context.type);
            }

            // Fill in some data for possible display in the "Replace Results"
            // tab and note the replacement.
            // XXX I don't believe these Note'd replacements are ever used
            //     since ReplaceAll was re-factored.
            var startLineNum = scimoz.lineFromPosition(startByte);
            var endLineNum = scimoz.lineFromPosition(endByte);
            var contextStartPos = scimoz.positionFromLine(startLineNum);
            var contextEndPos = scimoz.getLineEndPosition(endLineNum);
            replaceResult.context_ = scimoz.getTextRange(contextStartPos, contextEndPos);
            replaceResult.line = startLineNum + 1;
            replaceResult.column = replaceResult.start -
                                   scimoz.positionFromLine(startLineNum);
            gFindSession.NoteReplace(replaceResult);
        } else {
            findLog.info("find result is not genuine: skipping replace");
        }
    } else {
        findLog.info("no last find result to replace");
    }
}


function _FindAllInView(editor, view, context, pattern, resultsView)
{
    // Find all instances of "pattern" with "replacement" in the current
    // view context.
    //
    // Note: koIFindSession.StartFind() is skipped because we don't need to
    // use this state mechanism to know when to terminate the session. Session
    // termination is handled via .NoteUrl() and .HaveSearchedThisUrlAlready().
    //
    // No return value.

    var viewId = view.document.displayPath;
    var scimoz = view.scintilla.scimoz;
    var text; // the text to search (this is a subset of whole buffer if
              // searching within a selection)
    var contextOffset; // "text"s offset into the whole scimoz buffer
    if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
        text = scimoz.getTextRange(
                    scimoz.positionAtChar(0, context.startIndex),
                    scimoz.positionAtChar(0, context.endIndex));
        contextOffset = context.startIndex;
    } else {
        text = scimoz.text;
        contextOffset = 0;
    }

    findSvc.findallex(viewId, text, pattern, resultsView, contextOffset,
                      scimoz);
    gFindSession.NoteUrl(viewId);
}


function _MarkAllInView(editor, view, context, pattern)
{
    // Mark all instances of "pattern" with "replacement" in the current
    // view context.
    //
    // Note: koIFindSession.StartFind() is skipped because we don't need to
    // use this state mechanism to know when to terminate the session. Session
    // termination is handled via .NoteUrl() and .HaveSearchedThisUrlAlready().
    //
    // Returns true if hits were found in the view, false otherwise.

    var viewId = view.document.displayPath;
    var scimoz = view.scintilla.scimoz;
    var text; // the text to search (this is a subset of whole buffer if
              // searching within a selection)
    var contextOffset; // "text"s offset into the whole scimoz buffer
    if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
        text = scimoz.getTextRange(context.startIndex, context.endIndex);
        contextOffset = context.startIndex;
    } else {
        text = scimoz.text;
        contextOffset = 0;
    }

    var countObj = new Object();
    var lines = new Array();
    lines = findSvc.findalllines(viewId, text, pattern, contextOffset,
                                 scimoz, countObj);
    var foundSome = lines.length > 0;
    _MarkLinesForViewId(editor, viewId, lines);

    gFindSession.NoteUrl(viewId);
    return foundSome;
}


function _ReplaceAllInView(editor, view, context, pattern, replacement,
                           resultsView)
{
    // Replace all instances of "pattern" with "replacement" in the current
    // view context.
    //
    // Note: koIFindSession.StartReplace() is skipped because we don't need to
    // use this state mechanism to know when to terminate the session. Session
    // termination is handled via .NoteUrl() and .HaveSearchedThisUrlAlready().
    //
    // Returns the number of replacements made.

    var viewId = view.document.displayPath;
    var scimoz = view.scintilla.scimoz;
    var text; // the text to search (this is a subset of whole buffer if
              // searching within a selection)
    var contextOffset; // "text"s offset into the whole scimoz buffer
    if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
        text = scimoz.getTextRange(
                    scimoz.positionAtChar(0, context.startIndex),
                    scimoz.positionAtChar(0, context.endIndex));
        contextOffset = context.startIndex;
    } else {
        text = scimoz.text;
        contextOffset = 0;
    }

    // Find all possible replacements in this view.
    var numReplacementsObj = new Object();
    var replacementText = findSvc.replaceallex(viewId, text, pattern,
                                               replacement, gFindSession,
                                               resultsView, contextOffset,
                                               scimoz, numReplacementsObj);
    var numReplacements = numReplacementsObj.value;

    if (replacementText != null) {
        if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
            scimoz.targetStart = scimoz.positionAtChar(0, context.startIndex);
            scimoz.targetEnd = scimoz.positionAtChar(0, context.endIndex);
            scimoz.replaceTarget(-1, replacementText);
            // Restore the selection.
            scimoz.currentPos = scimoz.positionAtChar(0, context.startIndex);
            scimoz.anchor = scimoz.positionAtChar(0, context.startIndex)
                + stringutils_bytelength(replacementText);
        } else {
            var line = scimoz.lineFromPosition(scimoz.currentPos);
            scimoz.targetStart = 0;
            scimoz.targetEnd = stringutils_bytelength(text);
            scimoz.replaceTarget(-1, replacementText);
            // Put the cursor on the same line as before replacement.
            scimoz.anchor = scimoz.currentPos = scimoz.positionFromLine(line);
        }
    }

    //XXX Correct folds. See bug 25263.
    //    Are we losing other info like bookmarks?

    //XXX Is this bad if the working view is not visible???
    scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(scimoz.currentPos));

    gFindSession.NoteUrl(viewId);
    return numReplacements;
}


function _DisplayFindResult(editor, findResult)
{
    //dump("display find result: "+findResult.url+": " + findResult.start +
    //     "-" + findResult.end + ": '" + findResult.value + "'\n");

    // Make the view for this find result current if necessary.
    // XXX Note that we are presuming that Komodo does not yet allow multiple
    //     views per document, which it eventually will. When this is possibly
    //     we must switch to storing unique view IDs in the find session.
    // XXX We are currently using view.document.displayPath as the view's
    //     "unique id". The view system does not currently provide a way to
    //     getViewFromDisplayPath() so we roll our own.
    if (editor.gViewMgr.currentView.document.displayPath != findResult.url) {
        var view = _GetViewFromViewId(editor, findResult.url);
        if (view == null) {
            var err = "The view for a find result was closed before it could "+
                      "be displayed!";
            findLog.error(err);
            // Don't raise an error. Just attempt to limp along. This method
            // has no mechanism for signalling failure.
            return;
        }
        // we don't want to focus when find dialog is up.
        var noFocus = editor != window;
        view.makeCurrent(noFocus);
    }

    // Jump to the find result and select it.
    // Jump to the _end_ (i.e. the start if going backwards) of the
    // result so that repeated finds are non-overlapping.
    var scimoz = editor.gViewMgr.currentView.scintilla.scimoz;
    // Unfortunately this is going to be very slow -- I can't come up with a
    // way around this that will actually work without rewriting the find subsystem.
    var s = stringutils_bytelength(scimoz.text.slice(0, findResult.start));
    scimoz.setSel(s, s + stringutils_bytelength(findResult.value));

    // Ensure folded lines are expanded. We *could* conceivably have search
    // results that span lines so should ensure that the whole selection is
    // visible. XXX should include every line in btwn here
    scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(findResult.start))
    scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(findResult.end))

    scimoz.chooseCaretX();
}


function _MarkLinesForViewId(editor, viewId, lines)
{
    // Mark all the given find results.
    var view = _GetViewFromViewId(editor, viewId);
    if (!view) return;

    var isBookmarkable = false;
    try {
        view.QueryInterface(Components.interfaces.koIBookmarkableView);
        isBookmarkable = true;
    } catch (ex) {}
    if (!isBookmarkable) {
        findLog.warn("Could not bookmark find results for '"+viewId+
                     "', because that view does not support koIBookmarkableView.");
        return;
    }

    for (var i = 0; i < lines.length; ++i) {
        view.addBookmark(lines[i]);
    }
}


function _UiForCompletedFindSession(editor, context)
{
    var numFinds = gFindSession.GetNumFinds();
    var numReplacements = gFindSession.GetNumReplacements();

    // Pop up an alert box with an appropriate message.
    var msg = "Komodo has finished searching " + context.name + ".";
    if (numFinds == 0) {
        msg += " The search item was not found.";
    } else if (numReplacements > 0) {
        msg += " " + numReplacements + " of " + numFinds +
               " occurrence(s) were replaced.";
    } else {
        msg += " " + numFinds + " occurrence(s) were found.";
    }
    alert(msg);
}


// Report and error in the find service to the user.
//
//  "op" indicates if the current operation was during finding or replacement.
//      It must be one of "find" or "replace".
//
function _UiForFindServiceError(op /* = "find" */)
{
    if (typeof op == 'undefined' || op == null) op = "find";

    var msg = null;
    switch (op) {
    case "find":
        msg = "There was a syntax error in your search pattern: ";
        break;
    case "replace":
        msg = "There was a syntax error in your replacement pattern: ";
        break;
    default:
        log.error("unexpected 'op' argument sent to "+
                 "_UiForFindServiceError(): op='"+op+"'\n");
    }

    var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                        .getService(Components.interfaces.koILastErrorService);
    msg += lastErrorSvc.getLastErrorMessage();
    msg += ".";
    alert(msg);
}



//---- public functions

function Find_FindNextInMacro(editor, contexttype, pattern, patternType,
                              caseSensitivity, searchBackward, matchWord,
                              mode, quiet, useMRU)
{
    // Called by macros -- only difference with Find_FindNext is that
    // contexttype is
    // a constant, and the context needs to be created.
    // XXX the APIs should be factored out so that these "inmacro" versions aren't necessary.
    var old_patternType, old_caseSensitivity, old_searchBackward, old_matchWord;
    var context = Components.classes["@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
    context.type = contexttype;

    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    // Stash old options
    old_patternType = findSvc.options.patternType;
    old_caseSensitivity = findSvc.options.caseSensitivity;
    old_searchBackward = findSvc.options.searchBackward;
    old_matchWord = findSvc.options.matchWord;

    findSvc.options.patternType = patternType;
    findSvc.options.caseSensitivity = caseSensitivity;
    findSvc.options.searchBackward = searchBackward;
    findSvc.options.matchWord = matchWord;

    Find_FindNext(editor, context, pattern, mode, quiet, useMRU);

    // restore options
    findSvc.options.patternType = old_patternType;
    findSvc.options.caseSensitivity = old_caseSensitivity;
    findSvc.options.searchBackward = old_searchBackward;
    findSvc.options.matchWord = old_matchWord;
}

function Find_FindAllInMacro(editor, contexttype, pattern, patternType,
                             caseSensitivity, searchBackward, matchWord)
{
    // Called by macros -- only difference with Find_FindAll is that
    // contexttype is a constant, the context needs to be created, and options
    // need to be set (and unset).
    // XXX the APIs should be factored out so that these "inmacro" versions
    //     aren't necessary.

    var old_patternType, old_caseSensitivity, old_searchBackward, old_matchWord;
    var context = Components.classes["@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
    context.type = contexttype;

    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    // Stash old options
    old_patternType = findSvc.options.patternType;
    old_caseSensitivity = findSvc.options.caseSensitivity;
    old_searchBackward = findSvc.options.searchBackward;
    old_matchWord = findSvc.options.matchWord;

    findSvc.options.patternType = patternType;
    findSvc.options.caseSensitivity = caseSensitivity;
    findSvc.options.searchBackward = searchBackward;
    findSvc.options.matchWord = matchWord;

    Find_FindAll(editor, context, pattern);

    // restore options
    findSvc.options.patternType = old_patternType;
    findSvc.options.caseSensitivity = old_caseSensitivity;
    findSvc.options.searchBackward = old_searchBackward;
    findSvc.options.matchWord = old_matchWord;
}


function Find_MarkAllInMacro(editor, contexttype, pattern, patternType,
                             caseSensitivity, searchBackward, matchWord)
{

    // Called by macros -- only difference with Find_MarkAll is that
    // contexttype is a constant, and the context needs to be created.
    // XXX the APIs should be factored out so that these "inmacro" versions
    //     aren't necessary.

    var old_patternType, old_caseSensitivity, old_searchBackward, old_matchWord;
    var context = Components.classes["@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
    context.type = contexttype;

    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    // Stash old options
    old_patternType = findSvc.options.patternType;
    old_caseSensitivity = findSvc.options.caseSensitivity;
    old_searchBackward = findSvc.options.searchBackward;
    old_matchWord = findSvc.options.matchWord;

    findSvc.options.patternType = patternType;
    findSvc.options.caseSensitivity = caseSensitivity;
    findSvc.options.searchBackward = searchBackward;
    findSvc.options.matchWord = matchWord;

    Find_MarkAll(editor, context, pattern);

    // restore options
    findSvc.options.patternType = old_patternType;
    findSvc.options.caseSensitivity = old_caseSensitivity;
    findSvc.options.searchBackward = old_searchBackward;
    findSvc.options.matchWord = old_matchWord;
}

function Find_ReplaceInMacro(editor, contexttype, pattern, replacement,
                             patternType, caseSensitivity, searchBackward, matchWord)
{
    // Called by macros -- only difference with Find_Replace is that
    // contexttype is a constant, and the context needs to be created.
    // XXX the APIs should be factored out so that these "inmacro" versions
    //     aren't necessary.

    var old_patternType, old_caseSensitivity, old_searchBackward, old_matchWord;
    var context = Components.classes["@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
    context.type = contexttype;

    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    // Stash old options
    old_patternType = findSvc.options.patternType;
    old_caseSensitivity = findSvc.options.caseSensitivity;
    old_searchBackward = findSvc.options.searchBackward;
    old_matchWord = findSvc.options.matchWord;

    findSvc.options.patternType = patternType;
    findSvc.options.caseSensitivity = caseSensitivity;
    findSvc.options.searchBackward = searchBackward;
    findSvc.options.matchWord = matchWord;

    Find_Replace(editor, context, pattern, replacement);

    // restore options
    findSvc.options.patternType = old_patternType;
    findSvc.options.caseSensitivity = old_caseSensitivity;
    findSvc.options.searchBackward = old_searchBackward;
    findSvc.options.matchWord = old_matchWord;
}


function Find_ReplaceAllInMacro(editor, contexttype, pattern, replacement, quiet,
                                patternType, caseSensitivity, searchBackward, matchWord)
{
    // Called by macros -- only difference with Find_Replace is that
    // contexttype is a constant, and the context needs to be created.
    // XXX the APIs should be factored out so that these "inmacro" versions
    //     aren't necessary.

    var old_patternType, old_caseSensitivity, old_searchBackward, old_matchWord;
    var context = Components.classes["@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
    context.type = contexttype;

    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    // Stash old options
    old_patternType = findSvc.options.patternType;
    old_caseSensitivity = findSvc.options.caseSensitivity;
    old_searchBackward = findSvc.options.searchBackward;
    old_matchWord = findSvc.options.matchWord;

    findSvc.options.patternType = patternType;
    findSvc.options.caseSensitivity = caseSensitivity;
    findSvc.options.searchBackward = searchBackward;
    findSvc.options.matchWord = matchWord;

    Find_ReplaceAll(editor, context, pattern, replacement, quiet);

    // restore options
    findSvc.options.patternType = old_patternType;
    findSvc.options.caseSensitivity = old_caseSensitivity;
    findSvc.options.searchBackward = old_searchBackward;
    findSvc.options.matchWord = old_matchWord;
}


// Find (and move to) the next occurrence of the given pattern.
//
//  "editor" is a reference to the komodo.xul main window
//  "context" is a koIFindContext instance
//  "pattern" is the pattern being sought
//  "mode" (optional) is either "find" (the default) or "replace",
//      indicating if this is in the context of a replace UI.
//  "quiet" (optional) is a boolean indicating if UI (e.g. dialogs) may
//      be used for presenting info, e.g. wrapped the document. Interactive
//      search typically sets this to true. Default is false.
//  "useMRU" (optional) is a boolean indicating if the pattern should be
//      loaded into the find MRU. Default is true.
function Find_FindNext(editor, context, pattern, mode /* ="find" */,
                       quiet /* =false */, useMRU /* =true */)
{
    if (typeof(mode) == 'undefined' || mode == null) mode = "find";
    if (typeof(quiet) == 'undefined' || quiet == null) quiet = false;
    if (typeof(useMRU) == 'undefined' || useMRU == null) useMRU = true;
    findLog.info("Find_FindNext(editor, context, pattern='"+pattern+
                 "', mode="+mode+", quiet="+quiet+", useMRU"+useMRU+")");

    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"].
                  getService(Components.interfaces.koIFindService);
    }
    if (editor.gMacroMgr.mode == 'recording') {
        editor.gMacroMgr.appendCode("Find_FindNextInMacro(window, " +
                                    context.type + ", '" +
                                    pattern.replace('\'', '\\\'', 'g') + "', " +
                                    findSvc.options.patternType + ", " +
                                    findSvc.options.caseSensitivity + ", " +
                                    findSvc.options.searchBackward + ", " +
                                    findSvc.options.matchWord + ", " +
                                    mode + ", " +
                                    quiet + ", " +
                                    useMRU +
                                    ");\n");
    }

    // Find the next match given the current state of the editor buffer and
    // do all the UI work to display the found result.
    // "mode" is either 'find' or 'replace'. The maintenance of a find session
    //      needs to know if a find-next operation is in the context of a
    //      finding or replacing.
    // Returns true iff a result was found and displayed.

    // Find the next "hit". If no result is returned then the find session is
    // complete.
    var findResult = null;
    try {
        findResult = _SetupAndFindNext(editor, context, pattern, mode);
    } catch(e) {
        if (!quiet)
            _UiForFindServiceError();
        return null;
    }
    if (useMRU)
        MRU_add("find-patternMru", pattern, true);

    if (findResult) {
        _DisplayFindResult(editor, findResult);
    } else {
        if (!quiet)
            _UiForCompletedFindSession(editor, context);
        //log.debug("Reset find session, because 'FindNext' did not find "
        //         + "anything.");
        gFindSession.Reset();
    }
    return findResult != null;
}


function Find_FindAll(editor, context, pattern, patternAlias)
{
    findLog.info("Find_FindAll(editor, context, pattern='"+pattern+
                 "', patternAlias='"+patternAlias+"')");
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"].
                  getService(Components.interfaces.koIFindService);
    }
    if (editor.gMacroMgr.mode == 'recording') {
        editor.gMacroMgr.appendCode("Find_FindAllInMacro(window, " +
                                    context.type + ", '" +
                                    pattern.replace('\'', '\\\'', 'g') + "', " +
                                    findSvc.options.patternType + ", " +
                                    findSvc.options.caseSensitivity + ", " +
                                    findSvc.options.searchBackward + ", " +
                                    findSvc.options.matchWord + ", " +
                                    ");\n");
    }

    var preferredResultsTab = findSvc.options.displayInFindResults2 ? 2 : 1;
    var resultsMgr = editor.FindResultsTab_GetTab(preferredResultsTab);
    if (resultsMgr == null)
        return null;
    resultsMgr.configure(pattern, patternAlias, null, context,
                         findSvc.options);
    resultsMgr.show();

    resultsMgr.searchStarted();
    var numFiles = null;
    try {
        if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC
            || context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
            findLog.debug("Find_FindAll: find all in '"+
                          editor.gViewMgr.currentView.document.displayPath+"'\n");
            _FindAllInView(editor, editor.gViewMgr.currentView, context,
                           pattern, resultsMgr.view);
        } else if (context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS) {
            var view = editor.gViewMgr.currentView;
            var viewId;
            numFiles = 0;
            while (view) {
                viewId = view.document.displayPath;
                if (gFindSession.HaveSearchedThisUrlAlready(viewId)) {
                    findLog.debug("Find_FindAll: have already searched '"+
                                  viewId+"'\n");
                    break;
                }

                findLog.debug("Find_FindAll: find all in '"+viewId+"'\n");
                _FindAllInView(editor, view, context, pattern, resultsMgr.view);
                numFiles += 1;

                if (findSvc.options.searchBackward) {
                    view = _GetPreviousView(editor, view);
                } else {
                    view = _GetNextView(editor, view);
                }
            }
        } else {
            throw("unexpected context: name='" + context.name + "' type=" +
                  context.type);
        }
        resultsMgr.searchFinished(true, resultsMgr.view.rowCount, numFiles);
    } catch(ex) {
        _UiForFindServiceError();
        return null;
    }


    MRU_add("find-patternMru", pattern, true);
    var foundSome = resultsMgr.view.rowCount > 0;
    if (! foundSome) {
        var msg;
        if (typeof(patternAlias) != 'undefined' && patternAlias) {
            msg = "No "+patternAlias+" were found in "+context.name+".";
        } else {
            msg = "'"+pattern+"' was not found in "+context.name+".";
        }
        alert(msg);
    }
    gFindSession.Reset();

    return foundSome;
}


function Find_MarkAll(editor, context, pattern, patternAlias)
{
    // Returns true iff any matches were found and displayed.
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }
    if (editor.gMacroMgr.mode == 'recording') {
        editor.gMacroMgr.appendCode("Find_MarkAllInMacro(window, " +
                                    context.type + ", '" +
                                    pattern.replace('\'', '\\\'', 'g') + "', " +
                                    findSvc.options.patternType + ", " +
                                    findSvc.options.caseSensitivity + ", " +
                                    findSvc.options.searchBackward + ", " +
                                    findSvc.options.matchWord + ", " +
                                    ");\n");
    }

    var foundSome = false;
    var rv
    if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC
        || context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
        rv = _MarkAllInView(editor, editor.gViewMgr.currentView, context,
                            pattern);
        if (rv) foundSome = true;
    } else if (context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS) {
        var view = editor.gViewMgr.currentView;
        var viewId;
        while (view) {
            viewId = view.document.displayPath;
            if (gFindSession.HaveSearchedThisUrlAlready(viewId)) {
                findLog.debug("Find_MarkAll: have alread searched '"+
                              viewId+"'\n");
                break;
            }

            rv = _MarkAllInView(editor, view, context, pattern);
            if (rv) foundSome = true;

            if (findSvc.options.searchBackward) {
                view = _GetPreviousView(editor, view);
            } else {
                view = _GetNextView(editor, view);
            }
        }
    } else {
        throw("unexpected context: name='" + context.name + "' type=" +
              context.type);
    }

    MRU_add("find-patternMru", pattern, true);
    if (! foundSome) {
        var msg;
        if (typeof(patternAlias) != 'undefined' && patternAlias) {
            msg = "No "+patternAlias+"' were found in "+context.name+".";
        } else {
            msg = "'"+pattern+"' was not found in "+context.name+".";
        }
        alert(msg);
    }
    gFindSession.Reset();

    return foundSome;
}


function Find_Replace(editor, context, pattern, replacement)
{
    // Replace the currently selected find result (if there is one) and find
    // the next occurrence of the current pattern. Returns true iff a next
    // occurence of the pattern was found and hilighted.
    findLog.info("Find_Replace(editor, context, pattern='"+pattern+
                 "', replacement='"+replacement+"')");
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"].
                  getService(Components.interfaces.koIFindService);
    }

    if (editor.gMacroMgr.mode == 'recording') {
        editor.gMacroMgr.appendCode("Find_ReplaceInMacro(window, " +
                                    context.type + ", '" +
                                    pattern.replace('\'', '\\\'', 'g') + "', '" +
                                    replacement.replace('\'', '\\\'', 'g') + "', " +
                                    findSvc.options.patternType + ", " +
                                    findSvc.options.caseSensitivity + ", " +
                                    findSvc.options.searchBackward + ", " +
                                    findSvc.options.matchWord +
                                    ");\n");
    }

    var view = editor.gViewMgr.currentView;
    var findResult = null;
    if (view != null) {
        // Here is how a replace works:
        //  - the first "Replace" will just identify the replacement to make
        //  - the second "Replace" will make that replacement and find the next
        //  - and so on...

        // replace the current find result if there is one
        try {
            findLog.info("Find_Replace: replace last find result");
            _ReplaceLastFindResult(editor, context, pattern, replacement);
        } catch(e) {
            _UiForFindServiceError("replace");
            return null;
        }

        // find the next one
        try {
            findLog.info("Find_Replace: find next hit");
            findResult = _SetupAndFindNext(editor, context, pattern,
                                           'replace');
            if (findResult) {
                findLog.info("Find_Replace: found next hit: "+
                             findResult.start+"-"+findResult.end+": '"+
                             findResult.value+"'");
            }
        } catch(e) {
            _UiForFindServiceError();
            return null;
        }
        MRU_add("find-patternMru", pattern, true);

        // Do the appropriate UI work for results.
        if (findResult == null) {
            _UiForCompletedFindSession(editor, context);
            //findLog.debug("Reset find session, because 'Replace' action "
            //              + "returned no result.");
            gFindSession.Reset();
        } else {
            _DisplayFindResult(editor, findResult);
        }
    }
    return findResult != null;
}


function Find_ReplaceAll(editor, context, pattern, replacement,
                         showReplaceResults /* =false */)
{
    if (typeof(showReplaceResults) == "undefined") showReplaceResults = false;
    findLog.info("Find_ReplaceAll(editor, context, pattern='"+pattern+
                 "', replacement='"+replacement+"', showReplaceResults="+
                 showReplaceResults+")\n");

    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"].
                  getService(Components.interfaces.koIFindService);
    }

    if (editor.gMacroMgr.mode == 'recording') {
        editor.gMacroMgr.appendCode("Find_ReplaceAllInMacro(window, " +
                                    context.type + ", '" +
                                    pattern.replace('\'', '\\\'', 'g') + "', '" +
                                    replacement.replace('\'', '\\\'', 'g') + "', " +
                                    showReplaceResults + ", " +
                                    findSvc.options.patternType + ", " +
                                    findSvc.options.caseSensitivity + ", " +
                                    findSvc.options.searchBackward + ", " +
                                    findSvc.options.matchWord +
                                    ");\n");
    }

    var resultsMgr = null;
    var resultsView = null;
    var numFiles = null;
    if (showReplaceResults) {
        var preferredResultsTab = findSvc.options.displayInFindResults2 ? 2 : 1;
        resultsMgr = editor.FindResultsTab_GetTab(preferredResultsTab);
        if (resultsMgr == null)
            return false;
        resultsMgr.configure(pattern, null, replacement, context,
                             findSvc.options);
        resultsMgr.show();
        resultsView = resultsMgr.view;
        resultsMgr.searchStarted();

        // Should add results already in session.
        var countObj = new Object();
        var replaceResults = new Object();
        rrs = gFindSession.GetReplacements(countObj);
        var i, rr;
        for (i = 0; i < rrs.length; ++ i) {
            rr = rrs[i];
            resultsMgr.addReplaceResult(
                rr.url, rr.start, rr.end, rr.value, rr.replacement, rr.url,
                rr.line, rr.column, rr.context_);
        }
    }

    var numReplacements = gFindSession.GetNumReplacements();
    var nr;
    if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC
        || context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
        findLog.debug("Find_ReplaceAll: replace all in '"+
                      editor.gViewMgr.currentView.document.displayPath+"'\n");
        nr = _ReplaceAllInView(editor, editor.gViewMgr.currentView, context,
                               pattern, replacement, resultsView);
        numReplacements += nr;
    } else if (context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS) {
        var view = editor.gViewMgr.currentView;
        var viewId;
        numFiles = 0;
        while (view) {
            viewId = view.document.displayPath;
            if (gFindSession.HaveSearchedThisUrlAlready(viewId)) {
                findLog.debug("Find_ReplaceAll: have alread searched '"+
                              viewId+"'\n");
                break;
            }

            findLog.debug("Find_ReplaceAll: replace all in '"+viewId+"'\n");
            nr = _ReplaceAllInView(editor, view, context, pattern,
                                   replacement, resultsView);
            numReplacements += nr;
            numFiles += 1;

            if (findSvc.options.searchBackward) {
                view = _GetPreviousView(editor, view);
            } else {
                view = _GetNextView(editor, view);
            }
        }
    } else {
        throw("unexpected context: name='" + context.name + "' type=" +
              context.type);
    }
    if (showReplaceResults) {
        resultsMgr.searchFinished(true, numReplacements, numFiles);
    }

    MRU_add("find-patternMru", pattern, true);
    var msg;
    if (numReplacements == 0) {
        msg = "'"+pattern+"' was not found in "+context.name+
                  ". No changes were made.";
        alert(msg);
    } else {
        msg = "Made "+numReplacements+" replacements in "+
                  context.name+".";
        editor.StatusBar_AddMessage(msg, "find", 3000, true);
    }
    gFindSession.Reset();

    return numReplacements != 0;
}


function Find_FindAllInFiles(editor, context, pattern, patternAlias)
{
    findLog.info("Find_FindAllInFiles(editor, context, pattern='"+pattern+
                 "', patternAlias='"+patternAlias+"')");
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"].
                  getService(Components.interfaces.koIFindService);
    }

    //XXX:TODO macro recording stuff for this

    var preferredResultsTab = findSvc.options.displayInFindResults2 ? 2 : 1;
    var resultsMgr = editor.FindResultsTab_GetTab(preferredResultsTab);
    if (resultsMgr == null)
        return false;
    resultsMgr.configure(pattern, patternAlias, null, context,
                         findSvc.options);
    resultsMgr.show();

    // This will kick off a thread to fill in the given tree/outliner.
    try {
        findSvc.findallinfiles(resultsMgr.id, pattern, resultsMgr,
                           resultsMgr.view);
    } catch(e) {
        _UiForFindServiceError();
        resultsMgr.clear();
        gWidgets.panel.pattern.focus();
        return false;
    }
    return true;
//XXX Perhaps this should be done here? Or in the dialog.
//    MRU_add("find-patternMru", pattern, true);
}


