/* Copyright (c) 2000-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

// -*- Mode: JavaScript; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
//
// Assitance for Komodo's Run Command feature set.
//
// Notes on MRUs:
// - The Run Command system currently used three MRUs stored under the
//   following pref names:
//    'run-commandMru' - stores _encoded_ command strings, used for
//        Tools->Recent Commands
//    'run-commandStringMru' - stores just the command string, used for
//        command textbox MRU
//    'run-parseRegexMru' - stores just parse regex strings, used for the
//        parseRegex textbox MRU.
//

//---- globals

var runLog = getLoggingMgr().getLogger("run_functions");
//runLog.setLevel(LOG_DEBUG);

var _gRunSvc = Components.classes["@activestate.com/koRunService;1"]
               .getService(Components.interfaces.koIRunService);
var _run_gTerminationCallback = null;

//---- internal utility routines

function _Run_UpdateMRU(command, encodedCommand, cwd, parseRegex)
{
    MRU_add("run-commandMru", encodedCommand, true /* case-sensitive */);
    MRU_add("run-commandStringMru", command, true /* case-sensitive */);
    if (cwd) {
        MRU_add("run-cwdMru", cwd, true /* case-sensitive */);
    }
    if (parseRegex) {
        MRU_add("run-parseRegexMru", parseRegex, true /* case-sensitive */);
    }
}


// Keep a list of processes spawned by the RunCommand system. This is necessary
// to ensure that there are no hangs when Komodo closes because some process is
// still running. Before closing Komodo all process must be terminated.
function _Run_RegisterProcess(editor, command, process)
{
    //dump("_Run_RegisterProcess(command='"+command+"', process='"+
    //     process+"')\n");

    // XXX There is some timing issue with process termination. For quick
    //     commands (e.g. 'echo hi'), the process may be a zombie before it
    //     gets here. We MUST, however, maintain a reference to it, at least
    //     until the next time around because some related things need to clean
    //     up before the ProcessProxy object destructs. Otherwise Komodo will
    //     hang. Don't know the details.
    _Run_UnregisterZombieProcess(editor);

    var p = {};
    p['command'] = command;
    p['process'] = process;
    editor.gRunProcesses[editor.gRunProcesses.length] = p;
}


// Remove any processes that have terminated from the list.
function _Run_UnregisterZombieProcess(editor)
{
    var newList = [];
    for (var i=0; i < editor.gRunProcesses.length; i++) {
        var p = editor.gRunProcesses[i]['process'];
        try {
            p.wait(0);
            //dump("_Run_UnregisterZombieProcess: zombie: index="+i+
            //     ", command='"+editor.gRunProcesses[i]['command']+"'\n");
        } catch (ex) {
            // XXX Could getLastError here to determine if the wait failed
            //     because of WAIT_TIMEOUT or WAIT_FAILED. Should probably not
            //     ignore the latter, but punting on that for now.
            newList[newList.length] = editor.gRunProcesses[i];
        }
    }
    editor.gRunProcesses = newList;
}



//---- support classes

// koIRunTerminationListener implementation whose only job is to call
// notify on child termination.
// Note: This is psychotic how much code you need to do this. Do I have
//       to have a separate file for this? See
//       http://lxr.mozilla.org/seamonkey/source/xpcom/sample/nsSample.js
function _koRunTerminationListener() { }
_koRunTerminationListener.prototype = {
    init: function (editor, command, callback) {
        this._editor = editor;
        this._command = command;
        this._callback = callback;
    },
    onTerminate: function (retval) {
        //dump("_koRunTerminationListener::onTerminate(retval="+retval+")\n");
        this._editor.RunOutput_EndSession(retval);
        var msg = "'" + this._command + "' returned " + retval + ".";
        this._editor.StatusBar_AddMessage(msg, "run_command", 3000,
                                          retval ? 1 : 0);
        if (this._callback) {
            // Want the callback to come _after_ this thread of control so call
            // it in a timeout. This requires the use of a global so it is
            // *possible* that .onTerminate() calls from separate threads
            // *could* result in a callback not getting called. That would be a
            // huge fluke though.
            _run_gTerminationCallback = this._callback;
            window.setTimeout(
                "_run_gTerminationCallback("+retval+"); _run_gTerminationCallback=null;",
                10);
        }
    },
    QueryInterface: function (iid) {
        if (!iid.equals(Components.interfaces.koIRunTerminationListener) &&
            !iid.equals(Components.interfaces.nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    }
}

var _koRunTerminationListenerModule = {
    myCID: Components.ID("{6EB8049A-ED64-4959-8D6E-0ED16C3BF9EA}"),
    myProgID: "@activestate.com/koIRunTerminationListener;1",

    firstTime: true,
    registerSelf: function (compMgr, fileSpec, location, type) {
        if (this.firstTime) {
            this.firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        compMgr = compMgr.QueryInterface(
            Components.interfaces.nsIComponentRegistrar);
        compMgr.registerFactoryLocation(this.myCID,
                                        "Sample JS Component",
                                        this.myProgID,
                                        fileSpec,
                                        location,
                                        type);
    },

    getClassObject: function (compMgr, cid, iid) {
        if (!cid.equals(this.myCID))
            throw Components.results.NS_ERROR_NO_INTERFACE;
        if (!iid.equals(Components.interfaces.nsIFactory))
            throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
        return this.myFactory;
    },

    myFactory: {
        createInstance: function (outer, iid) {
            if (outer != null)
                throw Components.results.NS_ERROR_NO_AGGREGATION;
            return (new _koRunTerminationListener()).QueryInterface(iid);
        }
    },

    canUnload: function(compMgr) {
        return true;
    }
};

function NSGetModule(compMgr, fileSpec) {
    return _koRunTerminationListenerModule;
}



//---- public interface

function Run_BuildRecentCommandsMenu(popupWidget)
{
    var globalPrefs = Components.classes["@activestate.com/koPrefService;1"]
                      .getService(Components.interfaces.koIPrefService).prefs;

    var mruPrefName = "run-commandMru";
    var mruList = globalPrefs.hasPref(mruPrefName) ?
                  globalPrefs.getPref(mruPrefName) : null;

    // clear the current list
    while (popupWidget.childNodes.length != 0)  {
        var child = popupWidget.firstChild;
        popupWidget.removeChild(child);
    }

    // if the MRU is empty then say so
    var itemWidget = null;
    if (!mruList || mruList.length == 0) {
        itemWidget = document.createElement("menuitem");
        itemWidget.setAttribute("label", "(Empty)");
        itemWidget.setAttribute("disabled", "true");
        popupWidget.appendChild(itemWidget);
    }

    // otherwise, populate the menupopup
    else {
        for(var i = 0; i < mruList.length; i++) {
            itemWidget = document.createElement("menuitem");

            // special case: if there are no options then strip the leading
            // "[] " for display to the user (XXX this presumes knowledge of
            // the encoding scheme)
            var encodedCommand = mruList.getStringPref(i);
            var displayCommand = encodedCommand;
            if (encodedCommand.substring(0,3) == "[] ") {
                displayCommand = encodedCommand.substring(3,
                                   encodedCommand.length);
            }
            itemWidget.setAttribute("label", (i+1) + " " + displayCommand);

            // escape char that will get interpreter twice (XXX not confident
            // that this covers all cases)
            encodedCommand = encodedCommand.replace('\\', '\\\\', "g");
            encodedCommand = encodedCommand.replace("'", "\\'", "g");
            encodedCommand = encodedCommand.replace('"', '\\"', "g");
            var handler = "Run_RunEncodedCommand(window, '"
                          + encodedCommand + "');";
            itemWidget.setAttribute("oncommand", handler);

            if (i <= 8) {
                itemWidget.setAttribute("accesskey", ""+(i+1));
            } else if (i == 9) {
                itemWidget.setAttribute("accesskey", "0");
            }

            popupWidget.appendChild(itemWidget);
        }
    }
}


function Run_RunEncodedCommand(editor, encodedCommand)
{
    // decode the command string and then run it
    var commandObj = new Object();
    var cwdObj = new Object();
    var envObj = new Object();
    var insertOutputObj = new Object();
    var operateOnSelectionObj = new Object();
    var doNotOpenOutputWindowObj = new Object();
    var runInObj = new Object();
    var parseOutputObj = new Object();
    var parseRegexObj = new Object();
    var showParsedOutputListObj = new Object();
    _gRunSvc.Decode(encodedCommand, commandObj, cwdObj, envObj,
                    insertOutputObj, operateOnSelectionObj,
                    doNotOpenOutputWindowObj, runInObj,
                    parseOutputObj, parseRegexObj, showParsedOutputListObj);

    return Run_RunCommand(editor, commandObj.value, cwdObj.value, envObj.value,
                          insertOutputObj.value, operateOnSelectionObj.value,
                          doNotOpenOutputWindowObj.value, runInObj.value,
                          parseOutputObj.value, parseRegexObj.value,
                          showParsedOutputListObj.value);
}


// Run the given command.
//  "editor" is the main editor window JS object.
//  "command" is the command string to spawn.
//  "cwd" is the directory to start the child process in (null to inherit from
//      the parent).
//  "env" is an encoded string of environment changes to make for the child
//      (null to inherit from the parent). Encoding scheme:
//              "VAR1=value1\nVAR2=value2\n..."
//  "insertOutput" (boolean) specifies whether to insert the command's output
//      into the current insertion point.
//  "operateOnSelection" (boolean) specifies whether to use the current
//      selection as input to the given command.
//  "doNotOpenOutputWindow" (boolean) specifies that the Command Output window
//      should *not* be opened.
//  "runIn" (string) identifies where the command should be run, e.g. in a new
//      console window, in the command output pane (this is only meaningful if
//      the command will be run interactively). See the koIPart_command IDL for
//      a description of the exact strings to use.
//  "parseOutput" (boolean) indicates whether to parse output lines with the
//      given regular expression, "parseRegex".
//  "parseRegex" (string) is a regex pattern to use to parse the output lines
//      into the command output window tree.
//  "showParsedOutputList" (boolean) indicates whether to view parsed results
//      in the tree.
//  "name" (string) is an optional name for the command being run. Sometimes
//      this name is useful when interacting with the user.
//  "clearOutputWindow" (boolean). By default the command output window is
//      cleared before starting a new command. This option is only relevant
//      when runIn=="command-output-window". NOTE: This option is currently NOT
//      exposed via the Command UI, nor is it persisted in koIPart_command's.
//      YAGNI.
//  "terminationCallback" (function) is called after the command terminates.
//      XXX: This is currently only implemented for
//           runIn=="command-output-window" for simplicity sake.
//
// Returns true iff the command was launched successfully.
//
function Run_RunCommand(editor, command, cwd, env, insertOutput,
                        operateOnSelection, doNotOpenOutputWindow, runIn,
                        parseOutput /* =false */, parseRegex /* ='' */,
                        showParsedOutputList /* =false */,
                        name /* =null */,
                        clearOutputWindow /* =true */,
                        terminationCallback /* =null */,
                        saveInMRU, /* =true */
                        saveInMacro /* =true */)
{
try {
    if (typeof parseOutput == 'undefined' || parseOutput == null) parseOutput = false;
    if (typeof parseRegex == 'undefined' || parseRegex == null) parseRegex = "";
    if (typeof showParsedOutputList == 'undefined' || showParsedOutputList == null) showParsedOutputList = false;
    if (typeof name == 'undefined') name = null;
    if (typeof clearOutputWindow == 'undefined' || clearOutputWindow == null) clearOutputWindow = true;
    if (typeof terminationCallback == 'undefined') terminationCallback = null;
    if (typeof saveInMRU == 'undefined' || saveInMRU == null) saveInMRU = true;
    if (typeof saveInMacro == 'undefined' || saveInMacro == null) saveInMacro = true;
    //dump("Run_RunCommand(editor, command='"+command+"', cwd='"+cwd+
    //     "', env, insertOutput='"+insertOutput+
    //     "', operateOnSelection='"+operateOnSelection+
    //     "', doNotOpenOutputWindow='"+doNotOpenOutputWindow+
    //     "', runIn='"+runIn+"', parseOutput='"+parseOutput+
    //     "', parseRegex='"+parseRegex+
    //     "', showParsedOutputList='"+showParsedOutputList+
    //     "', name='"+name+
    //     "', clearOutputWindow='"+clearOutputWindow+
    //     "', terminationCallback='"+terminationCallback+"')\n");
    var obj, errmsg;

    var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                       .getService(Components.interfaces.koILastErrorService);

    var view = editor.gViewMgr.currentView;
    var scimoz = null;
    var filename = null;
    var url = null;
    if (view && view.getAttribute("type") == "editor") {
        scimoz = view.scintilla.scimoz;
        filename = view.document.displayPath;
        if (view.document.file) url = view.document.file.URI;
    }

    // Ensure that it is reasonable to run this command given the current state
    // of Komodo.
    if (operateOnSelection) {
        if (!view) {
            alert("Cannot 'operate on selection' because there is " +
                  "no current file.");
            return false;
        } else if (view.getAttribute("type") != "editor") {
            alert("Do not know how to 'operate on selection' on "+
                  "non-editor views.");
            return false;
        } else if (scimoz.selText == "") {
            alert("Cannot 'operate on selection' because the current " +
                  "file does not have a selection.");
            return false;
        }
    }
    if (insertOutput) {
        if (!view) {
            alert("Cannot 'insert output' because there is no current " +
                  "file in which to insert.");
            return false;
        } else if (view.getAttribute("type") != "editor") {
            alert("Do not know how to 'insert output' into "+
                  "non-editor views.");
            return false;
        }
    }

    var icommand = null;
    var icommandForDisplay = null;
    var icwd = null;
    var icwdForDisplay = null;
    var ienv = null;
    var ienvForDisplay = null;
    var env_in = [];
    if (env) env_in = [env];
    try {
        var istrings;
        if (cwd == null) {
            istrings = Interpolate_interpolate(editor, [command],
                                               env_in, name);
            icommand = istrings[0];
            icommandForDisplay = istrings[1];
            if (env) {
                ienv = istrings[2];
                ienvForDisplay = istrings[3];
            }
        } else {
            istrings = Interpolate_interpolate(editor, [command, cwd],
                                               env_in, name);
            icommand = istrings[0];
            icommandForDisplay = istrings[1];
            icwd = istrings[2];
            icwdForDisplay = istrings[3];
            if (env) {
                ienv = istrings[4];
                ienvForDisplay = istrings[5];
            }
        }
    } catch (ex) {
        var errno = lastErrorSvc.getLastErrorCode();
        if (errno == Components.results.NS_ERROR_ABORT) {
            // Command was cancelled.
        } else if (errno == Components.results.NS_ERROR_INVALID_ARG) {
            errmsg = lastErrorSvc.getLastErrorMessage();
            alert("Error running command: " + errmsg);
        } else {
            runLog.error(ex);
            alert("There was an unexpected error: " + ex);
        }
        return false;
    }

    // Run the command.
    var retval = null;
    var termListener = null;
    var process = null;
    var input = null;
    var output = null;
    var error = null;
    errmsg = null;
    if (operateOnSelection && scimoz && scimoz.selText != "") {
        input = scimoz.selText;
    }

    if (insertOutput) {
        // Start the command and monitor it. If the process is taking a long
        // time to complete then pop up a dialog to wait for its termination
        // giving the user the option to abort the command by killing the
        // process.
        //dump("XXX RunAndNotify: about to start '"+icommand+"'\n");
        try {
            process = _gRunSvc.RunAndNotify(icommand, icwd, ienv, input);
        } catch (ex) {
            //dump("XXX RunAndNotify: '"+icommand+"' failed to start\n");
            errmsg = lastErrorSvc.getLastErrorMessage();
            if (! errmsg) {
                errmsg = "Unknown error running: '"+icommand+"'";
            }
            alert(errmsg);
            return false;
        }
        //dump("XXX RunAndNotify: '"+icommand+"' started normally\n");

        var cancelled = false;
        try {
            process.wait(2);  // Give the command a little bit of time to run.
            //dump("XXX RunAndNotify: child finished within 2s\n");
        } catch (ex) {
            // XXX Could getLastError here to determine if the wait failed
            //     because of WAIT_TIMEOUT or WAIT_FAILED. Should probably not
            //     ignore the latter, but punting on that for now.
            //dump("XXX RunAndNotify: 2s passed, child has NOT finished\n");
            obj = new Object();
            obj.process = process;
            obj.command = icommand;
            openWindowUniqueInstance(
                "chrome://jaguar/content/run/waitfortermination.xul",
                "komodo_waitfortermination",
                "chrome,close=no,modal=yes,resizable=yes", obj);
            if (obj.retval == "killed") { // Cancelled.
                cancelled = true;
            }
        }

        if (! cancelled) {
            //dump("XXX RunAndNotify: get output and error\n");
            output = process.readStdout();
            error = process.readStderr();
            //dump("XXX RunAndNotify: finished getting output and error\n");
        }

    } else {
        if (runIn == "command-output-window") {
            try {
                editor.RunOutput_StartSession(icommandForDisplay, parseOutput,
                                              parseRegex, icwdForDisplay, filename,
                                              clearOutputWindow);
            } catch (ex) {
                alert(ex);
                return false;
            }
            var terminal = editor.RunOutput_GetTerminal();

            // Setup the termination listener.
            termListener = new _koRunTerminationListener();
            termListener.init(editor, icommandForDisplay, terminationCallback);

            // Start the command.
            try {
                process = _gRunSvc.RunInTerminal(icommand, icwd, ienv,
                                                 terminal, termListener,
                                                 input);
            } catch (ex) {
                errmsg = lastErrorSvc.getLastErrorMessage();
                if (! errmsg) {
                    errmsg = "Unknown error running: '"+icommandForDisplay+"'";
                }
                alert(errmsg);
                editor.RunOutput_EndSession(-1);
                return false;
            }

            // Register the command and show the output window.
            // XXX Should perhaps pass in cwd here and display that in
            //     the command description for the user.
            _Run_RegisterProcess(editor, icommandForDisplay, process);
            editor.RunOutput_SetProcessHandle(process);
            if (! doNotOpenOutputWindow) {
                editor.RunOutput_Show(editor, showParsedOutputList);
            }

        } else if (runIn == "new-console") {
            try {
                _gRunSvc.Run(icommand, icwd, ienv, true, input);
            } catch (ex) {
                errmsg = lastErrorSvc.getLastErrorMessage();
                if (! errmsg) {
                    errmsg = "Unknown error running: '"+icommandForDisplay+"'";
                }
                runLog.error(errmsg);
                alert(errmsg);
                return false;
            }

        } else if (runIn == "no-console") {
            try {
                _gRunSvc.Run(icommand, icwd, ienv, false, input);
            } catch (ex) {
                errmsg = lastErrorSvc.getLastErrorMessage();
                if (! errmsg) {
                    errmsg = "Unknown error running: '"+icommandForDisplay+"'";
                }
                runLog.error(errmsg);
                alert(errmsg);
                return false;
            }
        } else {
            log.error("Unexpected 'runIn' value: " + runIn);
            throw("Unexpected 'runIn' value: " + runIn + "\n");
        }
    }
    var encodedCommand = _gRunSvc.Encode(command, cwd, env, insertOutput,
                                         operateOnSelection,
                                         doNotOpenOutputWindow, runIn,
                                         parseOutput, parseRegex,
                                         showParsedOutputList);
    if (saveInMRU) {
        _Run_UpdateMRU(command, encodedCommand, cwd, parseRegex);
    }
    if (saveInMacro &&
        typeof(editor.gMacroMgr) != 'undefined' &&
        editor.gMacroMgr &&
        editor.gMacroMgr.mode == 'recording') {
        // escape char that will get interpreter twice (XXX not confident
        // that this covers all cases)
        encodedCommand = encodedCommand.replace('\\', '\\\\', "g");
        encodedCommand = encodedCommand.replace("'", "\\'", "g");
        encodedCommand = encodedCommand.replace('"', '\\"', "g");
        var runtxt = "Run_RunEncodedCommand(window, '"
                      + encodedCommand + "');";
        editor.gMacroMgr.appendCode(runtxt);
    }

    // Put focus back on previous view. SciMoz's undo action stuff requires
    // this and even if not inserting output, it is nice to have focus back in
    // the file being editted.
    if (view) {
        view.focus();
    }

    // Insert the returned output if requested and select it.
    if (insertOutput && view && output) {
        function min(a, b) { return a <= b ? a : b; }
        function max(a, b) { return a >= b ? a : b; }
        scimoz.targetStart = min(scimoz.anchor, scimoz.currentPos);
        scimoz.targetEnd = max(scimoz.anchor, scimoz.currentPos);
        scimoz.replaceTarget(output.length, output);
        scimoz.currentPos = scimoz.anchor + output.length;
    }

    // Raise an alert dialog if there was error output.
    if (error) {
        dialog_alert("The command returned the following error output:",
                     error);
    }

    // If the command looks like it was operating on the current file, then
    // notify that the file might have changed.
    if (url) {
        var markers = {"%f":1, "%F":1, "%(f":1, "%(F":1};
        var _gObserverSvc = Components.classes["@mozilla.org/observer-service;1"]
                            .getService(Components.interfaces.nsIObserverService);
        for (marker in markers) {
            if (command.indexOf(marker) != -1) {
                _gObserverSvc.notifyObservers(self, "file_status", url)
                break;
            }
        }
    }

    return true;
} catch (e) {
    log.exception(e);
}
    return false;
}



// Prepare to close Komodo. Return false if cannot yet close Komodo.
// If there are processes still running then must wait for or kill them.
// Note: 'gRunProcesses' is defined in komodo.js, therefore this method must
//       only be called from the komodo.xul window.
function Run_CanClose()
{
    var i;
    // If there are any still running processes, then ask the user if they want
    // to kill those processes (call kill on them and return true) or cancel
    // and wait for those processes to finish normally (return false).
    _Run_UnregisterZombieProcess(window);
    if (gRunProcesses.length > 0) {
        var commands = "";
        for (i=0; i < gRunProcesses.length; i++) {
            commands += gRunProcesses[i]['command'] + '\n';
        }
        var question = "The following commands are still running "+
                       "and must terminate before Komodo can "+
                       "exit. Would you like Komodo to terminate "+
                       "these commands?";
        var answer = dialog_okCancel(question, "OK", commands);
        if (answer == "OK") {
            try {
                for (i=0; i < gRunProcesses.length; i++) {
                    //dump("Run_CanClose(): killing '"+
                    //     gRunProcesses[i]['command']+"'\n");
                    gRunProcesses[i]['process'].kill(-1);
                }
            } catch (ex) {
                gRunLog.error("Error killing running process when closing Komodo:"+
                         ex);
                //XXX Should this be reported to the user?
                return false;
            }
        } else if (answer == "Cancel") {
            return false;
        } else {
            throw("Unexpected return value from OK/Cancel dialog in "+
                  "Run_CanClose: '"+answer+"'.\n");
        }
    }
    return true;
}

