# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

# An implementation of a ftp file objects
#
# Contributors:
# * Shane Caraveo
# * Todd Whiteman

from xpcom import file, components, COMException, ServerException, server, nsError
from xpcom.server.enumerator import SimpleEnumerator

# import tempfile
# import codecs
import logging
import threading

# needed for ftpfile
import re, os, stat
import ftplib, StringIO, string

import URIlib, remotefilelib


class koFTPConnection(remotefilelib.koRFConnection):
    _com_interfaces_ = [components.interfaces.koIFTPConnection]
    _reg_desc_ = "Komodo FTP Connection"
    _reg_contractid_ = "@activestate.com/koFTPConnection;1"
    _reg_clsid_ = "{76A0956E-044C-45fc-BD9B-07BC5C659E4A}"

    def __init__(self):
        remotefilelib.koRFConnection.__init__(self)
        self.log = logging.getLogger('koFTPConnection')
        #self.log.setLevel(logging.DEBUG)
        self.port = remotefilelib.koRFProtocolDefaultPort['ftp']
        self.protocol = 'ftp'
        #self._globalPrefs = \
        #    components.classes["@activestate.com/koPrefService;1"].\
        #    getService(components.interfaces.koIPrefService).prefs

    def __del__(self):
        try:
            self.log.debug("koFTP deleted")
            if self._connection:
                self._connection.quit()
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP CLOSE ERROR: %s", self._lasterror)
            pass
        del self.log

    def _setConnectionMode(self):
        # Most servers accept passive mode, so we'll try that first
        # Note: passive is enabled by default in Python 2.1 and above
        info = []
        try:
            self._connection.makepasv()
        except ftplib.all_errors, e:
            self.log.info("_setConnectionMode: Unable to set passive mode, exception: %s", e)
            error_code = e[-1][:3]
            if error_code == "550":
                self.log.info("_setConnectionMode: Trying active mode")
                self._connection.set_pasv(False)

    def do_openSocket(self):
        """Open the FTP connection socket to the remote site."""
        self._lasterror = ""
        # open the connection
        try:
            self._connection = ftplib.FTP()
            self._connection.connect(self.server, self.port)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP OPEN ERROR: %s", self._lasterror)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def do_authenticateWithAgent(self):
        """Not used for ftp"""
        return 0

    def do_authenticateWithPassword(self):
        # attempt authentication
        try:
            self._connection.login(self.username, self.password)
            self._setConnectionMode()
            self.log.debug("FTP login: Successful")
            return 1
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            if self._lasterror[:3] == "530":
                # username/password error, unset password so we can prompt again
                self.log.debug("FTP login: Invalid username/password")
                self.password = ''
            else:
                raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
        return 0

    def NOOP(self):
        return self._connection.voidcmd('NOOP')

    def verifyConnected(self):
        """Verify that we are connected to the remote server"""
        try:
            if not self._connection:
                self.open(self.server, self.port, self.username, self.password)
            else:
                self.NOOP()
        except ftplib.all_errors, e:
            # connection lost, reconnect if possible
            self.log.info("Re-opening the ftp connection because it was closed")
            self.open(self.server, self.port, self.username, self.password)

    def do_close(self):
        """Close the FTP connection"""
        self._lasterror = ""
        try:
            self.log.debug("Closing FTP connection")
            if self._connection:
                self._connection.quit()
                self._connection = None
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP CLOSE ERROR: %s", self._lasterror)
            # XXX - Raise an exception... ??
            pass

    basicModeForBaseDir = stat.S_IFDIR | stat.S_IREAD | stat.S_IWRITE
    def _createBaseDirectory(self, dirname):
        rf = components.classes['@activestate.com/koRemoteFileInfo;1'].\
                    createInstance(components.interfaces.koIRemoteFileInfo)
        rf.initFromStats(dirname, dirname, 0, 0, 0, self.basicModeForBaseDir, 0)
        return rf

    def _getRFElementForPath(self, path):
        pwd = None
        try:
            info = []
            dirname = os.path.dirname(path)
            basename = os.path.basename(path)
            self.log.debug("_getRFElementForPath for: '%s'", path)
            # We have two ways of doing this
            # 1 - Using LIST -lad, which just displays the directory
            # 1 - Note: Some servers may not use this option (I.e. windows)
            # 1 - Note: Also, some quoted paths may not work with this
            try:
                self._connection.retrlines("LIST -lad %s" % path, info.append)
                if len(info) == 1:
                    rf_pathinfo = self._createRFInfoFromListing(dirname, info[0])
                    if rf_pathinfo and rf_pathinfo.getFilename() == basename:
                        return rf_pathinfo
            except ftplib.all_errors, e:
                self.log.debug("LIST -lad failed: %s", e)
                pass
    
            # 2 - Do a LIST -al of the parent directory and find the child in the list
            pwd = self.currentDirectory()
            self._connection.cwd(dirname)
            info = []
            self._connection.retrlines("LIST -la", info.append)
            for filelisting in info:
                if filelisting.find(dirname):
                    rf_pathinfo = self._createRFInfoFromListing(dirname, filelisting)
                    if rf_pathinfo and rf_pathinfo.getFilename() == basename:
                        return rf_pathinfo
            return None
        finally:
            if pwd:
                self._connection.cwd(pwd)

    def _createRFInfoFromListing(self, dirname, fileinfo):
        rf = components.classes['@activestate.com/koRemoteFileInfo;1'].\
                    createInstance(components.interfaces.koIRemoteFileInfo)
        if not rf.initFromDirectoryListing(dirname, fileinfo):
            del rf
            return None
        if rf.isSymlink():
            # Find out the link target and populate the target info
            # into this file
            link_target = rf.getLinkTarget()
            self.log.debug("Symlink info for '%s' -> '%s'", rf.getFilepath(), link_target)
            ## We now know where the link goes, but not what the link is
            ## find out what it is
            info = []
            rf_link = self._getRFElementForPath(link_target)
            rf.initFromStats(rf.getDirname(), rf.getFilename(), rf_link.size,
                             rf_link.uid, rf_link.gid, rf_link.mode, 
                             rf_link.mtime)
            self.log.debug("File now: %s", rf)
        return rf

    def do_getPathInfo(self, path):
        try:
            # Get the remote file details
            rf_info = self._getRFElementForPath(path)
            if rf_info:
                return rf_info

            # Last chance, might be a basedir
            if path in ('/', '~', '~/'):
                return self._createBaseDirectory(path)

            self._lasterror = "FTP LIST ERROR: Unable to list path: '%s'" % (path)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.warn("FTP LIST ERROR: %s", self._lasterror)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def do_getDirectoryList(self, path):
        dirinfo = []
        pwd = None
        try:
            try:
                info = []
                self._connection.cwd(path)
                self._connection.dir("-la", info.append)
                for fileinfo in info:
                    rf_fileinfo = self._createRFInfoFromListing(path, fileinfo)
                    if rf_fileinfo and rf_fileinfo.getFilename() not in (".", ".."):
                        dirinfo.append(rf_fileinfo)
            except ftplib.all_errors, e:
                self._lasterror = e.args[-1]
                self.log.warn("FTP LIST ERROR: %s", self._lasterror)
                ## don't raise, this is expected behaviour if creating a directory or file
                #pass
                raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
        finally:
            if pwd:
                self._connection.cwd(pwd)
        return dirinfo

    def do_rename(self, oldName, newName):
        try:
            self._connection.rename(oldName, newName)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP RENAME ERROR: %s", self._lasterror)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def do_removeFile(self, name):
        try:
            self._connection.delete(name)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP REMOVE ERROR: %s", self._lasterror)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def do_changeDirectory(self, path):
        try:
            self._connection.cwd(path)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP CWD ERROR: %s", self._lasterror)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def do_currentDirectory(self):
        try:
            path = self._connection.pwd()
            self.log.debug("currentDirectory cwd %s",path)
            return path
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP PWD ERROR: %s", self._lasterror)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def do_getHomeDirectory(self):
        try:
            pwd = self._connection.pwd()
            self._connection.cwd("~")
            home = self._connection.pwd()
            self._connection.cwd(pwd)
            self.log.debug("getHomeDirectory cwd %s",home)
            return home
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.warn("FTP HOMEDIR ERROR: %s", self._lasterror)
        return "/"
            
    def do_removeDirectory(self, name):
        try:
            self._connection.rmd(name)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP REMOVE ERROR: %s", self._lasterror)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def do_createDirectory(self, name, permissions):
        try:
            self._connection.mkd(name) 
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP CREATE ERROR: %s", self._lasterror)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)


    def _read_cb(self, block):
        #print "read %r bytes"  % (len(block))
        self.filedata.write(block)

    def do_readFile(self, filename):
        try:
            # We want binary
            self._connection.sendcmd('TYPE I')
            if self.filedata.tell() == 0:
                # retreive the file
                self._connection.retrbinary("RETR "+filename, self._read_cb)
                self.filedata.seek(0,0)
            data = self.filedata.getvalue()
            # reset filedata
            self.filedata.close()
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP RETR ERROR: %s", self._lasterror)
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
        self.filedata = StringIO.StringIO()
        return data

    def do_writeFile(self, filename, data):
        self.filedata.write(data)
        self.filedata.seek(0,0)
        self._lasterror = ""
        try:
            self._connection.storbinary("STOR "+filename, self.filedata, 8192)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            self.log.error("FTP STOR ERROR: %s", self._lasterror)
            self.filedata.close()
            self.filedata = StringIO.StringIO()
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
        self.filedata.close()
        self.filedata = StringIO.StringIO()



class koFTPNode:
    _com_interfaces_ = [components.interfaces.koIFTPNode, components.interfaces.koIFTPNodeFile]
    _reg_desc_ = "Komodo FTP Node"
    _reg_contractid_ = "@activestate.com/koFTPNode;1"
    _reg_clsid_ = "{2747A621-9F3B-400d-B724-B6CA0F5F3027}"

    def __init__(self):
        self.log = logging.getLogger('koFTPNode')
        #self.log.setLevel(logging.DEBUG)
        self.connection = None
        self._parent = None
        self._children = []
        self._path = None
        self._readable = 0
        self._writeable = 0
        self._isDirectory = 0
        self._isFile = 0
        self._isHidden = 0
        self._isSymlink = 0
        self.fileSize = 0
        self.lastModifiedTime = 0
        self.leafName = ""
        self.protocol = ""
        self.server = ""
        self.port = 21
        self.username = ""
        self.password = ""
        self.followlinks = 1
        self.URL = ""
        self.diskSpaceAvailable = 0
        self.persistentDescriptor = ""
        self._target = ""
        
        # no support for the following attributes yet
        self.NORMAL_FILE_TYPE = 0
        self.DIRECTORY_TYPE = 1
        self.permissions = 0
        self.permissionsOfLink = 0
        self.lastModifiedTimeOfLink = 0
        self.fileSizeOfLink = 0

        self._globalPrefs = \
            components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs

    def init(self, protocol, server, port, username, password):
        self.log.debug("Init")
        if not self.connection:
            self.log.debug("attempt new connection %s %r %s",server,port,username)
            self.connection = remotefilelib.RFService.getConnection(protocol, server, port, username, password)
            self.protocol = protocol

    def normalize(self):
        self._path = remotefilelib.normalizePath(self._path)

    def getChildren(self):
        return self._children
    
    def get_parent(self):
        self._getParent()
        return self._parent
    
    def _getParent(self):
        if not self._parent and self._path != "/":
            self._parent = components.classes['@activestate.com/koFTPNode;1'].\
                        createInstance(components.interfaces.koIFTPNode)
            # get the parent path
            parts = string.split(self._path, "/")
            parent_path = ""
            if len(parts) > 2:
                parts = parts[:len(parts)-1]
                parent_path = string.join(parts, "/") + "/"
            if not parent_path:
                parent_path = "/"
            self.log.debug("getting parent for %s leafname: %s parent: %s",self._path,self.leafName,parent_path)
            self._parent.initWithConnection(self.connection, parent_path)
    
    def initClone(self, parent, path):
        self._parent = parent
        if parent: self._children = parent.getChildren()
        self.initWithConnection(parent.connection, path)
            
    def initWithConnection(self, connection, path):
        self.connection = connection
        self.protocol = self.connection.protocol
        self.set_path(path)

    def initChild(self, parent, info):
        self._parent = parent
        self.initWithConnectionInfo(parent.connection, info)

    def initWithConnectionInfo(self, connection, info):
        self.connection = connection
        self.protocol = self.connection.protocol
        self._loadRFInfo(info)
        #if self._isSymlink and self.followlinks:
        #    # automaticly resolve symlinks
        #    self.log.debug("initWithConnectionInfo attempting to resolve link: %s",self._target)
        #    leafname = self.leafName
        #    # This will grab the information about the link. We need to set back
        #    # the path for link and the leafName, as set_path changes it
        #    #self.set_path(remotefilelib.addslash(self._parent.path) + self._target)
        #    self.set_path(remotefilelib.addslash(self._parent.path) + self._target)
        #    if not self.leafName:
        #        if leafname[0]=="/": leafname = leafname[1:]
        #    self.leafName = leafname
        #    #self.log.debug("initWithConnectionInfo path : %s leafname : %s " ,self._path,self.leafName)
        ##else:
        self._path = remotefilelib.addslash(self._parent.path) + self.leafName
        self._generateURI()
    
    def set_URL(self,url):
        # XXX the path should be reset if URL is set
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    
    def _generateURI(self):
        # we do not include username/password in the URI!
        # those must be retrieved from prefs
        if self._path[0] == '~':
            uri = self.protocol+"://"+self.connection.server+":"+str(self.connection.port)+'/'+self._path
        else:
            uri = self.protocol+"://"+self.connection.server+":"+str(self.connection.port)+self._path
        self.URL = uri

    def _loadRFInfo(self, stats):
        # XXX - Does not parse up raw directory list anymore, but now uses
        #       RFInfo class.
        # XXX - koFTPNode() can be cleaner if these fields are removed and the
        #       RFInfo object is used instead
        self.log.info("D:%r  L:%r  R:%r  W:%r  S:%d  Name:%s", stats.isDirectory(), stats.isSymlink(), stats.isReadable(), stats.isWriteable(), stats.getFileSize(), stats.getFilename())
        self._stats = stats
        self._readable = stats.isReadable()
        self._writeable = stats.isWriteable()
        self._isDirectory = stats.isDirectory()
        self._isFile = stats.isFile()
        self._isHidden = stats.isHidden()
        self._isSymlink = stats.isSymlink()
        self._symlink_target = stats.getLinkTarget()
        self.fileSize = stats.getFileSize()
        self.leafName = stats.getFilename()
        self.lastModifiedTime = stats.getModifiedTime()

    def get_path(self):
        self.log.debug('get_path %s',self._path)
        return self._path
    
    def set_path(self, path):
        self.log.debug("set_path %s",path)
        # parse path for specific info
        if not path or path == "": path = "/"
        
        self._path = path
        self.normalize()
        if not self.connection:
            self.log.warn("No connection for set_path")
            return

        #if self._isSymlink and self.followlinks:
        #    self.log.info("Listing linked directory: %s -> %s", self.leafName, self._realpath)
        #else:
        # This can raise an exception when the path does not exist
        try:
            info = self.connection.list(self._path, 0)
            self._loadRFInfo(info)
        except COMException, ex:
            pass

        #num_resp = len(info)
        ## if num_resp == 0, file does not exist, we may be attempting
        ## to write a new file, so we will set defaults simular to directory
        #if num_resp <> 1:
        #    self.log.debug("set_path: directory found")
        #    #if file wasn't found, lastError will contain an error, otherwise
        #    # this is likely a directory
        #    if self.connection.lastError == "":
        #        # this is a directory
        #        # XXX, how do I get it's info?
        #        self._isDirectory = 1
        #        self._readable = 1
        #        self.lastModifiedTime = time.time()
        #        self.fileSize = 1
        #    else:
        #        self.log.error("ERROR: %s",self.connection.lastError)
        #        self._isDirectory = 0
        #        self._readable = 0
        #        self.lastModifiedTime = 0
        #        self.fileSize = 0
        #            
        #    self._writeable = 1
        #    self._isFile = 0
        #    self._isHidden = 0
        #    self._isSymlink = 0
        #    self.leafName = ""
        #        
        #elif num_resp==1:
        #    self.log.debug("set_path: file found")
        #    line = info[0]
        #    self._loadRFInfo(line)
        #    # we still may be looking at a directory with one entry, figure it out
        #    if not self._isDirectory and \
        #       os.path.basename(self._path) <> self.leafName:
        #        # this is a directory with a single entry, revert to a directory
        #        self._isDirectory = 1
        #        self.log.info("Setting to directory. '%s' <> '%s'", self._path, self.leafName)
        #        self._readable = 1
        #        self.lastModifiedTime = time.time()
        #        self.fileSize = 1
        #        self._writeable = 1
        #        self._isFile = 0
        #        self._isHidden = 0
        #        self._isSymlink = 0
        #        self.leafName = ""
        #            
        #    if self._isSymlink and self.followlinks:
        #        # automaticly resolve symlinks
        #        self.log.info("attempting to resolve link: %s",self._target)
        #        #self._path = remotefilelib.addslash(self._path) + self.leafName
        #        # self._path = remotefilelib.addslash(os.path.dirname(self._path)) + self.leafName
        #        # self._linked_target = remotefilelib.addslash(os.path.dirname(self._path)) + self._target
        #        # self._path = remotefilelib.addslash(os.path.dirname(self._path)) + self._target
        #        # self.normalize()
        #    self.log.debug("path : %s leafname : %s ",repr(self._path),self.leafName)

        self._generateURI()
        self.log.debug('URL:  %s',self.URL)
        self.log.debug('path: %s',self._path)
    
    def refresh(self):
        if self.connection:
            path = self._path
            self.connection.refresh(path)
            self.set_path(path)

    def _generateChildren(self, path):
        dirinfo = self.connection.list(self._path, 0)
        nodes = []
        # we have more listings lets make some nodes
        for info in dirinfo.getChildren():
            node = components.classes['@activestate.com/koFTPNode;1'].\
                        createInstance(components.interfaces.koIFTPNode)
            node.initChild(self, info)
            nodes.append(node)
        return nodes

    def get_directoryEntries(self):
        if not self.connection:
            return SimpleEnumerator([])
        if len(self._children) < 1:
            # or not self.connection.haveCacheEntry(self._path):
            self._children = self._generateChildren(self._path)
        return SimpleEnumerator(self._children)
    
    def exists(self):
        if not self.connection: return 0
        try:
            # This can raise an exception when the path does not exist
            exists = self.connection.havePath(self._path)
        except COMException, ex:
            exists = 0
        self.log.debug('exists %r',exists)
        return exists
    
    def close(self):
        self.log.debug('closing the connection')
        self.connection.close()
        self.connection = None
    
    def equals(self, node):
        if self._path == node.path and self.fileSize == node.fileSize \
            and self.lastModifiedTime == node.lastModifiedTime:
                return 1
        return 0
    
    def clone(self):
        node = components.classes['@activestate.com/koFTPNode;1'].\
                    createInstance(components.interfaces.koIFTPNode)
        parent = self.get_parent()
        if parent:
            node.initClone(parent, self._path)
        else:
            node.initWithConnection(self.connection, self._path)
        return node
    
    def append(self, relativePath):
        if not relativePath or relativePath[0] == "/":
            # see nsLocalFileWin.cpp for related functionality
            raise ServerException(nsError.NS_ERROR_FILE_UNRECOGNIZED_PATH, relativePath)
        path = self._path
        if path == "/":
            path += relativePath
        else:
            path += "/"+relativePath
        self.log.debug("new path is %s",path)
        self.set_path(path)
    
    def dump(self):
        print "node: %s leafname: %s isdir: %r uri: %s" % \
                (self._path, self.leafName, self._isDirectory, self.uri)

    #############################################################
    # koIFTPNodeFile interface functions
    #############################################################
    def initWithPath(self, filePath):
        self.log.debug("initWithPath %s",filePath)
        self.set_path(filePath)
    
    #[noscript] PRFileDescStar openNSPRFileDesc(in long flags, in long mode);
    def openNSPRFileDesc(self, flags, mode):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
            
    #[noscript] FILE           openANSIFileDesc(in string mode);
    def openANSIFileDesc(self, mode):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
            
    #[noscript] PRLibraryStar  load();
    def load(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    
    def appendRelativePath(self, relativeFilePath):
        self.append(relativeFilePath)
    
    def reveal(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    
    def launch(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    
    #############################################################
    # realy basic stuff
    #############################################################
    
    def get_target(self):
        if self._isSymlink:
            return self._target
        else:
            raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def isWritable(self):
        self.log.debug('isWritable %r',self._writeable)
        return self._writeable
    
    def isReadable(self):
        self.log.debug('isReadable %r',self._readable)
        return self._readable
    
    def isExecutable(self):
        return 0
    
    def isHidden(self):
        self.log.debug('isHidden %r',self._isHidden)
        return self._isHidden
    
    def isDirectory(self):
        self.log.debug('isDirectory %r',self._isDirectory)
        return self._isDirectory

    def isFile(self):
        self.log.debug('isFile %r',self._isFile)
        return self._isFile

    def isSymlink(self):
        self.log.debug('isSymlink %r',self._isSymlink)
        return self._isSymlink

    def isSpecial(self):
        return 0

    def moveTo(self, newParentDir, newName):
        # moveTo only does a rename in ftp
        parent = self.get_parent()
        if newParentDir is not None:
            newname = remotefilelib.addslash(newParentDir)+newName
        elif parent:
            newname = remotefilelib.addslash(parent.path)+newName
        else:
            path = re.search(r'^(.*)/', self._path)
            newname = path + "/"+newName

        if string.find(newname, "..") > -1:
            newname = remotefilelib.normalizePath(newName)
        
        self.connection.rename(self._path, newname)
        if parent:
            self.connection.refresh(parent.path)

    def remove(self, recursive):
        if self._isDirectory:
            self.connection.removeDirectory(self._path)
        else:
            self.connection.removeFile(self._path)
        parent = self.get_parent()
        if parent:
            self.connection.refresh(parent.path)

    def createUnique(self, type, permissions):
        # since we append the new directory name before calling createUnique,
        # self._isDirectory will always be false
        #if not self._isDirectory:
        #        raise COMException(nsError.NS_ERROR_NOT_IMPLEMENTED)
        if type == self.DIRECTORY_TYPE:
            self.connection.createDirectory(self._path, permissions)
            parent = self.get_parent()
            if parent:
                self.connection.refresh(parent.path)
        else:
            raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
            
    #############################################################
    # NOT IMPLEMENTED!
    #############################################################
    
    def create(self, type, permissions):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def copyTo(self, newParentDir, newName):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
            
    def copyToFollowingLinks(self, newParentDir, newName):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def spawn(self, args):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def contains(self, inFile,  recursive):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)


class koFTPNodeFile(koFTPNode):
    _com_interfaces_ = [components.interfaces.koIFTPNode, components.interfaces.koIFTPNodeFile, components.interfaces.nsILocalFile]
    _reg_desc_ = "Komodo FTP Node File"
    _reg_contractid_ = "@activestate.com/koFTPNodeFile;1"
    _reg_clsid_ = "{CA20E22B-99D9-4e40-B156-521017549B95}"
    
    def __init__(self):
        koFTPNode.__init__(self)

    #  def initWithPath(self, filePath):
    #          koFTPNode.initWithPath(self, filePath)
    #  
    #  #[noscript] PRFileDescStar openNSPRFileDesc(in long flags, in long mode);
    #  def openNSPRFileDesc(self, flags, mode):
    #          return koFTPNode.openNSPRFileDesc(self, flags, mode)
    #          
    #  #[noscript] FILE           openANSIFileDesc(in string mode);
    #  def openANSIFileDesc(self, mode):
    #          return koFTPNode.openANSIFileDesc(self, mode)
    #          
    #  #[noscript] PRLibraryStar  load();
    #  def load(self):
    #          return koFTPNode.load()
    #
    #  def appendRelativePath(self, relativeFilePath):
    #          koFTPNode.appendRelativePath(self, relativeFilePath)
    #  
    #  def reveal(self):
    #          koFTPNode.reveal(self)
    #  
    #  def launch(self):
    #          koFTPNode.launch(self)


# FTP File support
class FTPFile:
    _com_interfaces_ = [components.interfaces.koIFTPFile]
    _reg_desc_ = "Komodo FTP File"
    _reg_contractid_ = "@activestate.com/koFTPFile;1"
    _reg_clsid_ = "{010656B6-E830-45ae-B7D9-F22C1E4AFDEC}"

    def __init__(self):
        self.log = logging.getLogger('FTPFile')
        #self.log.setLevel(logging.INFO)
        self.protocol = None
        self.password = None
        self.username = None
        self._globalPrefs = None
        self.filename = ""
        self.filedata = StringIO.StringIO()
        self.ftpfile = None
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                            .getService(components.interfaces.koILastErrorService)

    def initServer(self, url, mode="r", server_alias=""):
        if not url:
            self.log.warn("Encountered undecipherable encoding %s" % (url))
        self.log.debug("initServer url='%s' mode='%s' server_alias='%s'" % (url, mode, server_alias))
        self.server_opts = None
        if server_alias:
            # get username and password from prefs
            if not self._globalPrefs:
                self._globalPrefs = \
                    components.classes["@activestate.com/koPrefService;1"].\
                    getService(components.interfaces.koIPrefService).prefs
            deploymentServerList = self._globalPrefs.getPref("deploymentServerList");

            for i in range(0, deploymentServerList.length):
                dserver = deploymentServerList.getPref(i).QueryInterface(components.interfaces.koIOrderedPreference)
                if server_alias == dserver.getStringPref(0):
                    # we found our match
                    so = dserver.getStringPref(1)
                    self.server_opts = so.split(":")
                    #print "server_opts=",so, repr(self.server_opts)
                    proto_name = self.server_opts[0].lower()
                    if proto_name not in remotefilelib.koHandledRFProtocols:
                            raise ValueError, "Unhandled Protocol: %s" % (proto_name)
                    self.protocol = proto_name
                    self.username = self.server_opts[4]
                    self.password = self.server_opts[5]
                    crypto = components.classes['@activestate.com/koPasswordCrypto;1'].\
                            getService(components.interfaces.koIPasswordCrypto)
                    # if the password is empty, that's ok, we'll ask for it later
                    if self.password:
                        try:
                            self.password = crypto.decodePassword(self.password)
                        except:
                            self.log.warn("Dropping password that could not be decoded. See bug 18752.")
                    #print "username: %s password %s" % (self.username, self.password)
                    break
        self.init(url, mode)
            
    def init(self, url, mode="r"):
        # Note: urlparse does not parse up sftp url's properly, so using our own URIlib
        try:
            uri = URIlib.URIParser(url)
            if uri.scheme not in remotefilelib.koHandledRFProtocols:
                raise ValueError, "Unknown Protocol :%s" % (uri.scheme)
            
            # get the username:password@domain
            #u = re.search("^(([^:@]+)(:([^:@]+))?@)?([^:@]+)(:(.*))?",uri.server).groups()
            self.log.debug("host info: %s" % (uri.server))
            ug = re.search("^(([^:@]+)(:([^:]+))?@)?([^:@]+)(:(.*))?",uri.server)
            if not ug:
                raise ValueError, "Regular Expression Failure in %s" % (uri.server)
                return
            u = ug.groups()
            #print "re: ",repr(u)
            if not self.username: self.username = u[1]
            if not self.password: self.password = u[3]
            self.domain = u[4]
            self.port = u[6] # not used yet
            if not self.port or self.port == "0": self.port = remotefilelib.koRFProtocolDefaultPort[uri.scheme]
            #print "%s:%s@%s" %(username,password,domain)
    
            self.fullpath = uri.path
            self.mode = mode
            # make the connection to the ftp site
            self.ftpfile = components.classes["@activestate.com/koFTPNodeFile;1"] \
                            .createInstance(components.interfaces.koIFTPNodeFile);
            self.ftpfile.init(self.protocol, self.domain, string.atoi(self.port), self.username, self.password)
            self.ftpfile.initWithPath(self.fullpath)
        except Exception, e:
            self.log.info("*" * 40)
            self.log.exception(e)
            self.log.info("*" * 40)

    def read(self, n = -1):
        self.log.debug("attempting to read ftp file, size %d" % (n))
        if not self.ftpfile.connection or not self.ftpfile.isReadable():
            self.log.warn("unable to read file %s, readable: %d",self.fullpath,self.ftpfile.isReadable())
            return None
        if not self.filedata:
            self.filedata = StringIO.StringIO()
        if self.filedata.tell() == 0:
            # retrieve the file
            data = self.ftpfile.connection.readFile(self.fullpath)
            #print "read in ",len(data)," bytes"
            self.filedata.write(data)
            self.filedata.seek(0,0)
        if n == -1:
            return self.filedata.getvalue()
        else:
            return self.filedata.read(n)

    def write(self, data):
        if not self.ftpfile.connection or not self.ftpfile.isWritable():
            self.log.warn("unable to write file %s, writeable %d",self.fullpath,self.ftpfile.isWritable())
            return None
        if not self.filedata:
            self.filedata = StringIO.StringIO()
        self.filedata.write(data)

    def _store(self):
        if self.mode.find('w') >= 0:
            self.filedata.seek(0,0)
            self.ftpfile.connection.writeFile(self.fullpath, self.filedata.getvalue())
        self.filedata.close()
        self.filedata = None

    def close(self):
        try:
            if self.ftpfile.connection and self.filedata:
                self._store()
        except COMException, ex:
            err = self.ftpfile.connection.lastError
            self.lastErrorSvc.setLastError(0, err)
            raise ServerException(nsError.NS_ERROR_FAILURE, err)

    def get_isWritable(self):
        if self.ftpfile:
            return self.ftpfile.isWritable()
        self.log.warn("no ftpfile %s available",self.fullpath)
        return 0
    
    def get_isReadable(self):
        if self.ftpfile:
            return self.ftpfile.isReadable()
        self.log.warn("no ftpfile %s available",self.fullpath)
        return 0



def _test():
    # Let's make sure our URIlib parses up our required protocols correctly.
    #import urlparse
    ftp_url = r'ftp://test:testuser@toddw:21/dir1/dir1.txt'
    uri_ftp = URIlib.URIParser(ftp_url)
    sftp_url = r'sftp://test:testuser@toddw:22/dir1/dir1.txt'
    uri_sftp = URIlib.URIParser(sftp_url)
    assert uri_ftp.scheme == 'ftp'
    assert uri_sftp.scheme == 'sftp'
    assert uri_ftp.leafName == uri_sftp.leafName == 'dir1.txt'
    assert uri_ftp.dirName == uri_sftp.dirName == '/dir1'
    assert uri_ftp.server == 'test:testuser@toddw:21'
    assert uri_sftp.server == 'test:testuser@toddw:22'
    #parts = list(urlparse.urlparse(url))
    #print parts
    #for uri in (uri_ftp, uri_sftp):
    #    ftpfile = components.classes["@activestate.com/koFTPNodeFile;1"] \
    #                    .createInstance(components.interfaces.koIFTPNodeFile);
    #    sp = uri.server.split("@", 1)
    #    username, password = sp[0].split(':', 1)
    #    hostname, port = sp[1].split(':', 1)
    #    ftpfile.init(uri.scheme, hostname, string.atoi(port), username, password)
    #    ftpfile.initWithPath('~' + uri.path)
    


if __name__ == '__main__':
    _test()
