
import sys, os
import fnmatch
from types import *
import timeline

from xpcom import components, nsError, ServerException, COMException
import xpcom.server
from xpcom.server import WrapObject, UnwrapObject
from xpcom.server.enumerator import SimpleEnumerator
import re
import logging
log = logging.getLogger('koLanguage')
#log.setLevel(logging.INFO)

class koLanguageItem:
    _com_interfaces_ = [components.interfaces.koIHierarchyItem]
    _reg_contractid_ = "@activestate.com/koLanguageItem;1"
    _reg_clsid_ = "{33532bc1-302a-4b2d-ad14-13c5eadf4d93}"
    
    def __init__(self, language, key):
        self.language = language
        self.name = language
        self.key = key
        
    def get_available_types(self):
        return components.interfaces.koIHierarchyItem.ITEM_STRING

    def get_item_string(self):
        return self.language

    def get_container(self):
        return 0

class koLanguageContainer:
    _com_interfaces_ = [components.interfaces.koIHierarchyItem]
    _reg_contractid_ = "@activestate.com/koLanguageContainer;1"
    _reg_clsid_ = "{878ed885-6274-4c07-9668-a9a01a0ae09c}"
    
    def __init__(self, label, languages):
        self.name = label
        self.languages = languages
    
    def getChildren(self):
        return self.languages

    def get_available_types(self):
        return 0
        return components.interfaces.koIHierarchyItem.ITEM_STRING

    def get_item_string(self):
        return 0
        return self.label

    def get_container(self):
        return 1
    
# The LanguageRegistryService keeps track of which languages/services
# are available.  It is used by packages of services to inform the
# system of their presence, and is used internally by the other
# language services classes to get information on the services
# available.

class koLanguageRegistryService:
    _com_interfaces_ = [components.interfaces.koILanguageRegistryService,
                        components.interfaces.nsIObserver]
    _reg_contractid_ = "@activestate.com/koLanguageRegistryService;1"
    _reg_clsid_ = "{4E76795E-CC92-47c6-8801-C9ACFC1B02E3}"

    primary = []
    internal = [] # Should 'Rx' be 'Regex' now?
    
    # Language-specific patterns that will identify the langauge based on
    # a match against the head of the document.
    #
    # Note that some shebang "line" matches require more than the first
    # line so pattern should be prepared to deal with that.
    # XXX this belongs in language services
    shebangPatterns = []
    
    # Mapping of local variable mode names (typically Emacs mode name)
    # to the appropriate language name in this registry.
    # - All we need to note here are exceptions in the naming scheme,
    #   like "mode: C" which corresponds to Komodo's C++ language.
    _modeName2LanguageName = {}
    
    def __init__(self):
        timeline.enter('koLanguageRegistryService.__init__')
        self.__languages = {}
        self.__accesskeys = {}
        self.__filters = None
        self.__filtersByLanguage = None
        self.__associations = None
        self.__absolutefilters = {}
        self.__initPrefs()
        self._resetFileAssociations()

        self._globalPrefSvc = components.classes["@activestate.com/koPrefService;1"].\
                            getService(components.interfaces.koIPrefService)
        self._globalPrefs = self._globalPrefSvc.prefs

        # XXX get from prefs?
        self.defaultLanguage = "Text"
        
        self.registerLanguages()
        timeline.leave('koLanguageRegistryService.__init__')
    
    def observe(self, aSubject, aTopic, someData):
        if someData == "fileAssociations":
            self.__filters = None
            self.__filtersByLanguage = None
            self.__associations = []
            self._resetFileAssociations()
            
    def _resetFileAssociations(self):
        timeline.enter('koLanguageRegistryService._resetFileAssociations')
        associations = self.__prefs.getPref("fileAssociations") \
                       .QueryInterface(components.interfaces.koIOrderedPreference)
        self.__filters = []
        self.__filtersByLanguage = {}
        self.__associations = []
        self.__absolutefilters = {}
        for i in range(0, associations.length):
            association = associations.getPref(i) \
                          .QueryInterface(components.interfaces.koIOrderedPreference)
            filterMatch = association.getStringPref(0)
            filterLanguage = association.getStringPref(1)

            # set filters
            self.__filters.append(filterMatch)

            # set filters by language
            if not self.__filtersByLanguage.has_key(filterLanguage):
                self.__filtersByLanguage[filterLanguage] = []
            self.__filtersByLanguage[filterLanguage].append(filterMatch)

            # get the file extensions only
            if filterMatch[:2] == "*.":
                self.__absolutefilters[filterMatch[2:]] = filterLanguage
            self.__absolutefilters[filterMatch] = filterLanguage
            
            # set associations
            self.__associations.append((filterMatch,filterLanguage))
        timeline.leave('koLanguageRegistryService._resetFileAssociations')

    def getLanguageHierarchy(self):
        # We could determine primary languages programmatically but seeing which
        # features each language offers, but that would involve creating an
        # XPCOM object for each language.

        secondary = [i for i in self.__languages.keys()
                        if i not in self.primary and
                           i not in self.internal]
        secondary.sort()

        otherContainer = koLanguageContainer('Other', [koLanguageItem(i, self.__accesskeys[i]) for i in secondary])
        return koLanguageContainer('', [koLanguageItem(i, self.__accesskeys[i]) for i in self.primary] + [otherContainer])
                
    def getLanguages(self):

        languages = self.__languages.keys()
        languages.sort()
        languages = [language for language in languages if language not in self.internal]
        return languages
    
    def getFilters(self):
        if self.__filters is None:
            self._resetFileAssociations()

        return self.__filters

    def getFiltersForLanguage(self, language):
        if self.__filtersByLanguage is None:
            self._resetFileAssociations()

        if self.__filtersByLanguage.has_key(language):
            filters = self.__filtersByLanguage[language]
        else:
            filters = []

        return filters

    def registerLanguages(self):
        """registerLanguages
        
        gets a list of files from the same directory this file
        is located in, and calls the register function.
        """
        import sys, os
        
        frame = sys._getframe()
        filename = frame.f_code.co_filename
        dirname = os.path.dirname(filename)
        # XXX eventually change this to look into a special
        # languages directory
        log.info("looking for languages in [%s]",dirname)
        import glob
        files = glob.glob(os.path.join(dirname, 'ko*Language.py'))
        for file in files:
            module = {}
            execfile(file, module)
            if 'registerLanguage' in module:
                module['registerLanguage'](self)
        
    def registerLanguage(self, language):
        name = language.name
        assert not self.__languages.has_key(name), \
               "Language '%s' already registered" % (name)
        log.info("registering language [%s]", name)
        
        self.__languages[name] = language
        language = UnwrapObject(language)
        self.__accesskeys[name] = language.accessKey
        # So that we can tell that, for example:
        #     -*- mode: javascript -*-
        # means language name "JavaScript".
        if language.modeNames:
            for modeName in language.modeNames:
                self._modeName2LanguageName[modeName.lower()] = name
        else:
            self._modeName2LanguageName[name.lower()] = name
        if language.primary:
            self.primary.append(name)
        if language.internal:
            self.internal.append(name)
        for pat in language.shebangPatterns:
            self.shebangPatterns.append((name, pat))

    def getAccessKeys(self):
        languages = self.__languages.keys()
        languages.sort()
        return [self.__accesskeys[language] for language in languages]
    
    def getLanguage(self, language):
        # return a koILanguage for language.  Create it if it does not
        # exist yet.
        if not language: language=self.defaultLanguage
        
        if language not in self.__languages:
            log.warn("Asked for unknown language: %r", language)
            if language != self.defaultLanguage:
                return self.getLanguage(self.defaultLanguage)
            # Trouble if we can't load the default language.
            return None

        if self.__languages[language] is None:
            contractid = "@activestate.com/koLanguage?language=%s;1" \
                         % (language)
            self.__languages[language] = components.classes[contractid] \
                    .createInstance(components.interfaces.koILanguage)

        return self.__languages[language]

    def suggestLanguageForFile(self, filename):
        if self.__associations is None:
            self._resetFileAssociations()

        language = ''  # default

        # get the file extension
        parts = filename.split(".")

        if len(parts) > 0:
            ext = parts[len(parts)-1];
            if self.__absolutefilters.has_key(ext):
                return self.__absolutefilters[ext]

        if self.__absolutefilters.has_key(filename):
            return self.__absolutefilters[filename]
        
        # check for file pattern matches        
        for (pat, lang) in self.__associations:
            if fnmatch.fnmatch(filename, pat):
                language = lang
                break
            
        return language

    def _sendStatusMessage(self, msg, timeout=3000, highlight=1):
        observerSvc = components.classes["@mozilla.org/observer-service;1"]\
                      .getService(components.interfaces.nsIObserverService)
        sm = components.classes["@activestate.com/koStatusMessage;1"]\
             .createInstance(components.interfaces.koIStatusMessage)
        sm.category = "language_registry"
        sm.msg = msg
        sm.timeout = timeout     # 0 for no timeout, else a number of milliseconds
        sm.highlight = highlight # boolean, whether or not to highlight
        try:
            observerSvc.notifyObservers(sm, "status_message", None)
        except COMException, e:
            # do nothing: Notify sometimes raises an exception if (???)
            # receivers are not registered?
            pass

    emacsLocalVars1_re = re.compile("-\*-\s*(.*?)\s*-\*-")
    # This regular expression is intended to match blocks like this:
    #    PREFIX Local Variables: SUFFIX
    #    PREFIX mode: Tcl SUFFIX
    #    PREFIX End: SUFFIX
    # Some notes:
    # - "[ \t]" is used instead of "\s" to specifically exclude newlines
    # - "(\r\n|\n|\r)" is used instead of "$" because the sre engine does
    #   not like anything other than Unix-style line terminators.
    emacsLocalVars2_re = re.compile(r"^(?P<prefix>(?:[^\r\n|\n|\r])*?)[ \t]*Local Variables:[ \t]*(?P<suffix>.*?)(?:\r\n|\n|\r)(?P<content>.*?\1End:)",
                                    re.IGNORECASE | re.MULTILINE | re.DOTALL)
    def _getEmacsLocalVariables(self, head, tail):
        """Return a dictionary of emacs local variables.
        
            "head" is a sufficient amount of text from the start of the file.
            "tail" is a sufficient amount of text from the end of the file.
        
        Parsing is done according to this spec (and according to some
        in-practice deviations from this):
            http://www.gnu.org/software/emacs/manual/html_chapter/emacs_33.html#SEC485
        Note: This has moved to:
            http://www.gnu.org/software/emacs/manual/emacs.html#File-Variables
            
        A ValueError is raised is there is a problem parsing the local
        variables.
        """
        localVars = {}

        # Search the head for a '-*-'-style one-liner of variables.
        if head.find("-*-") != -1:
            match = self.emacsLocalVars1_re.search(head)
            if match:
                localVarsStr = match.group(1)
                if '\n' in localVarsStr:
                    raise ValueError("local variables error: -*- not "
                                     "terminated before end of line")
                localVarStrs = [s.strip() for s in localVarsStr.split(';') if s.strip()]
                if len(localVarStrs) == 1 and ':' not in localVarStrs[0]:
                    # While not in the spec, this form is allowed by emacs:
                    #   -*- Tcl -*-
                    # where the implied "variable" is "mode". This form is only
                    # allowed if there are no other variables.
                    localVars["mode"] = localVarStrs[0].strip()
                else:
                    for localVarStr in localVarStrs:
                        try:
                            variable, value = localVarStr.strip().split(':', 1)
                        except ValueError:
                            raise ValueError("local variables error: malformed -*- line")
                        # Lowercase the variable name because Emacs allows "Mode"
                        # or "mode" or "MoDe", etc.
                        localVars[variable.lower()] = value.strip()

        # Search the tail for a "Local Variables" block.
        match = self.emacsLocalVars2_re.search(tail)
        if match:
            prefix = match.group("prefix")
            suffix = match.group("suffix")
            lines = match.group("content").splitlines(0)
            #print "prefix=%r, suffix=%r, content=%r, lines: %s"\
            #      % (prefix, suffix, match.group("content"), lines)
            # Validate the Local Variables block: proper prefix and suffix
            # usage.
            for i in range(len(lines)):
                line = lines[i]
                if not line.startswith(prefix):
                    raise ValueError("local variables error: line '%s' "
                                     "does not use proper prefix '%s'"
                                     % (line, prefix))
                # Don't validate suffix on last line. Emacs doesn't care,
                # neither should Komodo.
                if i != len(lines)-1 and not line.endswith(suffix):
                    raise ValueError("local variables error: line '%s' "
                                     "does not use proper suffix '%s'"
                                     % (line, suffix))
            # Parse out one local var per line.
            for line in lines[:-1]: # no var on the last line ("PREFIX End:")
                if prefix: line = line[len(prefix):] # strip prefix
                if suffix: line = line[:-len(suffix)] # strip suffix
                line = line.strip()
                try:
                    variable, value = line.split(':', 1)
                except ValueError:
                    raise ValueError("local variables error: missing colon "
                                     "in local variables entry: '%s'" % line)
                # Do NOT lowercase the variable name, because Emacs only
                # allows "mode" (and not "Mode", "MoDe", etc.) in this block.
                localVars[variable] = value.strip()

        return localVars

    htmldoctype_re = re.compile('<!DOCTYPE\s+html',
                                re.IGNORECASE)
    def guessLanguageFromContents(self, head, tail):
        """Guess the language (e.g. Perl, Tcl) of a file from its contents.
        
            "head" is a sufficient amount of text from the start of the file
                where "sufficient" is undefined (although, realistically
                at least the first few lines should be passed in to get good
                coverage).
            "tail" is a sufficient amount of text from the end of the file,
                where "sufficient" is as above. (Usually the tail of the
                document is where Emacs-style local variables. Emacs'
                documentation says this block should be "near the end of
                the file, in the last page.")

        This method returns a list of possible languages with the more
        likely, or more specific, first. If no information can be gleaned an
        empty list is returned.
        """
        langs = []

        # Specification of the language via Emacs-style local variables
        # wins, so we check for it first.
        #   http://www.gnu.org/manual/emacs-21.2/html_mono/emacs.html#SEC486
        if self._globalPrefs.getBooleanPref("emacsLocalModeVariableDetection"):
            # First check for one-line local variables.
            # - The Emacs spec says this has to be in the _first_ line,
            #   but in practice this seems to be "near the top".
            try:
                localVars = self._getEmacsLocalVariables(head, tail)
            except ValueError, ex:
                self._sendStatusMessage(str(ex))
            else:
                if localVars.has_key("mode"):
                    mode = localVars["mode"]
                    try:
                        langName = self._modeName2LanguageName[mode.lower()]
                    except KeyError:
                        log.warn("unknown emacs mode: '%s'", mode)
                    else:
                        langs = [langName]

        # Detect if this is an XML file.
        if (self._globalPrefs.getBooleanPref('xmlDeclDetection') and
            not langs and head.lower().startswith(u'<?xml')):
            # it may be an XHTML file
            if self.htmldoctype_re.search(head):
                langs = ['HTML', 'XML']
            else:
                langs = ['XML', 'XSLT']

        # Detect the type from a possible shebang line.
        if (self._globalPrefs.getBooleanPref('shebangDetection') and
            not langs and head.startswith(u'#!')):
            shebangLangs = []
            for language, pattern in self.shebangPatterns:
                if pattern.match(head):
                    shebangLangs.append(language)
            if len(shebangLangs) > 1:
                self._sendStatusMessage("language determination error: "
                                        "ambiguous shebang (#!) line")
            else:
                langs = shebangLangs

        return langs
        
    def __initPrefs(self):
        self.__prefs = components \
                       .classes["@activestate.com/koPrefService;1"] \
                       .getService(components.interfaces.koIPrefService)\
                       .prefs
        # Observers will be QI'd for a weak-reference, so we must keep the
        # observer alive ourself, and must keep the COM object alive,
        # _not_ just the Python instance!!!
        # XXX - this is a BUG in the weak-reference support.
        # It should NOT be necessary to do this, as the COM object is
        # kept alive by the service manager.  I suspect that this bug
        # happens due to the weak-reference being made during
        # __init__.  FIXME!
        self._observer = xpcom.server.WrapObject(self,
                                      components.interfaces.nsIObserver)
        self.__prefs.addObserver(self._observer)



class koSourceFile:
    _com_interfaces_ = [components.interfaces.koISourceFile]
    _reg_contractid_ = "@activestate.com/koSourceFile;1"
    _reg_clsid_ = "{D579D901-7DBF-4fa6-AE7A-1A85DAC9CA83}" 

    _registry = None
    def __init__(self):
        if koSourceFile._registry is None:
            koSourceFile._registry = components.classes["@activestate.com/koLanguageRegistryService;1"].getService(components.interfaces.koILanguageRegistryService) 
        
    def init(self, scintilla):
        self._scin = scintilla
        self.suggestLanguage("Text")

    def getLanguageBlocks(self):
        #return SimpleEnumerator(self.__languageBlocks)
        return self._languageBlocks

    def getLanguageBlockAtPosition(self, position):
        # for now, we only allow one language block per file
        assert 0 <= position <= self._scin.length, "position out of range"
        assert len(self._languageBlocks) == 1, "internal inconsistency"
        
        return self._languageBlocks[0]

    def suggestLanguage(self, language):
        self._languageBlocks = [
            self._createLanguageBlock(language, 0, self._scin.length)
            ]
        
    def suggestFilename(self, basename):
        # make the association and create a language block appropriately
        language = koSourceFile._registry.suggestLanguageForFile(basename)
        self.suggestLanguage(language or 'Text')

    def _createLanguageBlock(self, language, offset, length):
        return koLanguageBlock(self._scin, language, offset, length, koSourceFile._registry)


class koLanguageBlock:
    _com_interfaces_ = [components.interfaces.koILanguageBlock]

    def __init__(self, scintilla, language, offset, length, registry):
        # for now, we only allow one language block per file
        assert offset == 0
        assert length == scintilla.length
        
        self.__scin = scintilla
        self.offset = offset
        self.length = length
        self.__language = language
        self.__registry = registry


    def get_language(self):
        return self.__registry.getLanguage(self.__language)
    
    def get_text(self):
        assert self.offset+self.length <= self.__scin.length
        
        return self.__scin.text[self.offset:self.offset+self.length]


    def getRange(self, start, end):
        assert self.offset+start <= self.__scin.length
        assert self.offset+end <= self.__scin.length
        assert start <= end
        
        return self.get_text()[start:end]




#---- self-test suite for the guessLanguageFromContents implementation

if __name__ == "__main__":
    import unittest
    from xpcom.server import UnwrapObject

    regSvc = components.classes['@activestate.com/koLanguageRegistryService;1'].\
             getService(components.interfaces.koILanguageRegistryService)
    regSvc = UnwrapObject(regSvc)
    
    class GuessLangTestCase(unittest.TestCase):
        def _assertPossibleLanguagesAre(self, head, tail, expected):
            actual = regSvc.guessLanguageFromContents(head, tail)
            if actual != expected:
                errmsg = """guessed possible languages do not match expected results:
Expected: %s
Got:      %s
---------------- head of document --------------------------------
%s
---------------- tail of document --------------------------------
%s
------------------------------------------------------------------
""" % (expected, actual, head, tail)
                self.fail(errmsg)

        def _assertEmacsLocalVarsAre(self, head, tail, expected):
            actual = regSvc._getEmacsLocalVariables(head, tail)
            if actual != expected:
                errmsg = """Emacs-style local variable results are not as expected:
Expected: %s
Got:      %s
---------------- head of document --------------------------------
%s
---------------- tail of document --------------------------------
%s
------------------------------------------------------------------
""" % (expected, actual, head, tail)
                self.fail(errmsg)

        def test_perl_shebang(self):
            perlHeads = [
                "#!perl",
                "#!perl5 -w",
                "#!/bin/perl",
                "#! /bin/perl -dwh  ",
                "#!/bin/PeRl",
                "#!/bin/myperl",
            ]
            for head in perlHeads:
                self._assertPossibleLanguagesAre(head, "", ["Perl"])
            notPerlHeads = [
                "foo",
                "#!/bin/erl",
            ]
            for head in notPerlHeads:
                self._assertPossibleLanguagesAre(head, "", [])
        def test_python_shebang(self):
            pythonHeads = [
                "#!python",
                "#!python22 -w",
                "#!/bin/python",
                "#! /bin/python -dwh  ",
                "#!/bin/PyThOn",
                "#!/bin/mypython",
            ]
            for head in pythonHeads:
                self._assertPossibleLanguagesAre(head, "", ["Python"])
            notPythonHeads = [
                "foo",
                "#!/bin/ython",
            ]
            for head in notPythonHeads:
                self._assertPossibleLanguagesAre(head, "", [])
        def test_tcl_shebang(self):
            tclHeads = [
                "#!tclsh",
                "#!tclsh82",
                "#!/bin/tclsh",
                "#!/bin/expect",
                "#! /bin/wish -v  ",
                "#!/bin/TcLSh",
                "#!/bin/mytclsh",
                """\
#!/bin/sh
# the next line restarts using tclsh \\
exec tclsh "$0" "$@"
""",
                """\
#!/bin/sh
# the next line restarts using tclsh \\
exec wish "$0" "$@"
""",
                """\
#!/bin/sh
# the next line restarts using tclsh \\
exec expect "$0" "$@"
""",
            ]
            for head in tclHeads:
                self._assertPossibleLanguagesAre(head, "", ["Tcl"])
            notTclHeads = [
                "foo",
                "#!/bin/clsh",
            ]
            for head in notTclHeads:
                self._assertPossibleLanguagesAre(head, "", [])

        def test_emacs_local_variables(self):
            # Ensure the pref to use emacs-style local variables is on.
            globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
                          getService(components.interfaces.koIPrefService).prefs
            emacsLocalModeVariableDetection = globalPrefs.getBooleanPref("emacsLocalModeVariableDetection")
            globalPrefs.setBooleanPref("emacsLocalModeVariableDetection", 1)

            headsAndTailsAndVarsAndLangs = [
                ("# -*- mode: Tcl -*-", "", {"mode": "Tcl"}, ["Tcl"]),
                ("# -*- Mode: Tcl -*-", "", {"mode": "Tcl"}, ["Tcl"]),
                ("# -*- mode: tcl -*-", "", {"mode": "tcl"}, ["Tcl"]),
                ("# *- mode: Tcl -*- blah blah", "", {}, []),
                ("", """\
# Using a prefix and suffix.
PREFIX Local Variables: SUFFIX
PREFIX mode: Tcl SUFFIX
PREFIX End: SUFFIX
""", {"mode": "Tcl"}, ["Tcl"]),
                ("", """\
Local Variables:
mode: Tcl
End:
""", {"mode": "Tcl"}, ["Tcl"]),
                ("", """\
# Using a realistic prefix.
# Local Variables:
# mode: Tcl
# End:
""", {"mode": "Tcl"}, ["Tcl"]),
                ("", """\
# Make sure the "End:" in a variable value does not screw up parsing.
PREFIX Local variables: SUFFIX
PREFIX foo: End: SUFFIX
PREFIX tab-width: 4 SUFFIX
PREFIX End: SUFFIX
""", {"foo": "End:", "tab-width": "4"}, []),
                ("", """\
# Must use proper prefix.
PREFIX Local Variables: SUFFIX
 mode: Tcl SUFFIX
PREFIX End: SUFFIX
""", ValueError, []),
                ("", """\
# Must use proper suffix.
PREFIX Local Variables: SUFFIX
PREFIX mode: Tcl
PREFIX End: SUFFIX
""", ValueError, []),
                ("", """\
# Whitespace after the prefix and before the suffix is allowed.
# The suffix on the "End:" line is not checked.
PREFIX Local Variables: SUFFIX
PREFIX    mode: Tcl	SUFFIX
PREFIX End:
""", {"mode": "Tcl"}, ["Tcl"]),
                ("", """\
# In the Local Variables block variable names are NOT lowercase'd, therefore
# the user must specify the lowercase "mode" to properly set the mode.
PREFIX Local Variables: SUFFIX
PREFIX MoDe: tcl SUFFIX
PREFIX End: SUFFIX
""", {"MoDe": "tcl"}, []),
                ("# -*- Mode: perl -*-", """\
# The local variables block beats the one-liner.
Local Variables:
mode: Tcl
End:
""", {"mode": "Tcl"}, ["Tcl"]),
            ]
            for head, tail, vars, langs in headsAndTailsAndVarsAndLangs:
                if isinstance(vars, dict):
                    self._assertEmacsLocalVarsAre(head, tail, vars)
                elif issubclass(vars, Exception):
                    self.assertRaises(ValueError, regSvc._getEmacsLocalVariables,
                                      head, tail)
                else:
                    raise "Unexpected test case 'vars' type: %s" % type(vars)
                self._assertPossibleLanguagesAre(head, tail, langs)

            # Restore the pref.
            globalPrefs.setBooleanPref("emacsLocalModeVariableDetection",
                                       emacsLocalModeVariableDetection)


        def test_bug28775(self):
            # The problem here was that no path info between the "exec"
            # and the "wish" was allowed.
            head = """\
#!/bin/sh
# aem \\
exec $AUTOTEST/bin/wish "$0" ${1+"$@"} &
"""
            self._assertPossibleLanguagesAre(head, "", ["Tcl"])

            # The problem here was that if there was a blank line with an
            # LF line terminator (i.e. Unix EOLs) before the "Local Variables"
            # line, then the LF would get included in the "prefix" and
            # subsequent lines would look like they didn't have the proper
            # prefix.
            tail = """
# ;;; Local Variables: ***
# ;;; mode: tcl ***
# ;;; End: ***
"""
            self._assertEmacsLocalVarsAre("", tail, {"mode": "tcl"})


    sys.argv.append('-v') # hackily make the test suite run in verbose mode
    unittest.main()



# Local Variables:
# mode: Python
# End:
