#!/usr/bin/env python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import sys, os, re, string, stat
import tempfile
from xpcom import components, ServerException, COMException, nsError
import process
import koprocessutils


#---- components


class koPHPConfigurator:
    _com_interfaces_ = [components.interfaces.koIPHPConfigurator]
    _reg_clsid_ = "2C8CD9F6-DDF4-44e8-89E7-D8B26F42BE26"
    _reg_contractid_ = "@activestate.com/koPHPConfigurator;1"
    _reg_desc_ = "PHP Debugger Configurator"

    def _make_full_version(self,version):
        try:
            ver = version.split('.')
            lv = len(ver)
            if lv == 1:
                return [string.atoi(ver[0]),0,0]
            elif lv == 2:
                return [string.atoi(ver[0]),string.atoi(ver[1]),0]
            elif lv == 3:
                return [string.atoi(ver[0]),string.atoi(ver[1]),string.atoi(re.match(r"(\d+)", ver[2]).group())]
        except:
            pass
        return None
    
    def _get_source_extension_path(self,filelist,version,komododir,origver=None):
        if not origver:
            origver = tuple(version)
            
        #verdir = "%d.%d.%d"%tuple(version)
        possible_dir = possible_ver = None
        for dir in filelist:
            dirver = self._make_full_version(dir)
            if not dirver: continue
            # Is my version too early for this version? if yes, skip
            if version < dirver: continue

            # is the previously found match a better match than
            # this one? (ie previous is greater)
            if possible_ver and possible_ver >= dirver: continue
                
            possible_ver = dirver
            possible_dir = dir
                
        if possible_dir:
            return possible_dir
                
        if version[2] > 0:
            version[2]-=1
            return self._get_source_extension_path(filelist,version,komododir,origver)
        if version[1] > 0:
            version[2] = origver[2]
            version[1]-=1
            return self._get_source_extension_path(filelist,version,komododir,origver)
        if version[0] > 4:
            version[1] = origver[1]
            version[2] = origver[2]
            version[0]-=1
            return self._get_source_extension_path(filelist,version,komododir,origver)
        return None

    def configureDebugger(self, executable, oldini, inipath, extensiondir):
        # I want a seperate instance of infoex, since we'll be doing some
        # extra special stuff.
        phpInfoEx = components.classes['@activestate.com/koPHPInfoInstance;1'].\
                    createInstance(components.interfaces.koIPHPInfoEx);
        import shutil
        if not os.path.exists(executable):
            return 'Executable path %s does not exist' % executable
        if not os.path.isfile(oldini):
            return 'INI path %s does not exist' % executable
        phpInfoEx.executablePath = executable
        
        # get the version so we can figure out
        # which extension to install
        versionText = phpInfoEx.version;
        if not versionText:
            return 'Unable to retrieve PHP version'
        version = self._make_full_version(versionText)
        if not version:
            return 'Unable to retrieve PHP version [',versionText,'].'
        
        # build the path to the extension in komodo/php
        koDirSvc = components.classes["@activestate.com/koDirs;1"]\
                         .getService(components.interfaces.koIDirs)
        koSupportDir = koDirSvc.supportDir
        # XXX only windows and linux supported here!
        if sys.platform.startswith('win'):
            extensionname = 'php_xdebug.dll'
            extensionlabel = 'zend_extension_ts'
        else:
            extensionname = 'xdebug.so'
            extensionlabel = 'zend_extension'
        
        source_extension = os.path.join(koSupportDir, 'php', 'debugging')
        try:
            dirlist = os.listdir(source_extension)
        except:
            return 'Unable to find debugger extensions at [%s]'%source_extension
        source_extension_ver = self._get_source_extension_path(dirlist, version, koSupportDir)
       
        if not source_extension:
            return 'Debugger extension for PHP version %s is not available.' % versionText

        source_extension = os.path.join(source_extension,source_extension_ver,extensionname)
        
        # copy the old ini to the debug location
        try:
            if not os.path.isdir(inipath):
                os.makedirs(inipath)
        except:
            return 'Unable to make directory [%s]' %inipath
        ininame = os.path.join(inipath,'php.ini')
        try:
            shutil.copy(oldini,ininame)
        except:
            return 'Unable to copy [%s] to [%s]' %(oldini,inipath)
        
        try:
            _stat = os.stat(ininame)
            mode = _stat[stat.ST_MODE]
            writable = mode & stat.S_IWRITE == stat.S_IWRITE
            if not writable:
                os.chmod(ininame, mode | stat.S_IWUSR)
        except:
            return 'Unable to make ini writable [%s]' %inipath

        # modify the ini file
        try:
            ini = open(ininame)
            contents = ini.read()
            ini.close()
        except:
            return 'Unable to read ini at [%s]' % inipath
        
        # remove previous configurations of xdebug
        contents=contents.replace('xdebug',';xdebug')
        # disable any other zend extensions, they tend to interfere with debuggers
        # these are not normal php extensions
        contents=contents.replace('zend_extension',';zend_extension')
        # remove old activedebug stuff
        contents=contents.replace('extension=activedebug',';extension=activedebug')
        contents=contents.replace('activedebug',';activedebug')
        
        newcontents = """; xdebug config added by Komodo
%s=%s
xdebug.remote_enable=1
xdebug.remote_handler=dbgp
xdebug.remote_mode=req
xdebug.remote_port=9000
%s
""" % \
            (extensionlabel,
            os.path.join(extensiondir, extensionname),
            contents)
        
        try:
            ini = open(ininame,'w+')
            ini.write(newcontents)
            ini.close()
            if sys.platform[:3] != 'win':
                os.chmod(ininame,0644)
        except:
            return 'Unable to write ini at [%s]' % inipath
        
        # copy the debugger extension to the extension dir
        try:
            if not os.path.isdir(extensiondir):
                os.makedirs(extensiondir)
        except:
            return 'Unable to create directory [%s]' % extensiondir
        dest = os.path.join(extensiondir,extensionname)
        try:
            shutil.copyfile(source_extension,dest)
            if sys.platform[:3] != 'win':
                os.chmod(dest,0755)
        except:
            return 'Unable to copy [%s] to [%s]'%(source_extension,dest)
        
        phpInfoEx.cfg_file_path = inipath
        # validate that this all worked!
        if phpInfoEx.isDebuggerExtensionLoadable:
            return ''
        return 'PHP is configured, but was unable to load the debugger extension.'


