# Copyright (c) 2005-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

# A holder of remote file information, I.e. for ftp, sftp etc...
#
# Contributors:
# * Todd Whiteman

from xpcom import file, components, COMException, ServerException, server, nsError
from xpcom.server.enumerator import SimpleEnumerator
from xpcom.client import WeakReference

import os, string, time, stat
import strptime
import re
import logging
import threading

# Regex parsers for processing directory listings
re_parse_unix_ftp_line     = re.compile(r'^.([+rwxXstugo-]*)\s+(.*?\s+.*?\s+.*?)\s*(\d*?)\s+(\w+)\s+(\d+)\s+([\d:]+)\s(.*)$')
re_parse_unix_ftp_line_alt = re.compile(r'^.([+rwxXstugo-]*)\s+(.*?\s+.*?\s+.*?)\s+(\d*?)(\s+)([\d-]+)\s+([\d:]+)\s(.*)$')
re_parse_dos_ftp_line      = re.compile(r'^(..-..-..\s+..:....)\s+(.*?)\s+(.*)$')

# koRemoteFileInfo will store information about a remote file/directory.
# This includes permissions, file type, size and path.
class koRemoteFileInfo:
    """Remote file information"""
    _com_interfaces_ = [components.interfaces.koIRemoteFileInfo]
    _reg_desc_ = "Komodo Remote File Info"
    _reg_contractid_ = "@activestate.com/koRemoteFileInfo;1"
    _reg_clsid_ = "{de59c592-44b5-44c6-aa15-b9c6059416be}"

    def __init__(self):
        # Not all these may be set, depending upon the remote OS
        # and it's SFTP server
        self.log = logging.getLogger('koRemoteFileInfo')
        #self.log.setLevel(logging.DEBUG)
        self.path     = ''      # Full path to the file
        self.filename = ''      # Just the name of the file without directory
        self.dirname  = ''      # The directory the file is help within
        self.st_size  = 0
        self.st_uid   = 0
        self.st_gid   = 0
        self.st_mode  = 0
        self.st_mtime = 0
        self.children = []
        self.link_target = None
        self.lastListedTime = 0

    # Boolean checker methods
    def isFile(self):
        return stat.S_ISREG(self.st_mode)
    def isDirectory(self):
        return stat.S_ISDIR(self.st_mode)
    def isSymlink(self):
        return stat.S_ISLNK(self.st_mode)
    def isReadable(self):
        return (self.st_mode & stat.S_IREAD) == stat.S_IREAD
    def isWriteable(self):
        return (self.st_mode & stat.S_IWRITE) == stat.S_IWRITE
    def isExecutable(self):
        return (self.st_mode & stat.S_IEXEC) == stat.S_IEXEC
    def isHidden(self): # Hmmm...
        # XXX - Whats hidden for, how is it used??
        return 0
    def needsDirectoryListing(self):
        # Must be a directory that has not had it's children set
        return (self.isDirectory() and not self.lastListedTime)

    # Getter methods
    def getFilepath(self):
        return self.path
    def getFilename(self):
        return self.filename
    def getDirname(self):
        return self.dirname
    def getLinkTarget(self):
        return self.link_target
    def getFileSize(self):
        return self.st_size
    def getModifiedTime(self):
        return self.st_mtime
    def getChildren(self):
        return self.children

    # Setter methods
    def setChildren(self, children):
        self.children = children
        self.lastListedTime = time.time()
    def setLinkTarget(self, link_target):
        self.link_target = link_target
    def setMode(self, mode):
        self.st_mode = mode

    # Internal for class usage
    def _setpathinfo(self, dirname, filename, link_target=None):
        filename = os.path.basename(filename)
        if filename not in ('', '/', '~', '~/', '.', './'):
            self.path = dirname + '/' + filename
            self.dirname = dirname
        else:
            self.dirname = filename
            self.path = filename
        self.filename = filename
        if link_target:
            if link_target[0] in ('/', '~'):
                # Absolute path
                self.link_target = link_target
            else:
                # Relative path
                self.link_target = self.dirname + '/' + link_target

    # Initializers
    # Init from file attributes, same as used by the stat and os modules
    def initFromStats(self, dirname, filename, size, uid, gid, mode, mtime):
        self._setpathinfo(dirname, filename)
        self.st_size  = size
        self.st_uid   = uid
        self.st_gid   = gid
        self.st_mode  = mode
        self.st_mtime = mtime
        self.log.debug("initFromStats: %s", str(self))

    # Init from raw ftp directory listing
    def initFromDirectoryListing(self, dirname, line):
        self.st_mode = 0
        self.dirname = dirname
        link_target = None

        if not line:
            log.debug("initFromDirectoryListing: No line to process!")
            return False

        if line[0] in ["-","d","l"]: # unix style lines
            fig = re_parse_unix_ftp_line.search(line)
            if not fig:
                fig = re_parse_unix_ftp_line_alt.search(line)
                if not fig:
                    return False
            fi = fig.groups()

            if line[0] == "d":
                self.st_mode |= stat.S_IFDIR    # directory
            elif line[0] == "-":
                self.st_mode |= stat.S_IFREG    # regular file
            else:
                self.st_mode |= stat.S_IFLNK    # symlink
                self.log.debug("THIS IS A SYMLINK!!!")
            # we unfortunately dont know if we are owner or group member, so guess again
            if "r" in fi[0]: self.st_mode |= stat.S_IREAD
            if "w" in fi[0]: self.st_mode |= stat.S_IWRITE
            if "x" in fi[0]: self.st_mode |= stat.S_IEXEC

            if fi[2]: self.st_size = string.atoi(fi[2]) # File size

            # Date is in fi[3], I.e Nov 30 or 2005-11-30 format
            # do we have time, or year?
            sp = fi[4].split('-', 2)
            if len(sp) == 3:
                # 2005-11-30 format
                year, month, day = sp
                hour = fi[5]
                date = string.join([year,month,day,hour]," ")
                t = strptime.strptime(date,'%Y %m %d %H:%M')
            else:
                month,day,year = fi[3:6]
                if year.find(":") > -1:
                    hour = year
                    # fix time to be 5 digit always
                    year = "%d"%(time.gmtime(time.time())[0])
                else:
                    hour = "00:00"
                
                if len(day) < 2: day = "0"+day            
                if len(hour) < 5: hour = "0"+hour

                date = string.join([month,day,year,hour]," ")
                t = strptime.strptime(date,'%b %d %Y %H:%M')
            
            # print "date trans: ",repr(t)
            try:
                self.st_mtime = time.mktime(t)
            except:
                self.st_mtime = 0  # Unknown date, set it as 0, and 1970 it is.

            if self.isSymlink():
                filename, link_target = fi[6].split(" -> ")
                self.log.debug("Symlink %s -> %s", filename, link_target)
            else:
                filename = fi[6]

        else: # msdos style lines
            fig = re_parse_dos_ftp_line.search(line)
            if not fig: return False
            fi = fig.groups()
            # we simply don't know, so say yes!

            self.st_mode |= stat.S_IREAD
            self.st_mode |= stat.S_IWRITE
            self.st_mode |= stat.S_IEXEC
            # self.st_mode |= stat.S_IFDIR    # directory
            self.st_mode |= stat.S_IFREG    # regular file
            # self.st_mode |= stat.S_IFLNK    # symlink
            
            try:
                self.st_size = string.atoi(fi[1])
            except ValueError, e:
                # this is a directory
                if str(fi[1][:5]) == '<DIR>':
                    self.st_mode |= stat.S_IFDIR    # directory
                    self.st_mode -= stat.S_IFREG    # regular file
                    self.st_size = 0
                else:
                    pass
            t = strptime.strptime(fi[0],'%m-%d-%y  %I:%M%p')
            # print "date trans: ",repr(t)
            try:
                self.st_time = time.mktime(t)
            except:
                self.st_mtime = 0  # Unknown date, set it as 0, and 1970 it is.
            filename = fi[2]

        # Set the full path
        # XXX - Windows/DOS filenames break ??
        self._setpathinfo(dirname, filename, link_target)
        self.log.debug("initFromDirectoryListing: %s", str(self))
        return True

    # Readonly attributes
    def get_size(self):
        return self.st_size
    def get_mode(self):
        return self.st_mode
    def get_uid(self):
        return self.st_uid
    def get_gid(self):
        return self.st_gid
    def get_mtime(self):
        return self.st_mtime

    def __str__(self):
        s = ""
        if self.isDirectory():
            s += "D"
        if self.isFile():
            s += "F"
        if self.isReadable():
            s += "r"
        if self.isWriteable():
            s += "w"
        if self.isSymlink():
            return "%s (link -> %s) m:%s sz:%d c:%d %s fp:'%s'" % (self.filename, self.link_target, s, self.st_size, len(self.children), time.asctime(time.localtime(self.st_mtime)), self.path)
        else:
            return "%s m:%s sz:%d c:%d %s fp:'%s'" % (self.filename, s, self.st_size, len(self.children), time.asctime(time.localtime(self.st_mtime)), self.path)

    def __repr__(self):
        return str(self)



# Test function
def _test():
    pass

# When run from command line
if __name__ == '__main__':
    _test()

