#!python
# Copyright (c) 2004-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

"""Get and manage Code Intelligence data about source code of many languages.

The Code Intelligence system is one for generating and managing code
structure information on given source code. See the spec for more details:
    http://specs.activestate.com/Komodo_3.0/func/code_intelligence.html
Usage examples follow.

Scan a file and return data in CIX format:

    import codeintel
    cix = codeintel.scan("foo.py", "Python")

General program structure for a runtime C.I. system:

    # Initialize the manager at application startup.
    manager = codeintel.Manager()
    manager.initialize()

    # Request scans on certain files.
    # - Scans are performed asynchronously and data is loaded into the
    #   database (CIDB).
    r = codeintel.ScanRequest(os.path.join("play", "sample.py"),
                              "Python",
                              codeintel.PRIORITY_OPEN)
    manager.addRequest(r)

    #XXX Describe how to get data out, both synchronously and asynchronously
    #    (by getting a notification when requested scans are completed).

    # Finalize the manager when shutting down app.
    manager.finalize()
"""
# Dev Notes on PySQLite and Multi-threading:
# - There seem to be some caveats when using multi-threading (particularly
#   on Un*x) with PySQLite/SQLite. I have submitted one crash report to
#   pysqlite-users: avoid usage of Connection.db.sqlite_busy_handler().
#   I think multi-threading will work fine, but we should be able to move
#   to all database modification coming from one thread if necessary: have
#   a results Queue and the scheduler thread can feed those results into the
#   database; rather than the current situation where each _ScanRunner thread
#   does the db updating. c.f. the THREADSAFE flags for building PySQLite
#   and/or SQLite.
# - We _do_ have to properly handle SQL_BUSY retvals -> OperationalError
#   exceptions in the Python binding.

import os
import sys
import time
import md5
import pprint
import threading
import stat
import types
import shutil

import logging
import apsw
import process

from codeintel.common import *
from codeintel.scheduler import Scheduler, BatchScheduler
from codeintel.database import Database
from codeintel.engines import LanguageEngineRegistry
from codeintel import citdl
from codeintel import rwlock


#---- global variables

_version_ = (0, 2, 0)
log = logging.getLogger("codeintel.__init__")
#log.setLevel(logging.INFO)



#---- public convenience methods

def scan(filename, language, md5sum=None, mtime=None):
    """Scan the given filename and return code intelligence data (CIX).
    
        "filename" is the path to a file to scan.
        "language" is a string indicating what kind of content the file
            holds. This is used to get an appropriate Language Engine Driver
            from the LanguageEngineRegistry. Generally the proper string
            is just the name of the language, e.g., "Python", "Perl".
        "md5sum" (optional) if the MD5 hexdigest has already been calculated
            for the content, it can be passed in here. Otherwise this
            is calculated.
        "mtime" (optional) is a modified time for the file (in seconds since
            the "epoch"). If it is not specified the file is stat'd.

    If there are no language engines for the given content type a
    CodeIntelError is raised.
    """
    log.debug("scan(filename='%s', language='%s', md5sum=%r, mtime=%r)",
              filename, language, md5sum, mtime)
    if mtime is None:
        mtime = int(os.stat(filename)[stat.ST_MTIME])
    fin = open(filename, "r")
    try:
        s = fin.read()
    finally:
        fin.close()
    return scanString(s, language, path=filename, md5sum=md5sum, mtime=mtime)


def scanString(s, language, path="<string>", md5sum=None, mtime=None):
    """Scan the given filename and return code intelligence data (CIX).
    
        "s" is the content to scan.
        "language" is a string indicating what kind of content the file
            holds. This is used to get an appropriate Language Engine Driver
            from the LanguageEngineRegistry. Generally the proper string
            is just the name of the language, e.g., "Python", "Perl".
        "path" (optional) is the path from which the content originates.
            This path appears in the resulting XML so may be relevant. If
            not specified "<string>" is used.
        "md5sum" (optional) if the MD5 hexdigest has already been calculated
            for the content, it can be passed in here. Otherwise this
            is calculated.
        "mtime" (optional) is a modified time for the file (in seconds since
            the "epoch"). If it is not specified the _current_ time is used.
            Note that the default is not to stat() the file and use that
            because the given content might not reflect the saved file state.

    If there are no language engines for the given content type a
    CodeIntelError is raised.
    """
    if md5sum is None: md5sum = md5.new(s).hexdigest()
    if mtime is None: mtime = int(time.time())
        
    registry = LanguageEngineRegistry()
    drivers = registry.getDrivers(language)
    if not drivers:
        raise CodeIntelError("There are no Code Intelligence Language "
                             "Engines registered for '%s' content"
                             % language)
    mergedCIX = ""
    for driver in drivers:
        r = ScanRequest(path, language, PRIORITY_IMMEDIATE,
                        content=s, md5sum=md5sum, mtime=mtime)
        cix = driver.scan(r)
        if not mergedCIX:
            mergedCIX = cix
        else:
            raise NotImplementedError("Can't yet merge CIX documents.")
    return mergedCIX



#---- internal support stuff

_gPyFuncTypes = (
    types.BuiltinFunctionType,
    types.BuiltinMethodType,
    types.FunctionType,
    types.GeneratorType,
    types.LambdaType,
    types.MethodType,
    types.UnboundMethodType,
    types.ModuleType,
    types.ClassType,
)
def _getTypeNameForPyObj(obj):
    if isinstance(obj, _gPyFuncTypes):
        typeName = "function"
    elif isinstance(obj, types.ModuleType):
        typeName = "module"
    elif isinstance(obj, types.ClassType):
        typeName = "class"
    else:
        typeName = "variable"
    return typeName

def _getPyClassMembers(obj):
    # Recurse into base classes looking for attributes
    typesAndMembers = []
    try:
        members = dir(obj)
        for b in obj.__bases__:
            for m in _getPyClassMembers(b):
                if m not in members:
                    members.append(m)
    except AttributeError:
        pass
    return members

def _findPyClassConstructor(classObj):
    # Given a class object, return a function object used for the
    # constructor (ie, __init__() ) or None if we can't find one.
    try:
        return classObj.__init__.im_func
    except AttributeError:
        for base in classObj.__bases__:
            rc = _findPyClassConstructor(base)
            if rc is not None:
                return rc
    return None

def _getPyObjCallTip(obj):
    """Return a suitable call tip for the given Python object."""
    if obj is None:
        return ""

    name = ""
    argStr = ""
    doc = None
    argOffset = 0
    if type(obj) == types.ClassType:
        # Look for the highest __init__ in the class chain.
        funcObj = _findPyClassConstructor(obj)
        if funcObj is None:
            funcObj = lambda: None
        else:
            argOffset = 1
        name = "%s.%s" % (obj.__name__, funcObj.__name__)
    elif type(obj) == types.MethodType:
        # bit of a hack for methods - turn it into a function
        # but we drop the "self" param.
        funcObj = obj.im_func
        argOffset = 1
        name = funcObj.__name__
    else:
        funcObj = obj
        try:
            name = funcObj.__name__
        except AttributeError:
            pass
    # Try and build one for Python defined functions
    if type(funcObj) in (types.FunctionType, types.LambdaType):
        try:
            realArgs = funcObj.func_code.co_varnames[argOffset:funcObj.func_code.co_argcount]
            defaults = funcObj.func_defaults or []
            defaults = list(map(lambda name: "=%s" % name, defaults))
            defaults = [""] * (len(realArgs)-len(defaults)) + defaults
            items = map(lambda arg, dflt: arg+dflt, realArgs, defaults)
            if funcObj.func_code.co_flags & 0x4:
                items.append("...")
            if funcObj.func_code.co_flags & 0x8:
                items.append("***")
            argStr = ", ".join(items)
            argStr = "(%s)" % argStr
        except:
            pass
    # See if we can use the docstring.
    if hasattr(obj, "__doc__") and obj.__doc__:
        pos = obj.__doc__.find("\n")
        if pos<0 or pos>70: pos=70
        doc = obj.__doc__[:pos]
    # Put it together.
    calltip = name+argStr
    if doc:
        if calltip: calltip += "\n"  # calltip can be empty for, e.g., str(
        calltip += doc
    return calltip


#---- public Manager class

class Manager:
    """Manage a long-running interaction with the Code Intelligence system.
    
    This class can be sub-classed to customize some handling. The following
    methods are useful for overriding:
        requestCompleted(request)
    """
    MIN_CIDB_VERSION = (1, 0)  # minimum supported database version

    def __init__(self, registry=None, coreLanguages=[]):
        """Create a Code Intel Manager.
        
            "registry" (optional) can be used to specify a customized
                LanguageEngineRegistry instance to use for mapping languages
                to CILEs (Code Intel Language Engines). If not specified the
                default set of mappings is used.
            "coreLanguages" (optional) is a list of languages that will be
                considered "core" to this managed CodeIntel system.
                Effects:
                - the CIDB will endeavor to always have the standard CIX
                  files for core languages loaded
                - when loading CIX for this language, scan of imported
                  modules will be scheduled
        """
        if registry is None:
            self._cileRegistry = LanguageEngineRegistry()
        else:
            self._cileRegistry = registry
        self.coreLanguages = coreLanguages
        self._importHandlers = {}
        self._scheduler = None # the scheduler thread, started as required
        self._dbPath = None # the path to the SQLite database file
        self._dbLock = rwlock.RWLock() # guard for the CIDB
        
        # Used for batch updates.
        self._batchScheduler = None
        self._lastBatchUpdateErrors = None
        self._batchProgressCache = None
        self._batchProgressQuiet = False
        self._batchUpdateCompletedEvent = threading.Event()
        # Boolean indicating if a late-created main scheduler should NOT be
        # started immediately (because a BatchScheduler is already running).
        self._waitToStartScheduler = False
        
        # All caching (mostly for CITDL evaluation) is done in this dict.
        # Any subsystem using this cache will/should carve out it own area
        # by creating a sub-dictionary at a unique key.
        self.cache = {}
        self.factory = citdl.ThingyFactory(self.cache)

    def getCIDBUpgradeInfo(self, dbPath=None):
        """Returns information indicating if a CIDB upgrade is necessary and
        possible.
        
            "dbPath" (optional) is the full path to the SQLite database
                to use. If not specified the default, "./codeintel.db",
                is used.

        Returns one of the following:
            (CIDB_UPGRADE_NOT_NECESSARY, None)
            (CIDB_UPGRADE_NOT_POSSIBLE, "<reason>")
            (CIDB_UPGRADE_NECESSARY, None)
        
        This method is intended to be used as the first step of the
        initialization of the CodeIntel system (unless upgrades are not
        desired).
        """
        if not dbPath:
            dbPath = os.path.abspath("codeintel.db")

        if not os.path.isfile(dbPath):
            return (CIDB_UPGRADE_NOT_NECESSARY, None)
        
        db = Database(self)
        return db.getUpgradeInfo(dbPath, self.MIN_CIDB_VERSION)

    def resetCIDB(self, dbPath=None, backup=True):
        """Move the given database out of the way to make way for a new one.
        
            "dbPath" (optional) is the full path to the SQLite database
                to reset. If not specified the default, "./codeintel.db",
                is used.
            "backup" (optional, default True) is a boolean indicating if
                the original database should be backed up. If so, the backup
                is $dbPath+".err".
        """
        if not dbPath:
            dbPath = os.path.abspath("codeintel.db")

        if not os.path.isfile(dbPath):
            return
        if backup:
            errDBPath = dbPath + ".err"
            log.info("backing up current database to '%s'", errDBPath)
            if os.path.exists(errDBPath):
                os.remove(errDBPath)
                for i in range(10): # Try to avoid OSError from slow-deleting NTFS
                    if not os.path.exists(errDBPath): break
                    time.sleep(1)
            os.rename(dbPath, errDBPath)
        else:
            os.remove(dbPath)

    def upgradeCIDB(self, dbPath=None, restore=True):
        """Upgrade the given database to the current version.
        
            "dbPath" (optional) is the full path to the SQLite database
                to use. If not specified the default, "./codeintel.db",
                is used.
            "restore" (optional, default True) is a boolean indicating if
                the original database should be restored on failure.

        Typically this is only called if .getCIDBUpgradeInfo() returns
        CIDB_UPGRADE_NECESSARY.
        """
        if not dbPath:
            dbPath = os.path.abspath("codeintel.db")

        if not os.path.isfile(dbPath):
            raise CodeIntelError("cannot upgrade '%s': file does not exist" % dbPath)

        if restore:
            errDBPath = dbPath + ".err"
            if os.path.exists(errDBPath):
                os.remove(errDBPath)
            backupDBPath = dbPath + ".bak"
            log.info("backing up current database to '%s'", backupDBPath)
            if os.path.exists(backupDBPath):
                os.remove(backupDBPath)
            shutil.copy(dbPath, backupDBPath)

        try:
            self.batchUpdateRequest("upgrade", dbPath)
            self.batchUpdateStart(quiet=True)
            self.batchUpdateWait()
            errors = self.batchUpdateGetErrors()
            if errors:
                raise CIDBError("errors upgrading CIDB: '%s'"
                                % "', '".join(errors))
        except:
            if restore:
                if os.path.exists(errDBPath): # should be rare as it is removed at start
                    os.remove(errDBPath)
                for i in range(10): # Try to avoid OSError from slow-deleting NTFS
                    if not os.path.exists(errDBPath): break
                    time.sleep(1)
                log.info("saving failed upgrade to '%s'", errDBPath)
                os.rename(dbPath, errDBPath)
                log.info("restoring original database to '%s'", dbPath)
                os.rename(backupDBPath, dbPath)
            raise
        else:
            if restore:
                log.info("removing backup database '%s'", backupDBPath)
                os.remove(backupDBPath)

    def initialize(self, dbPath=None):
        """Initialize the Code Intelligence system.
        
            "dbPath" (optional) is the full path to the SQLite database
                to use. If not specified the default, "./codeintel.db",
                is used.
        
        If the database path does not exist when this is called then a new
        database is created. A CodeIntelError, DatabaseError, or
        EnvironmentError may be raised if the database path cannot be read or
        written to. The database is wiped if it is unsupported (e.g.,
        insufficient version, incorrect platform).
        """
        if not dbPath:
            self._dbPath = os.path.abspath("codeintel.db")
        else:
            self._dbPath = dbPath

        # For now we just trash the existing database if it is too old or
        # whatever. Eventually we may want to provide a richer mechanism:
        # - error handling in a separate (overridable) method with default of
        #   trashing the database
        # - think about accommodating a UI so that a status dialog can be put
        #   up while a conversion is done
        if os.path.isfile(self._dbPath):
            try:
                # Ensure we support this database version.
                log.info("open CIDB at '%s'" % self._dbPath)
                db = Database(self)
                if db.getVersionInfo() < self.MIN_CIDB_VERSION:
                    raise CodeIntelError(
                        "database '%s' version, %s, is unsupported"
                        % (self._dbPath, db.getVersion()))
                if db.getHostname().lower() != getHostname().lower():
                    raise CodeIntelError(
                        "database '%s' hostname, '%s', does not match "
                        "the current hostname: '%s'"
                        % (self._dbPath, db.getHostname(),
                           getHostname()))
                if db.getPlatform() != getPlatform():
                    raise CodeIntelError(
                        "database '%s' platform, '%s', does not match "
                        "the current platform: '%s'"
                        % (self._dbPath, db.getPlatform(),
                           getPlatform()))
            except CodeIntelError, ex:
                log.warn("%s: removing database" % ex)
                self._removeDatabase()

        if not os.path.isfile(self._dbPath):
            self._createNewDatabase()

    def _removeDatabase(self):
        try:
            os.remove(self._dbPath)
        except EnvironmentError, ex:
            # Try a little harder.
            time.sleep(1)
            try:
                os.chmod(self._dbPath, 0777)
                os.remove(self._dbPath)
                if os.path.exists(self._dbPath):
                    raise EnvironmentError("File remains after attempted "
                                           "removal: '%s'" % self._dbPath)
            except EnvironmentError, ex2:
                raise CodeIntelError("Could not remove database: it must "
                                     "be removed manually: %s" % ex2)

    def _createNewDatabase(self):
        # Create a new database.
        log.info("create new CIDB at '%s'" % self._dbPath)
        db = Database(self)
        db.create()

    def finalize(self):
        if self._batchScheduler:
            self._batchScheduler.stop()
        if self._scheduler:
            self._scheduler.stop()
        self._connections = {}

    def _startScheduler(self):
        if self._dbPath is None:
            raise CodeIntelError("must initialize the Manager before "
                                 "starting the scheduler")
        self._scheduler = Scheduler(Scheduler.MODE_DAEMON, self,
                                    self.requestCompleted)
        if not self._waitToStartScheduler:
            self._scheduler.start()

    def acquireCIDBReadLock(self):
        return self._dbLock.acquire('r')
    def releaseCIDBReadLock(self):
        return self._dbLock.release('r')
    def acquireCIDBWriteLock(self):
        return self._dbLock.acquire('w')
    def releaseCIDBWriteLock(self):
        return self._dbLock.release('w')

    def getDatabaseConnection(self, dbPath=None):
        """Return a connection to the database.
        
            "dbPath" (optional) is the path to a SQLite database to use.
                By default the path configured for this Manager instance is
                used.
        
        Note that one must properly acquire and release an appropriate lock
        on the CIDB before using this connection with the
        {acquire|release}CIDB{Read|Write}Lock() methods.
        """
        if dbPath is None:
            dbPath = self._dbPath
        return apsw.Connection(dbPath)

    def getCILEDrivers(self, language):
        """Return all registered Language Engine Drivers for this language.

        Returns a list of LanguageEngineDriver instances. If there are no
        engines for this langage an empty list is returned.
        """
        return self._cileRegistry.getDrivers(language)

    def getImportHandler(self, language):
        """Return an "import"-handler for the given language.
        
        Returns None if don't know how to handle imports for this language.
        Each language-specific ImportHandler object is a singleton.
        """
        if language not in self._importHandlers:
            handler = citdl.getImportHandler(language)
            self._importHandlers[language] = handler
        return self._importHandlers[language]

    def isSupportedLanguage(self, language):
        """Return true iff the given language is supported."""
        return language in self._cileRegistry.getLanguages()

    def getSupportedLanguages(self):
        """Return a list of the supported languages."""
        return self._cileRegistry.getLanguages()

    def addRequest(self, request):
        """Make a scan request.
        
            "request" is a ScanRequest instance.
        """
        if self._scheduler is None:  # lazily create the scheduler thread
            self._startScheduler()
        self._scheduler.addRequest(request)

    def stageRequest(self, request):
        """Stage a scan request.
        
            "request" is a ScanRequest instance.
        
        Staging a scan request involves putting it "on deck" for a certain
        period (called the "staging delay") before actually queuing it.
        Subsequently staged requests for the same file within that period
        replace the earlier request on deck. The point is to reduce
        unnecessary scanning for a rapidly changing file.
        """
        if self._scheduler is None:  # lazily create the scheduler thread
            self._startScheduler()
        self._scheduler.stageRequest(request)

    def requestCompleted(self, request):
        """Called when a scan request is successfully completed.
        
            "request" is the ScanRequest instance.

        This is made available for sub-classes to override. By default this
        method does nothing. Note: This is called in the sub-thread used to
        complete that request.
        """
        pass

    def batchUpdateRequest(self, type, path, language=None, extra=None):
        """Queue up a request for a particular batch update of the db.
        
            "type" is one of:
                language    scan a language installation
                cix         import a CIX file
                directory   scan a source directory
                upgrade     upgrade a CIDB file to the current version
            "path", depending on "type" is the full path to:
                language    a language installation
                cix         a CIX file
                directory   a source directory
                upgrade     a CIDB file
            "language", depending on "type" is:
                language    the language of the language installation
                cix         (not relevant, should be None)
                directory   the language of the source directory
                upgrade     (not relevant, should be None)
            "extra" is an optional (null if not used) extra value depending
                on the type, path and/or language of the request that may be
                request for processing it. For example, a PHP language batch
                update request uses the "extra" field to specify the
                "php.ini"-config-file path.

        Note: this cannot be called between .bacthUpdateStart() and
        .batchUpdateCompleted().
        """
        if self._batchScheduler is None:
            self._batchScheduler = BatchScheduler(self,
                                                  self.batchUpdateProgress,
                                                  self._batchUpdateCompleted,
                                                  self.batchUpdateError)
        self._batchScheduler.addRequest(type, path, language, extra)

    def batchUpdateStart(self, quiet=False):
        """Start the batch update.
        
            "quiet" (optional, default false) is a boolean indicating if
                batch update progress should be shown on the console.
        """
        # Setup thread(s) to handle the batch requests and return. (I.e. this
        # is asynchronous.
        # Put the main scheduler on hold.
        if self._scheduler is not None:
            self._scheduler.pause()
        else:
            self._waitToStartScheduler = True
        self._batchUpdateCompletedEvent.clear()
        self._batchProgressQuiet = quiet
        self._lastBatchUpdateErrors = []
        self._batchProgressCache = {}
        self._batchScheduler.start()

    def batchUpdateStop(self):
        """Stop the current batch update."""
        self._batchScheduler.stop()

    def batchUpdateWait(self):
        """Convenience method to wait for batch update completion."""
        self._batchUpdateCompletedEvent.wait()

    def batchUpdateGetErrors(self):
        """Return a list of errors from the last batch update."""
        return self._lastBatchUpdateErrors

    def _batchUpdateCompleted(self, reason):
        self._waitToStartScheduler = False
        if self._scheduler is not None:
            if self._scheduler.isAlive():
                self._scheduler.resume()
            else:
                self._scheduler.start()
        self._lastBatchUpdateErrors = self._batchScheduler.errors
        self._batchScheduler = None
        self.batchUpdateCompleted(reason)
        self._batchUpdateCompletedEvent.set()

    _SHOW_PROGRESS_ON_ONE_LINE = True
    def batchUpdateProgress(self, stage, obj):
        """Called when progress is made on a set of batch update requests.
        
            "stage" is a string defining the current processing stage.
            "obj" is some object relevant to a particular stage to describe
                is state of progress.
        
        Subclasses can override this for custom handling.
        """
        if self._batchProgressQuiet:
            return

        cache = self._batchProgressCache or {}
        if stage == "Preparing database": # a.k.a. removing db indices
            name, current, total = obj
            line = "%s: %s" % (stage, name)
        elif stage == "Restoring indices":
            name, current, total = obj
            line = "%s: %s" % (stage, name)
        elif stage == "Gathering files":
            line = "Gathering files: %s files" % obj
        elif stage == "Scanning": # 'obj' is a request object
            # Scanning: [#####        ] (1/12) ...tel\xsltcile.py ETA: 5 min
            remaining = self._batchScheduler.getNumFilesToProcess()+1
            total = cache.setdefault("total_files", remaining)
            # Progress meter section
            percent = float(total-remaining)/float(total)
            METER_WIDTH = 20
            meterTemplate = "%%-%ds" % METER_WIDTH
            meter = meterTemplate % ("#"*int(METER_WIDTH*percent))
            # Count and file section
            # 79: display width; 25: other stuff, e.g. "Scanning", "ETA: ..."
            file = obj.path
            COUNT_AND_FILE_WIDTH = 79 - 25 - METER_WIDTH
            count = "(%d/%d)" % (total-remaining+1, total)
            FILE_WIDTH = COUNT_AND_FILE_WIDTH - len(count) - 1
            if len(file) > FILE_WIDTH:
                file = "..."+file[-FILE_WIDTH+3:]
            countAndFile = "%s %s" % (count, file)
            # ETA section
            now = time.time()
            if "starttime" in cache:
                elapsedsecs = now - cache["starttime"]
                totalsecs = elapsedsecs / percent
                remainsecs = totalsecs - elapsedsecs
                if remainsecs < 60.0:
                    eta = "ETA: %d sec" % int(remainsecs)
                elif (remainsecs/60.0 < 60.0):
                    eta = "ETA: %d min" % int(remainsecs/60.0)
                else:
                    eta = "ETA: %d hr" % int(remainsecs/3600.0)
            else:
                cache["starttime"] = now
                eta = ""
            # Put sections together
            line_template = "Scanning: [%%s] %%-%ds %%s" % COUNT_AND_FILE_WIDTH
            line = line_template % (meter, countAndFile, eta)
        else:
            line = "%s: %s" % (stage, obj)
        if self._SHOW_PROGRESS_ON_ONE_LINE and "last_line" in cache:
            length = len(cache["last_line"])
            sys.stdout.write( "\b \b"*length )
        sys.stdout.write(line)
        if not self._SHOW_PROGRESS_ON_ONE_LINE:
            sys.stdout.write('\n')
        cache["last_line"] = line
        self._batchProgressCache = cache

    def batchUpdateCompleted(self, reason):
        """Called when a set of batch update requests are completed.
        
        Subclasses can override this for custom handling.
        """
        if self._batchProgressQuiet:
            return

        cache = self._batchProgressCache or {}
        if "starttime" in cache:
            totalsecs = time.time() - cache["starttime"]
            if totalsecs < 60.0:
                elapsed = "%d seconds" % int(totalsecs)
            elif (totalsecs/60.0 < 60.0):
                elapsed = "%d minute(s)" % int(totalsecs/60.0)
            else:
                elapsed = "%d hour(s)" % int(totalsecs/3600.0)
        else:
            elapsed = None
        errors = self._lastBatchUpdateErrors
        if self._SHOW_PROGRESS_ON_ONE_LINE and "last_line" in cache:
            length = len(cache["last_line"])
            sys.stdout.write( "\b \b"*length )
        if reason == "completed":
            sys.stdout.write("Batch update completed.")
        elif reason == "stopped":
            sys.stdout.write("Batch update was cancelled.")
        elif reason == "error":
            sys.stdout.write("Batch update errored out.")
        else:
            sys.stdout.write("Batch update completed (%s)." % reason)
        if elapsed: sys.stdout.write(" Running time: %s." % elapsed)
        if errors: sys.stdout.write(" There were %d errors:" % len(errors))
        sys.stdout.write("\n")
        if errors:
            print "\t"+"\n\t".join(errors)

    def batchUpdateError(self, errmsg):
        """Called when there is an error in batch update processing.
        
            "errmsg" is a short string describing the error.
        
        Note that all errors can be retrieved after a batch update is
        complete by calling batchUpdateGetErrors()

        Subclasses can override this for custom handling.
        """
        if self._batchProgressQuiet:
            return
        print "Batch update error: %s" % errmsg

    def _parseCITDLExpr(self, expr):
        """Parse a CITDL expression into <symbol ref> and <scope> parts."""
        #XXX This should really be in citdl.py.
        return expr.split(None, 1)

    def _getMembersPyEvalFallback(self, symbolRef):
        """See if the symbol reference matches one of the modules in the
        current process run-time. ...to ensure being no worse that previous
        AutoComplete implementations in Komodo.
        
        Returns None if nothing suitable was found.
        """
        try:
            namespace = sys.modules.copy()
            #NOTE: the PythonWin equivalent code also has:
            #      - namespace.update(__main__.__dict__); and
            #      - Adds the current context from its Interactive Shell.
            #      Do we want to add those?
            obj = eval(symbolRef, namespace)
        except:
            return None
        typesAndMembers = None
        if obj is not None:
            members = []
            try:
                members += dir(obj)
            except AttributeError:
                pass # object has no __dict__
            if hasattr(obj, "__class__"):
                members += _getPyClassMembers(obj.__class__)
            typesAndMembers = []
            for m in members:
                if m[0] == '_':
                    continue # drop pseudo-private and special attrs
                typeName = _getTypeNameForPyObj(getattr(obj, m))
                typesAndMembers.append((typeName, m))

            # The object may be a COM object with typelib support - lets see
            # if we can get its props. (contributed by Stefan Migowsky)
            try:
                # Get the automation attributes
                typesAndProps = [("function", p) for p in
                                 obj.__class__._prop_map_get_.keys()]
                # See if there is an write only property 
                # could be optimized
                for p in obj.__class__._prop_map_put_.keys():
                    tp = ("function", p)
                    if tp not in typesAndProps:
                        typesAndProps.append(tp)
                typesAndMembers += typesAndProps
            except AttributeError:
                pass
            
            if not typesAndMembers:
                typesAndMembers = None
        return typesAndMembers

    def _getMembersFallback(self, expr, content, line, language):
        """A dumb fallback mechanism for finding members.
        
            "expr" is the CITDL expression.
            "content" is the file content.
            "line" is the line in the content on which AutoComplete was
                triggered.
            "language" is the file's language.
        
        Basic algorithms from PythonWin. Returns a list of 2-tuples:
            (<codeintel-type-name>, <member-name>)
        of found members; the empty list if none found.
        """
        symbolRef = self._parseCITDLExpr(expr)[0]
        typesAndMembers = None
        if language == "Python" and '(' not in symbolRef:
            #XXX Technically eval'ing any "symbolRef" is potentially
            #    dangerous. __getattr__ code *will* be run. We specifically
            #    disallow eval'ing expressions with "()" because those are
            #    more likely to have a side-effect.
            typesAndMembers = self._getMembersPyEvalFallback(symbolRef)
        if typesAndMembers is None:
            # Heuristics a-la "complete word": The idea is to find other
            # usages of the current binding and assume, that it refers to the
            # same object (or at least, to an object of the same type)
            # Contributed to PythonWin by Vadim Chugunov [vadimch@yahoo.com]

            #NOTE: PythonWin limits the search to the current class using
            #      its class browser. Perhaps we should do that here with
            #      our scope information?

            if language == "Python":
                pattern = re.compile(r"\b%s(\.)(\w+)(\s*\()?" % re.escape(symbolRef))
            elif language == "Perl":
                pattern = re.compile(r"%s(->|::)(\w+)(\s*\()?" % re.escape(symbolRef))
            else:
                raise CodeIntelError("don't know how to make a suitable "
                                     "dumb-get-members pattern for %s"
                                     % language)
            try:
                hits = pattern.findall(content)
            except re.error:
                # With the added "re.escape()" (when this code was adapted
                # for use in the CodeIntel system) this should not raise, but
                # let's be paranoid.
                hits = []

            # If the word currently to the right of the dot is the only
            # such occurrence, it is probably a red-herring.
            lines = content.splitlines()
            herrings = re.findall(pattern, lines[line-1])
            for herring in herrings: # Can be multiple on this one line.
                if herring in hits:
                    hits.remove(herring)

            # Remove duplicates names.
            unique = {}
            for hit in hits:
                unique[hit[1]] = hit
            hits = unique.values()

            typesAndMembers = []
            if language == "Python":
                for trigger,name,call in hits:
                    # Only filter out __*__ members and not _* members because
                    # local private and protected attributes on a class should be
                    # included.
                    if name[:2]=="__" and name[-2:]=="__": continue
                    typeName = call and "function" or "variable"
                    typesAndMembers.append( (typeName, name) )
            elif language == "Perl":
                symbolPrefix = re.search(r'(.?)\b\w', symbolRef).group(1)
                for trigger,name,call in hits:
                    if call or trigger == "->" or symbolPrefix in ' &':
                        typeName = "function"
                    elif symbolPref:
                        typeName = "variable"
                    typesAndMembers.append( (typeName, name) )
            else:
                raise NotImplementedError("fallback member type determination "
                                          "not implemented for '%s'" % language)

        citdl.log.info("    fallback members: %r", typesAndMembers)
        return typesAndMembers or []

    def getMembers(self, language, path, line, expr,
                   scopeFileId=None, scopeTable=None, scopeId=None,
                   content=None):
        """Return the likely members for the given expression and scope.

            "language" is a language name (e.g., "Python", "Perl") of the
                working file. This can be relevant for some things, e.g.
                import semantics.
            "path" and "line" together indicate the scope from which to
                interpret the expression. Line is 1-based.
            "expr" is a CITDL expression identifying the object/symbol/type
                for which to return members.
            "scopeFileId", "scopeTable" and "scopeId" (optional) can be
                passed in to express the scope more accurately than "path"
                and "line" -- if, say, the caller is adjusting for recent
                edits not in the database.
            "content" (optional) is the file content. This may be used, if
                passed in, for fallback "dumb" completion handling.

        Returns a list of 2-tuples: (<type-name>, <member-name>)
        The <type-name> may be None, to indicate that the type is unknown.
        """
        #start = time.clock() #XXX
        context = citdl.Context(self, language, path, line)
        try:
            cx = self.getDatabaseConnection()
            cu = cx.cursor()
            self.acquireCIDBReadLock()
            try:
                members = context.getMembers(cu, expr, scopeFileId,
                                             scopeTable, scopeId)
            finally:
                self.releaseCIDBReadLock()
                cu.finalize()
                cx.close()
        except CodeIntelError:
            if content:
                members = self._getMembersFallback(expr, content,
                                                   line, language)
                if members:
                    return members
            raise
        if not members and content:
            #XXX Typically we should trust citdl.Context's evaluation
            #    but we don't. Try "self.cx." in Database.create, for
            #    example. In any case, as long as determining members
            #    is not perfect it is better to have false positives
            #    than nothing.
            members = self._getMembersFallback(expr, content,
                                               line, language)
        #print "XXX getMembers took: %f seconds" % (time.clock() - start)
        return members

    def _getCallTipsPyEvalFallback(self, symbolRef):
        """See if the symbol reference matches one of the modules in the
        current process run-time. ...to ensure being no worse that previous
        AutoComplete implementations in Komodo.
        
        Returns None if nothing suitable was found.
        """
        try:
            namespace = sys.modules.copy()
            #NOTE: the PythonWin equivalent code also has:
            #      - namespace.update(__main__.__dict__); and
            #      - Adds the current context from its Interactive Shell.
            #      Do we want to add those?
            obj = eval(symbolRef, namespace)
        except:
            return None
        calltips = None
        if obj is not None and type(obj) != types.ModuleType:
            calltip = _getPyObjCallTip(obj)
            if calltip: calltips = [calltip]
        return calltips

    def _getCallTipsFallback(self, expr, content, line, language):
        """A dumb fallback mechanism for finding calltips.
        
            "expr" is the CITDL expression.
            "content" is the file content.
            "line" is the line in the content on which a CallTip was
                triggered.
            "language" is the file's language.
        
        Basic algorithms from PythonWin. Returns a list of found CallTips
        (the empty list if none found).
        """
        symbolRef = self._parseCITDLExpr(expr)[0]
        calltips = None
        if language == "Python" and '(' not in symbolRef:
            #XXX Technically eval'ing any "symbolRef" is potentially
            #    dangerous. __getattr__ code *will* be run. We specifically
            #    disallow eval'ing expressions with "()" because those are
            #    more likely to have a side-effect.
            calltips = self._getCallTipsPyEvalFallback(symbolRef)
        citdl.log.info("    fallback calltips: %r", calltips)
        return calltips or []

    def getCallTips(self, language, path, line, expr,
                    scopeFileId=None, scopeTable=None, scopeId=None,
                    content=None):
        """Return the likely calltips for the given expression and scope.

            "language" is a language name (e.g., "Python", "Perl") of the
                working file. This can be relevant for some things, e.g.
                import semantics.
            "path" and "line" together indicate the scope from which to
                interpret the expression. Line is 1-based.
            "expr" is a CITDL expression identifying the object/symbol/type
                for which to return calltips.
            "scopeFileId", "scopeTable" and "scopeId" (optional) can be
                passed in to express the scope more accurately than "path"
                and "line" -- if, say, the caller is adjusting for recent
                edits not in the database.
            "content" (optional) is the file content. This may be used, if
                passed in, for fallback "dumb" completion handling.

        Typically there is just one calltip, but there _can_ be many: for
        overloaded methods in C++ for example, for ambiguous symbol references
        in dynamic languages. Generally returns an empty list if there is
        no appropriate calltip or if no calltips could be determined.
        """
        log.debug("Manager.getCallTips('%s', '%s', line=%r, expr=%r, ...)",
                  language, os.path.basename(path), line, expr)
        #start = time.time() #XXX
        context = citdl.Context(self, language, path, line)
        try:
            cx = self.getDatabaseConnection()
            cu = cx.cursor()
            self.acquireCIDBReadLock()
            try:
                calltips = context.getCallTips(cu, expr, scopeFileId,
                                               scopeTable, scopeId)
            finally:
                self.releaseCIDBReadLock()
                cu.finalize()
                cx.close()
        except CodeIntelError:
            if content:
                calltips = self._getCallTipsFallback(expr, content,
                                                     line, language)
                if calltips:
                    return calltips
            raise
        if not calltips and content:
            calltips = self._getCallTipsFallback(expr, content,
                                                 line, language)
        #print "XXX getCallTips took: %f seconds" % (time.time() - start)
        return calltips

    def getSubimports(self, language, module, cwd, explicit):
        """Return a list of available modules to import under the given
        module prefix.
        
            "language" is a language name (e.g., "Python", "Perl") of the
                working file. This can be relevant for some things, e.g.
                import semantics.
            "module" is the current module prefix.
            "cwd" is the current working directory. This is sometimes
                required to resolve relative import paths.
        """
        log.debug("Manager.getSubimports('%s', module=%r, cwd=%r, explicit=%r)",
                  language, module, cwd, explicit)
        importer = self.getImportHandler(language)
        return importer.findSubImportsOnDisk(module, cwd)

    def getScopeForFileAndLine(self, path, line):
        """This is a rawer form of getScopeInfoForFileAndLine.
        
        Returns (<type>, <file_id>, <id>, <row>) where
            <type> is "module" or "symbol";
            <file_id> is the CIDB file id to this scope belongs;
            <id> is a CIDB row id into that table; and
            <row> is the row data.
        Returns (None, 0, 0, None) if can't determine.
        """
        try:
            cx = self.getDatabaseConnection()
            cu = cx.cursor()
            self.acquireCIDBReadLock()
            try:
                scope = self.factory.createScopeForFileAndLine(cu, path, line)
            finally:
                self.releaseCIDBReadLock()
                cu.finalize()
                cx.close()
        except CodeIntelError, ex:
            return (None, 0, 0, None)
        else:
            return (scope.table, scope.file_id, scope.id, scope.row)

    def getScopeInfoForFileAndLine(self, path, line, language):
        """Return info for presenting the current scope is a UI.
        
            "path" is the full (absolute) path to the file.
            "line" is a 1-based line number.
            "language" is the language of the file.
            
        Returns: (<typeName>, <name>, <image url>, <description>)
        
        If the line is at the global level then the return value is
            ("module", <module name>, None, None)
        If the scope cannot be determined the return value is
            (None, None, None, None)
        """
        from codeintel.citdl import M_NAME, S_TYPE, S_ID, S_SCOPE, S_NAME, S_ATTRIBUTES
        try:
            cx = self.getDatabaseConnection()
            cu = cx.cursor()
            self.acquireCIDBReadLock()
            try:
                scope = self.factory.createScopeForFileAndLine(cu, path, line)
            finally:
                self.releaseCIDBReadLock()
                cu.finalize()
                cx.close()
        except CodeIntelError, ex:
            #XXX Commonly get the following. Need to look into it.
            #       WARN: codeintel.__init__: could not find file in CIDB: 'D:\trentm\as\Apps\Komodo-devel\Blackfile.py'
            #    Silence the warning for now though.
            #log.warn(str(ex))
            if 1: #XXX
                import traceback
                traceback.print_exception(*sys.exc_info())
            retval = (None, None, None, None)
        else:
            if scope.table == "module":
                retval = ("module", scope.row[M_NAME], None, None)
            else:
                row = scope.row
                import cb
                type, id, scope, name, attrStr =\
                    row[S_TYPE], row[S_ID], row[S_SCOPE], row[S_NAME], row[S_ATTRIBUTES]
                attributes = parseAttributes(attrStr)
                typeName = symbolType2Name(type)
                imageURL = cb.getImageURLForSymbol(typeName, attributes)
                desc = cb.getDescForSymbol(typeName, name, attributes,
                                           scope, language)
                retval = (typeName, name, imageURL, desc)
        return retval



#---- mainline/self-test

def test_PriorityQueue():
    import pythoncom
    
    class Foo(object):
        def __init__(self, name):
            self.id = str(pythoncom.CreateGuid())
            self.name = name
        def __repr__(self):
            return "<Foo '%s' (%s)>" % (self.name, self.id)
    queue = _PriorityQueue(0)
    queue.put( (2, time.time(), Foo("second")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("zzz")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("first")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("asdf")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("aqer")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("xzcv")) )
    time.sleep(0.1)
    t = Foo("third")
    queue.put( (3, time.time(), t) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("awer")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("opiu")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("asdf")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("qwer")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), t) )
    time.sleep(0.1)
    for i in range(5):
        print queue.get()
    queue.put( (0, time.time(), Foo("sentinel")) )
    while queue.queue:
        print queue.get()


def test_ScanRequestQueue():
    queue = _ScanRequestQueue(0)
    r1 = ScanRequest("foo.py", "Python", PRIORITY_IMMEDIATE)
    r1.id = 1
    queue.put( (r1.priority, time.time(), r1) )
    time.sleep(0.1)
    r2 = ScanRequest("bar.py", "Python", PRIORITY_IMMEDIATE)
    r2.id = 2
    queue.put( (r2.priority, time.time(), r2) )
    time.sleep(0.1)
    # This new request should replace the earlier (and higher priority) r1
    # and, hence, get onto the queue ahead of r2.
    r3 = ScanRequest("foo.py", "Python", PRIORITY_BACKGROUND)
    r3.id = 3
    queue.put( (r3.priority, time.time(), r3) )

    p, t, r = queue.get()
    assert r.id == r3.id, "Expected r3 to get first on the queue. That didn't happen."
    assert len(queue.queue) == 1, "Expected only one item (r2) to remain on queue."
    print "test_ScanRequestQueue... ok"

def test_StagingScanRequestQueue():
    sys.stdout.write("test_StagingScanRequestQueue...")
    queue = _StagingScanRequestQueue(0, 1.0)
    # With a staging delay of one second only the last ScanRequest for
    # "foo.py" should get remain on the queue, and then only after the second
    # has elapsed.
    for i in range(5):
        r = ScanRequest("foo.py", "Python", PRIORITY_IMMEDIATE)
        r.id = i
        #sys.stdout.write("[stage %d]" % i)
        queue.stage( (r.priority, time.time(), r) )
        time.sleep(0.1)

    # Cheat a little bit and poke into the queue's internal data structure
    # to test that none of these are on the queue yet.
    #sys.stdout.write("[assert non staged]")
    assert len(queue.queue) == 0, "Expected no requests to be on queue yet."

    # After at least one "staging delay" there should be just one request
    # on the queue and it should be that last one (id=4)
    #sys.stdout.write("[wait for staging]")
    time.sleep(2.0)
    assert (len(queue.queue) == 1,
            "Expected only one request for foo.py to be on the queue by now: "
            "len(queue.queue)==%d" % len(queue.queue))
    #sys.stdout.write("[get item]")
    p, t, r = queue.get()
    assert r.id==4 and r.path=="foo.py", "Unexpected request on the queue: id=%d, path=%r"

    queue.finalize()
    print "ok"

def _play():
    print "_play..."
    dbpath = os.path.abspath("codeintel.db")
    if os.path.exists(dbpath):
        # Start with a fresh database everytime.
        os.remove(dbpath)

    class MyManager(Manager):
        def requestCompleted(self, request):
            print "completed scan of '%s' (%s)"\
                  % (request.path, request.language)

    m = MyManager()
    m.initialize()
    try:
        hiPy = os.path.join("play", "hi.py")
        files = [
            (os.path.join("play", "raven.py"), "Python"),
            (os.path.join("play", "assign.py"), "Python"),
            (hiPy, "Python"),
            (os.path.join("play", "sample.py"), "Python"),
            (os.path.join("test", "scan_inputs", "php5_sample.php"), "PHP"),
            (os.path.join("test", "test_eval_citdl", "hi.py"), "Python"),
        ]
        for file, lang in files:
            print "request scan of '%s' (%s)" % (file, lang)
            r = ScanRequest(file, lang, PRIORITY_OPEN, force=1)
            m.addRequest(r)

        ## Cheat and test pause/resume of the Scheduler
        #print "_play: pausing Scheduler..."
        #m._scheduler.pause()
        #time.sleep(6)
        #print "_play: ...resuming Scheduler"
        #m._scheduler.resume()

        print "_play: start batch update..."
        #m.batchUpdateRequest("language", r"C:\Python22\python.exe",
        #                     "Python")
        #m.batchUpdateRequest("directory", r"D:\applications\Python-2.2\Lib",
        #                     "Python")
        m.batchUpdateRequest("directory", r"D:\trentm\as\Apps\Komodo-devel\src\codeintel",
                             "*")
        m.batchUpdateStart()
        if 1:
            m.batchUpdateWait()
        else:
            print "_play: ...cancelling batch update in 25 seconds..."
            time.sleep(25)
            m.batchUpdateStop()
        print "_play: ...done batch update"

        #pythonCIX = os.path.join("codeintel", "python.cix")
        #print "_play: import '%s'..." % pythonCIX
        #m.batchUpdateRequest("cix", pythonCIX)
        #m.batchUpdateStart()
        #m.batchUpdateWait()
        #errors = m.batchUpdateGetErrors()
        #if errors:
        #    print "Errors importing '%s':\n\t%s" % (pythonCIX, "\n\t".join(errors))
        #print "_play: ...done importing '%s'" % pythonCIX

        ## Request the scan of hi.py again, it should get dropped based
        ## on mtime.
        #time.sleep(1)
        #print "request scan of '%s' (Python) again" % hiPy
        #r = ScanRequest(hiPy, "Python", PRIORITY_OPEN)
        #m.addRequest(r)

        #time.sleep(1)
        #fin = open(hiPy, 'r')
        #content = fin.read()
        #fin.close()
        #content += "\n\ndef poke(): pass # added this method\n"
        #r = ScanRequest(hiPy, "Python", PRIORITY_OPEN, content=content)
        #m.addRequest(r)
    
        while m._scheduler and m._scheduler.getNumRequests(): # let scheduler finish its scans
            time.sleep(1)
        time.sleep(5)
    
        ## Test that looking in the built-in scope works.
        #expr = "__version__"; path = hiPy; line = 5
        #print "get members for '%s' at %s:%d (to test built-in scope)"\
        #      % (expr, path, line)
        #members = m.getMembers("Python", os.path.abspath(path), line, expr)
        #print "members are:", members

        ## Dump the Code Browser
        #cx = m.getDatabaseConnection()
        #cu = cx.cursor()
        #print "\n\n--------- Code Browser ---------"
        #import cb
        #root = cb.CBRootNode()
        #for file, lang in files:
        #    root.addPath(cu, os.path.abspath(file))
        #root.pprint()
        #print "-----------------------------------"

        ## Try out some of the info retrieval routines.
        #ravenPy = os.path.abspath(os.path.join("play", "raven.py"))
        #print
        #for line in (1, 12, 14, 17, 18, 19, 33):
        #    retval = m.getScopeInfoForFileAndLine(ravenPy, line)
        #    print "getScopeInfoForFileAndLine('.../raven.py', line=%r): %r"\
        #          % (line, retval)
        #print
        #triggerPoints = [
        #    (31, "Raven"),
        #    (29, "raven"),
        #    (15, "quoth"),
        #    (16, "quoth"),
        #    (30, "random"),
        #    (30, "hi"),
        #]
        #for line, citdl in triggerPoints:
        #    try:
        #        retval = m.getMembers("Python", ravenPy, line, citdl)
        #    except CodeIntelError, ex:
        #        retval = "error: %s" % ex
        #    print "getMembers('.../raven.py', line=%r, citdl=%r): %s\n"\
        #          % (line, citdl, retval)
        #print
    finally:
        m.finalize()
        print "done"

def _test_bug_32305():
    print "_test_bug_32305..."
    REGEN = 1
    dbpath = os.path.abspath("codeintel.db")
    if REGEN:
        if os.path.exists(dbpath):
            # Start with a fresh database everytime.
            os.remove(dbpath)

    m = Manager()
    m.initialize()
    try:
        if REGEN:
            pythonCIX = os.path.join("codeintel", "python.cix")
            print "_play: import '%s'..." % pythonCIX
            m.batchUpdateRequest("cix", pythonCIX)
            m.batchUpdateStart()
            m.batchUpdateWait()
            errors = m.batchUpdateGetErrors()
            if errors:
                print "Errors importing '%s':\n\t%s" % (pythonCIX, "\n\t".join(errors))
            print "_play: ...done importing '%s'" % pythonCIX
    
            files = [r"C:\Python22\Lib\os.py"]
            for file in files:
                print "request scan of '%s' (Python)" % file
                r = ScanRequest(file, "Python", PRIORITY_OPEN)
                m.addRequest(r)
    
            while m._scheduler and m._scheduler.getNumRequests(): # let scheduler finish its scans
                time.sleep(1)
            time.sleep(5)
    

        # Test that looking in the 'os' module works as expected.
        fooPy = "foo.py"
        tests = [
            #{"content": "import os, _winreg\n\n",
            # "exprs": ["os", "os.path"]},
            #{"content": "from os import path",
            # "exprs": ["path"]},
            #{"content": "from os import *\n\n",
            # "exprs": ["path"]},
            #{"content": "from os.path import *\n\n",
            # "exprs": ["abspath"]},
            {"content": "import re\n\n",
             "exprs": ["re"]},
        ]
        
        for test in tests:
            print "request scan of '%s' (Python)" % fooPy
            r = ScanRequest(fooPy, "Python", PRIORITY_OPEN,
                            content=test["content"], force=1)
            m.addRequest(r)
            time.sleep(1)

            for expr in test["exprs"]:
                path = fooPy; line = 3
                print "\nget members for '%s' at %s:%d (to test built-in scope)"\
                      % (expr, path, line)
                try:
                    members = m.getMembers("Python", os.path.abspath(path), line, expr)
                    members.sort()
                    print "\nmembers are:", members
                except CodeIntelError, ex:
                    log.error(str(ex))
    finally:
        m.finalize()
        print "done"

if __name__ == "__main__":
    __file__ = os.path.abspath(sys.argv[0])
    logging.basicConfig()
    #test_PriorityQueue()
    #test_ScanRequestQueue()
    #test_StagingScanRequestQueue()
    _test_bug_32305()
    #_play()

