-- Copyright (c) 2004-2006 ActiveState Software Inc.
-- See the file LICENSE.txt for licensing information.
--
-- SQL commands to create the tables for a Code Intelligence Database (CIDB)
--
-- CIDB version 1.0
-- http://specs.activestate.com/komodo30/func/code_intelligence.html#database-schema
--
-- Notes:
-- - c.f. http://www.hwaci.com/sw/sqlite/datatypes.html
--   regarding INTEGER PRIMARY KEY

-- Meta information about the database, such as version, platform, host.
CREATE TABLE meta (
    name TEXT,
    value TEXT
);


-- One entry for every language for which there is a file in the file table.
CREATE TABLE language (
    id INTEGER PRIMARY KEY,
    name TEXT
);


-- XXX It might make sense to add a 'basename' value to optimize some queries.
CREATE TABLE file (
    id INTEGER PRIMARY KEY,
    path TEXT,
    compare_path TEXT,
    language_id INTEGER
);
CREATE INDEX file_compare_path_index ON file (compare_path);
CREATE INDEX file_language_id_index ON file (language_id);
CREATE INDEX file_compare_path_and_language_id_index ON file (compare_path, language_id);


-- There is one scan row for every file/generator combination.
-- If the last scan for a file/generator failed then "error" is a string
-- describing why. However, even if the last scan failed there may still be
-- scan data: previous successful scan results may intentionally not have
-- been overwritten. There is scan data if there is an md5 value.
CREATE TABLE scan (
    id INTEGER PRIMARY KEY,
    file_id INTEGER,
    -- Name of the generator that did this scan.
    generator TEXT,
    -- Last modified time of the file in seconds since the "epoch" (this is
    -- generally what the standard time() function, in any language, returns).
    mtime INTEGER,
    -- MD5 hex digest of the file content for the last successful scan.
    md5 TEXT,
    -- Error message from the last scan of this file by this generator,
    -- if there was an error.
    error TEXT
);


-- In most scripting languages the file _is_ the module. This isn't always
-- true though. For example, Perl can switch between modules in one file
-- with the 'package' statement. If this system grows to compiled languages
-- then a distinction between files and modules (e.g. C++ translation units)
-- might be useful.
CREATE TABLE module (
    id INTEGER PRIMARY KEY,
    file_id INTEGER,
    scan_id INTEGER,
    name TEXT,
    line INTEGER,
    attributes TEXT,
    doc TEXT
);
CREATE INDEX module_file_id_index ON module (file_id);
CREATE INDEX module_scan_id_index ON module (scan_id);


CREATE TABLE symbol (
    id INTEGER PRIMARY KEY,
    file_id INTEGER,
    scan_id INTEGER,
    module_id INTEGER,
    -- The 'id' of the parent symbol. If this symbol has global scope
    -- (within the module) then it is set to 0.
    scope_id INTEGER,
    -- text representation of the scope
    scope TEXT,
    name TEXT,
    -- Optional symbol signature. Just wanted for 'function' (and maybe
    -- 'class') symbols for use in calltip generation.
    -- Deprecation Note:
    -- Pre-CIDB-0.15 signatures were targeted for hover tips on symbols in
    -- something like Komodo's Code Browser. This is no longer the target.
    -- as a result, signatures on other symbol types are useless.
    signature TEXT,
    -- 1-based position of symbol declaration. These are both optional
    -- (to allow CIX data for binary modules) but strongly recommended for
    -- scanned text files.
    line INTEGER,
    lineend INTEGER,
    -- Type of symbol being declared. It must be one of the codeintel.ST_*
    -- constants (actually defined in "codeintel/common.py").
    type INTEGER,
    attributes TEXT,
    doc TEXT
);
CREATE INDEX symbol_file_id_index ON symbol (file_id);
CREATE INDEX symbol_scan_id_index ON symbol (scan_id);
CREATE INDEX symbol_module_id_index ON symbol (module_id);
CREATE INDEX symbol_scope_id_index ON symbol (scope_id);
-- Maybe index on "name" as well, and perhaps combinations of the above.


-- Base classes and interfaces for symbols in the symbol table.
CREATE TABLE base (
    id INTEGER PRIMARY KEY,
    scan_id INTEGER,
    symbol_id INTEGER,
    name TEXT,
    type INTEGER -- one of the BT_CLASSREF or BT_INTERFACEREF constants
);
CREATE INDEX base_scan_id_index ON base (scan_id);
CREATE INDEX base_symbol_id_index ON base (symbol_id);


-- Each variable (also includes symbols of type ST_ARGUMENT) or base
-- class/interface can have one or more possible types (more than one for
-- dynamic languages). A "type" is represented as a CITDL (Code Intelligence
-- Type Determination Language) expression. See
--      http://specs.activestate.com/Komodo_3.0/func/code_intelligence.html#CITDL
CREATE TABLE type (
    id INTEGER PRIMARY KEY,
    scan_id INTEGER,
    -- One of symbol_id or base_id is defined, indicate to which row in the
    -- 'symbol' or 'base' table this type applies.
    symbol_id INTEGER,
    base_id INTEGER,
    citdl TEXT,
    -- An optional numeric score that _may_ be used to indicate higher
    -- probabilities of some type guesses over others.
    score INTEGER
);
CREATE INDEX type_scan_id_index ON type (scan_id);
CREATE INDEX type_symbol_id_index ON type (symbol_id);
CREATE INDEX type_base_id_index ON type (base_id);


CREATE TABLE import (
    id INTEGER PRIMARY KEY,
    file_id INTEGER,
    scan_id INTEGER,
    module_id INTEGER,
    -- The scope in which the import occurs. A value of 0 indicates global
    -- scope in the module. Otherwise it indicates an 'id' in the
    -- 'symbol' table.
    scope_id INTEGER,
    line INTEGER,
    -- Name of the module (or file) being imported.
    module TEXT,
    -- The name of a particular symbol being import from the module, if any.
    -- If this is blank then the meaning of the import is language-dependent
    -- (import semantics vary widely).
    symbol TEXT,
    -- An optional alias specified for the 'name', if any, or the 'module',
    -- if not. E.g.:
    --      import xml as XML
    --      from time import time as clock
    alias TEXT
);
CREATE INDEX import_scan_id_index ON import (scan_id);
CREATE INDEX import_scope_id_index ON import (scope_id)

