#!/usr/bin/env python
# Copyright (c) 2005-2006 ActiveState Software Inc.
#
# Contributors:
#   Eric Promislow (EricP@ActiveState.com)

"""
    rubycile - a Code Intelligence Language Engine for the Ruby language

    Module Usage:
        from rubycile import scan
        mtime = os.stat("foo.rb")[stat.ST_MTIME]
        content = open("foo.rb", "r").read()
        scan(content, "foo.rb", mtime=mtime)
    
    Command-line Usage:
        rubycile.py [<options>...] [<Ruby files>...]

    Options:
        -h, --help          dump this help and exit
        -V, --version       dump this script's version and exit
        -v, --verbose       verbose output, use twice for more verbose output
        -f, --filename <path>   specify the filename of the file content
                            passed in on stdin, this is used for the "path"
                            attribute of the emitted <file> tag.
        --md5=<string>      md5 hash for the input
        --mtime=<secs>      modification time for output info, in #secs since
                            1/1/70.
        -c, --clock         print timing info for scans (CIX is not printed)

    One or more Ruby files can be specified as arguments or content can be
    passed in on stdin. A directory can also be specified, in which case
    all .rbfiles in that directory are scanned.

    This is a Language Engine for the Code Intelligence (codeintel) system.
    Code Intelligence XML format. See:
        http://specs.activestate.com/Komodo_3.0/func/code_intelligence.html
    
    The command-line interface will return non-zero iff the scan failed.
"""

import os
import os.path
import sys
import getopt
import md5
import re
import logging
import glob
import time
import stat

from SilverCity import ScintillaConstants

from codeintel import ruby_lexer, ruby_parser, util
from codeintel.parseutil import getAttrStr, xmlencode, cdataescape


#---- exceptions

class RubyCILEError(Exception):
    pass


#---- global data

_version_ = (0, 1, 0)
log = logging.getLogger("rubycile")
#log.setLevel(logging.DEBUG)

_gClockIt = 0   # if true then we are gathering timing data
_gClock = None  # if gathering timing data this is set to time retrieval fn
_gStartTime = None   # start time of current file being scanned


#---- internal support
# This code has intimate knowledge of the code objects defined in
# ruby_parser.py

def get_common_attrs(node):
    attrs = {}
    attrs = {"name": node.name}
    try: attrs["line"] = node.line_num;
    except: pass
    try: attrs["lineend"] = node.lineend;
    except: pass
    try:
        if node.is_classmethod:
            attrs["attributes"] = "__classmethod__"
        elif node.is_constructor:
            attrs["attributes"] = "__ctor__"
        else:
            attrs["attributes"] = node.attributes;
    except: pass
        
    return attrs

def dump_docstring(node, cix, indent_level):
    if len(node.doc_lines) >= 1:
        summarylines = util.parseDocSummary(node.doc_lines)
        if len(summarylines) > 0:
            str2 = "\n".join(summarylines)
            doc = cdataescape(xmlencode(str2))
            cix.append("%s<doc><![CDATA[%s]]></doc>\n" % (" " * indent_level, doc))

def dump_signature(node, cix, indent_level):
    try:
        signature = cdataescape(xmlencode(node.signature))
        if len(signature) > 0:
            cix.append("%s<signature><![CDATA[%s]]></signature>\n" % (" " * (indent_level), signature))
    except:
        pass

def dump_imports(node, cix, indent_level):
    try:
        for imp in node.imports:
            cix.append("%s<import module=\"%s\" line=\"%d\" symbol=\"*\" />\n" % \
                       (" " * indent_level, imp.name, imp.line_num))
        # Nothing to do with the include statements yet, that bring
        # all the names in a namespace into the calling namespace.
    except:
        pass

def sort_by_lines(adict):
    intermed = [(adict[k].line_num, adict[k].type, k) for k in adict.keys()]
    intermed.sort()
    return intermed

def dump_var(cix, attrs, var_type, indent_level):
    ind_str = " " * indent_level
    if var_type:
        cix.append("%s<variable%s >\n" % (ind_str, getAttrStr(attrs)))
        cix.append("%s  <type%s />\n" % (ind_str, getAttrStr({"type" : var_type})))
        cix.append("%s</variable>\n" % (ind_str,))
    else:
        cix.append("%s<variable%s />\n" % (ind_str, getAttrStr(attrs)))

def dump_globals(node, cix, indent_level):
    try:
        for line_no, var_type, var_name in sort_by_lines(node.global_vars):
            attrs = {"name": var_name, "line": line_no}
            dump_var(cix, attrs, var_type, indent_level)
    except Exception, ex:
        log.error(str(ex))

def dump_vars(node, cix, indent_level):
    # Locals
    for line_no, var_type, var_name in sort_by_lines(node.local_vars):
        attrs = {"name": var_name, "line":line_no}
        if var_name[0].islower() or var_name[0] == "_":
            attrs["attributes"] = "__local__"
        dump_var(cix, attrs, var_type, indent_level)
    # Class vars
    for line_no, var_type, var_name in sort_by_lines(node.class_vars):
        attrs = {"name": var_name, "line":line_no,
                 "attributes" : "__local__" }
        dump_var(cix, attrs, var_type, indent_level)
    # Instance vars
    for line_no, var_type, var_name in sort_by_lines(node.instance_vars):
        attrs = {"name": var_name, "line":line_no,
                 "attributes" : "__instancevar__ __local__"}
        dump_var(cix, attrs, var_type, indent_level)        

def module_cix(node, cix, indent_level):
    attrs = get_common_attrs(node)
    attrs['attributes'] = '__namespace__'
    cix.append("%s<class%s>\n" % (" " * indent_level, getAttrStr(attrs)))
    dump_docstring(node, cix, indent_level + 2)
    dump_signature(node, cix, indent_level + 2)
    dump_imports(node, cix, indent_level + 2)
    dump_vars(node, cix, indent_level + 2)
    #XXX: What else?
    visit_children(node, cix, indent_level + 2)
    cix.append("%s</class>\n" % (" " * indent_level,))

def class_cix(node, cix, indent_level):
    attrs = get_common_attrs(node)
    cix.append("%s<class%s>\n" % (" " * indent_level, getAttrStr(attrs)))
    dump_docstring(node, cix, indent_level + 2)
    dump_signature(node, cix, indent_level + 2)
    dump_imports(node, cix, indent_level + 2)
    dump_vars(node, cix, indent_level + 2)

    ind_str = " " * (indent_level + 2)
    # Put the includes in before the parents
    # Array of Name_LineNum
    for ref in node.includes:
        cix.append("%s<classref line=\"%d\" name=\"%s\" attributes=\"__mixin__\" >\n" % (ind_str, ref.line_num, ref.name))
        cix.append("  %s<type type=\"%s\" />\n" % (ind_str, ref.name))
        cix.append("%s</classref>\n" % (ind_str,))
        
    # Classrefs
    classrefs = node.classrefs
    if len(classrefs) > 0:
        for ref in classrefs:
            classref_type = ref[2]
            if classref_type:
                attrs = {}
                attrs["name"] = ref[0]
                attrs["line"] = ref[1]
                cix.append("%s<classref%s >\n" % (ind_str, getAttrStr(attrs)))
                cix.append("  %s<type type=\"%s\" />\n" % (ind_str, classref_type))
                cix.append("%s</classref>\n" % (ind_str,))
            else:
                cix.append("%s<classref line=\"%s\" name=\"%s\" >\n" % (ind_str, ref[1], ref[0]))
                cix.append("  %s<type type=\"%s\" />\n" % (ind_str, ref[0]))
                cix.append("%s</classref>\n" % (ind_str,))
    else:
        cix.append("%s<classref line=\"%d\" name=\"Object\" >\n" % (ind_str, node.line_num))
        cix.append("  %s<type type=\"Object\" />\n" % (ind_str,))
        cix.append("%s</classref>\n" % (ind_str,))

    #XXX: Get classrefs, doc, symbols(?)
    visit_children(node, cix, indent_level + 2)
    cix.append("%s</class>\n" % (" " * indent_level,))

def dump_arguments(node, cix, indent_level):
    for c in node.args:
        attrs = get_common_attrs(c)
        attrs['name'] = c.get_full_name()
        cix.append("%s<argument%s />\n" % (" " * indent_level, getAttrStr(attrs)))

def method_cix(node, cix, indent_level):
    attrs = get_common_attrs(node)
    cix.append("%s<function%s>\n" % (" " * indent_level, getAttrStr(attrs)))
    dump_docstring(node, cix, indent_level + 2)
    dump_signature(node, cix, indent_level + 2)
    #XXX: Get classrefs, doc, symbols(?)
    dump_arguments(node, cix, indent_level + 2)
    dump_imports(node, cix, indent_level + 2)
    dump_vars(node, cix, indent_level + 2)
    visit_children(node, cix, indent_level + 2)
    cix.append("%s</function>\n" % (" " * indent_level,))

def produce_cix(parse_tree, filename, md5sum, mtime):
    cix = [u'<?xml version="1.0" encoding="UTF-8"?>\n<codeintel version="0.1">\n']
    fileAttrs = {"language": "Ruby",
                 "generator": "Ruby",
                 "path": filename,
                 "md5": md5sum,
                 "mtime": mtime}
    cix.append("  <file%s>\n" % getAttrStr(fileAttrs))
    cix.append("    <module name=\"%s\">\n" % (os.path.basename(filename).split('.')[0], ))
    indent_level=6
    dump_docstring(parse_tree, cix, indent_level)
    dump_imports(parse_tree, cix, indent_level)
    dump_globals(parse_tree, cix, indent_level)
    dump_vars(parse_tree, cix, indent_level)
    visit_children(parse_tree, cix, indent_level)
    cix.append("    </module>\n")
    cix.append("  </file>\n</codeintel>\n")
    return u''.join(cix)

cb_hash = {"Module" : module_cix,
           "Class" : class_cix,
           "Method" : method_cix}

def visit_children(node, cix, indent_level):
    for c in node.children:
        cb_hash[c.class_name](c, cix, indent_level)


def scan(content, filename, md5sum=None, mtime=None):
    """Scan the given Ruby content and return Code Intelligence data
    conforming the the Code Intelligence XML format.
    
        "content" is the Ruby content to scan
        "filename" is the source of the Ruby content (used in the
            generated output).
        "md5sum" (optional) if the MD5 hexdigest has already been calculated
            for the content, it can be passed in here. Otherwise this
            is calculated.
        "mtime" (optional) is a modified time for the file (in seconds since
            the "epoch"). If it is not specified the _current_ time is used.
            Note that the default is not to stat() the file and use that
            because the given content might not reflect the saved file state.
        XXX Add an optional 'eoltype' so that it need not be
            re-calculated if already known.
    
    This can raise RubyCILEError if there was an error processing.
    """
    log.info("scan '%s'", filename)
    content = content.expandtabs(8)
    tokenizer = ruby_lexer.Lexer(content)
    parser = ruby_parser.Parser(tokenizer)
    if 1:
        parse_tree = parser.parse()
        if mtime is None:
            actual_mtime = int(time.time())
        else:
            actual_mtime = mtime
        if md5sum is None:
            actual_md5 = md5.new(content).hexdigest()
        else:
            actual_md5 = md5sum
        return produce_cix(parse_tree, filename, actual_md5, actual_mtime)
    if 0: #except Exception, e:
        print "Error: " + e
        sys.exit(1)   


#---- mainline

def main(argv):
    logging.basicConfig()
    # Parse options.
    try:
        opts, args = getopt.getopt(argv[1:], "Vvhf:c",
            ["version", "verbose", "help", "filename=", "md5=", "mtime=",
             "clock"])
    except getopt.GetoptError, ex:
        log.error(str(ex))
        log.error("Try `rubycile --help'.")
        return 1
    numVerboses = 0
    stdinFilename = None
    md5sum = None
    mtime = None
    global _gClockIt
    for opt, optarg in opts:
        if opt in ("-h", "--help"):
            sys.stdout.write(__doc__)
            return
        elif opt in ("-V", "--version"):
            ver = '.'.join([str(part) for part in _version_])
            print "rubycile %s" % ver
            return
        elif opt in ("-v", "--verbose"):
            numVerboses += 1
            if numVerboses == 1:
                log.setLevel(logging.INFO)
            else:
                log.setLevel(logging.DEBUG)
        elif opt in ("-f", "--filename"):
            stdinFilename = optarg
        elif opt in ("--md5",):
            md5sum = optarg
        elif opt in ("--mtime",):
            mtime = optarg
        elif opt in ("-c", "--clock"):
            _gClockIt = 1
            global _gClock
            if sys.platform.startswith("win"):
                _gClock = time.clock
            else:
                _gClock = time.time

    if len(args) == 0:
        contentOnStdin = 1
        filenames = [stdinFilename or "<stdin>"]
    else:
        contentOnStdin = 0
        paths = []
        for arg in args:
            paths += glob.glob(arg)
        filenames = []
        for path in paths:
            if os.path.isfile(path):
                filenames.append(path)
            elif os.path.isdir(path):
                rbfiles = [os.path.join(path, n) for n in os.listdir(path)
                           if os.path.splitext(n)[1] == ".rb"]
                rbfiles = [f for f in rbfiles if os.path.isfile(f)]
                filenames += rbfiles

    try:
        for filename in filenames:
            if contentOnStdin:
                log.debug("reading content from stdin")
                content = sys.stdin.read()
                log.debug("finished reading content from stdin")
                if mtime is None:
                    mtime = int(time.time())
            else:
                if mtime is None:
                    mtime = int(os.stat(filename)[stat.ST_MTIME])
                content = open(filename, 'r').read()

            if _gClockIt:
                sys.stdout.write("scanning '%s'..." % filename)
                global _gStartTime
                _gStartTime = _gClock()
            data = scan(content, filename, mtime, md5sum)
            if _gClockIt:
                sys.stdout.write(" %.3fs\n" % (_gClock()-_gStartTime))
            elif data:
                sys.stdout.write(data)
    except KeyboardInterrupt:
        log.debug("user abort")
        return 1
    except Exception, ex:
        log.error(str(ex))
        if log.isEnabledFor(logging.DEBUG):
            print
            import traceback
            traceback.print_exception(*sys.exc_info())
        return 1

if __name__ == "__main__":
    sys.exit(main(sys.argv))

