import xml.dom.minidom as minidom
import logging

log = logging.getLogger("koXMLStateMachineReader")

class BasicNamespace:
    def __init__(self):
        self.elements = {}

    def element_info(self, element_name):
        return self.elements[element_name]

    def possible_children(self, element_name):
        return self.elements.keys()

    def possible_attributes(self, element_name):
        return []

    def possible_attribute_values(self, element_name, attribute_name):
        return []
    
    def all_element_types(self):
        return self.elements.keys()

class state_machine_info(BasicNamespace):
    def __init__(self, filename):
        dom = minidom.parse(filename)
        handler().handle_schema(dom.documentElement)

        literal_state = child_state("#LITERAL", "ELEMENTSTATE_UNCHANGED", "ELEMENTSTATE_LITERAL")
        all_child_states.append(literal_state)

        elements={}
        for a_child_state in all_child_states:
            if not elements.has_key(a_child_state.element_name):
                el = Element()
                elements[a_child_state.element_name] = el
                for child in getChildren(all_element_states[a_child_state.state]):
                    el.possible_children[child] = None
                if all_element_states[a_child_state.state].literal_state!="ELEMENTSTATE_INVALID":
                    el.possible_children["#LITERAL"] = None
            
        items = elements.items()
        items.sort()

        for attr in all_attributes:
            for elementname in attr.attribute_set.owners:
                elements[elementname].attributes[attr.name]=attr
        self.elements = elements

    def element_info(self, element_name):
        return self.elements.get(element_name, None)

    def possible_children(self, element_name):
        return self.elements.get(element_name, Element()).possible_children.keys()

    def possible_attributes(self, element_name):
        return self.elements.get(element_name, Element()).attributes.keys()

    def possible_attribute_values(self, element_name, attribute_name):
        el = self.elements.get(element_name, Element())
        attr = el.attributes.get(attribute_name, None)
        if attr:
            return attr.suggestions
        else:
            return []

    def all_element_types(self):
        return self.elements.keys()

    def dump(self, stream):
        stream.write("<elements>\n")
        elements = self.elements.items()
        elements.sort()
        for name, el in elements:
            children = el.possible_children.keys()
            children.sort()
            stream.write('<element name="%s" children="%s">\n' % (
                                name, " ".join(children)))
            for attribute in el.attributes.values():
                text = '  <attribute name="%s" required="%s" suggestions="%s"/>\n' 
                text = text % ( attribute.name, attribute.required, 
                                    " ".join(attribute.suggestions))
                stream.write(text)
            stream.write("</element>\n")
        stream.write("</elements>\n")



all_element_states = {}
class element_state:
    def __init__(self, identifier, endable):
        self.identifier = identifier
        self.endable = endable
        self.attribute_sets=[]
        self.children_states = "#CHILDREN_NONE"
        all_element_states[identifier]=self

    def __repr__(self):
        return "<element_state %s>" % self.identifier

all_children_states = {}
class children_states:
    def __init__(self, identifier):
        self.identifier = identifier
        self.child_states = []
        all_children_states[identifier]=self

all_child_states = []
class child_state:
    def __init__(self, elementname, parentstate, state):
        self.element_name = elementname
        self.parent_state = parentstate
        self.state = state
        all_child_states.append(self)

all_attribute_sets = {}
class attribute_set:
    def __init__(self, identifier):
        self.identifier = identifier
        self.attrs = []
        self.owners = []
        all_attribute_sets[identifier] = self

all_attributes = []
class attribute:
    def __init__(self, name, required, set):
        self.name = name
        self.suggestions = []
        self.required = required
        self.attribute_set = set
        all_attributes.append(self)

    def __repr__(self):
        return "<attribute "+self.name+"=["+"|".join(self.suggestions)+"]>"

class handler:
    def processChild(self, node, *args):
        name = node.nodeName.replace("-", "_").replace("#", "_")
        handler = getattr(self, "handle_"+ name, None)
        if handler:
            handler(node, *args)
        else:
            self.handle__unknown(node, *args)
    
    def processChildren(self, node, types=["*"]):
        for child in node.childNodes:
            if child.nodeName in types or "*" in types:
                self.processChild(child)

    def handle_schema(self, node):
        self.processChildren(node)
     
    def handle_element(self, node):
        el = ElementType(node.getAttribute("name"))
        self.processChildren(node, ["complexType"])

    def handle_state_machine(self, node):
        self.procesChildren(node)

    def handle_element_state(self, node):        
        self.element_state = element_state(node.getAttribute("identifier"),
                                           node.getAttribute("endable"))
        self.processChildren(node)

    def handle_attribute_set(self, node):
        if node.parentNode.nodeName=="element-state":
            self.element_state.attribute_sets.append(node.getAttribute("ref"))
        else:
            self.attribute_set = attribute_set(node.getAttribute("identifier"))
            self.processChildren(node)

    def handle_belongs_to(self, node):
        self.attribute_set.owners.append(getChildText(node))

    def handle_attribute(self, node):
        self.processChildren(node)

    def handle_name(self, node):
        if node.parentNode.nodeName == "attribute":
            self.attr = attribute(getChildText(node), 
               node.parentNode.getAttribute("required")=="TRUE", self.attribute_set)
            self.attribute_set.attrs.append(self.attr)

    def handle_value(self, node): pass            

    def handle_suggest(self, node):
        self.attr.suggestions.append(getChildText(node))

    def handle_literal_state(self, node):
        self.element_state.literal_state = node.getAttribute("ref")

    def handle_children_states(self, node):
        if node.parentNode.nodeName=="element-state":
            assert self.element_state.children_states=="#CHILDREN_NONE"
            self.element_state.children_states = node.getAttribute("ref")
        else:
            self.children_states = children_states(
                node.getAttribute("identifier"))
            self.processChildren(node)

    def handle_child_state(self, node):
        self.children_states.child_states.append(
            child_state(node.getAttribute("element-name"),
              node.getAttribute("parent-state"),
              node.getAttribute("state")))

    def handle__text(self, node):
        if node.nodeValue.strip():
            pass
        #log.debug("Unhandled text: %s", node.nodeValue)

    def handle__unknown(self, node, *args):
        log.debug("unknown type: %s", node.nodeName)

class Element:
    def __init__(self):
        self.possible_children = {}
        self.attributes = {}

def getChildren(element_state):
    rc = []
    for state in all_children_states[
            element_state.children_states].child_states:
        rc.append(state.element_name)
        if state.parent_state!="ELEMENTSTATE_UNCHANGED":
            rc.extend(getChildren(all_element_states[state.parent_state]))
    return rc
        
def getChildText(node):
    return "".join([child.nodeValue for child in node.childNodes 
                    if child.nodeType == node.TEXT_NODE])

if __name__=="__main__":
    import sys
    if len(sys.argv)>1:
        filename = sys.argv[1]
        machine = state_machine_info(filename)
    else:
        filename = "XSLTSyntaxChecker\\src\\state-machine.xml"
        machine = state_machine_info(filename)
        assert "template" in machine.possible_children("stylesheet")
        assert "text" in machine.all_element_types()
        assert machine.possible_children("text")==[]
        assert machine.possible_children("garbage")==[]
        assert "version" in machine.possible_attributes("transform")
        assert machine.possible_attributes("garbage")==[]
        assert "upper-first" in machine.possible_attribute_values("sort", "case-order")
        assert machine.possible_attribute_values("garbage", "garbage") == []
        assert machine.possible_attribute_values("garbate", "case-order") == []
        ## filename = "..\\languages\\xhtml\\xhtml-state-machine.xml"
        ## machine = state_machine_info(filename)
        ## for element in machine.all_element_types():
            ## if element!="#LITERAL":
                ## assert "lang" in machine.possible_attributes(element), "no lang on %s" % element

    machine.dump(sys.stdout)
