# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

# An implementation of remote file objects
#
# Contributors:
# * Shane Caraveo
# * Todd Whiteman

from xpcom import file, components, COMException, ServerException, server, nsError
from xpcom.server.enumerator import SimpleEnumerator
from xpcom.client import WeakReference

import logging
import threading

# needed for ftpfile
import urlparse, urllib, re, os
import StringIO, string
import time, sys, strptime

koRFProtocolDefaultPort = {
    'ftp': 21,



}
koHandledRFProtocols = koRFProtocolDefaultPort.keys()

def addslash(path):
    if len(path) < 1: return "/"
    if path[len(path)-1] != "/": path += "/"
    return path

# This is quite ugly. You will not be able to open a filename that contains
# a backslash in the name (in unix you can have this type of filename).
def normalizePath(path):
    path = os.path.normpath(path) #normalize the path
    if sys.platform == "win32":
        path = string.replace(path, "\\","/") # os.normpath changes slashes on win
    if path[0] != "/" and path != '~':
        if path[:2] != '~/':
            path = "/"+path
    elif path[:2] == '/~':
        path = path[1:]
    return path

#must persist the last username/password in order to properly open files from koEditorWrapper
class _RFService:
    
    _connections = {}
    _sessionData = {}
    _lock = None
    def __init__(self):
        self._lock = threading.Lock()
        self.log = logging.getLogger('koRFConnectionService')
        #self.log.setLevel(logging.DEBUG)

    def getConnection(self, protocol, server, port, username, password):
        c = None
        self._lock.acquire()
        if password:
            self.log.debug("getConnection: %s %s:%s@%s:%r" %
                           (protocol, username, '*' * len(password), server, port))
        else:
            self.log.debug("getConnection: %s %s@%s:%r" %
                           (protocol, username, server, port))
        try:
            if port == 0:
                port = koRFProtocolDefaultPort[protocol]
            conn_key = "%s:%s:%s:%s"%(protocol,server,port,username)
            if conn_key in self._connections and \
                self._connections[conn_key]():
                self.log.debug("getConnection, found cached connection")
                return self._connections[conn_key]()

            if protocol == 'ftp':
                c = components.classes['@activestate.com/koFTPConnection;1'].\
                    createInstance(components.interfaces.koIFTPConnection)





            else:
                self._lasterror = "Unhandled protocol type: %s" % (protocol)
                raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

            c.open(server, port, username, password)
            self._connections[conn_key] = WeakReference(c)
        finally:
            self._lock.release()
        return c

    def saveSession(self, key, data):
        self._sessionData[key] = data
        
    def getSession(self, key):
        if key in self._sessionData:
            return self._sessionData[key]
        return None
    
    
RFService = _RFService()

class koRFConnection:
    def __init__(self):
        self._connection = None
        self.protocol = ""
        self.server = ""
        self.port = 0
        self.username = ""
        self.password = ""
        self._lasterror = ""
        self._directoryCache = {}
        self.key = ""
        # can only read/write one file at a time
        self.filedata = StringIO.StringIO()
        self._globalPrefs = \
            components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
    
    #def __del__(self):
    #    self._connection.quit()


    # Note: The do_* functions must be implemented by the actual
    # connection that is handling this protocol type. Thus these local
    # do_* functions should never get called unless the protocol does
    # not implement the function
    def do_authenticateWithPassword(self): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_authenticateWithAgent(self): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_openSocket(self): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_verifyConnected(self): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_getPathInfo(self, path): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_getDirectoryList(self, path): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_rename(self, oldName, newName): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_removeFile(self, name): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_changeDirectory(self, path): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_currentDirectory(self): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_getHomeDirectory(self): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_removeDirectory(self, name): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_createDirectory(self, name, permissions): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_close(self): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_readFile(self, filename): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def do_writeFile(self, filename, data): raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    #
    # Internal methods
    #
    #
    def addFileToDirectoryCache(self, fileinfo):
        fpath = fileinfo.getFilepath()
        if not self._directoryCache.has_key(fpath):
            self.log.debug("addFileToDirectoryCache: cached file '%s'", fpath)
        #else:
            #self.log.debug("addFileToDirectoryCache: file already cached '%s'", fpath)
        self._directoryCache[fpath] = fileinfo

    def addFolderToDirectoryCache(self, path, dirinfo):
        if not self._directoryCache.has_key(path):
            self.log.debug("addFolderToDirectoryCache: new to cache '%s'", path)
            
        #else:
        #    self.log.debug("addFolderToDirectoryCache: file already cached '%s'", path)
        #    dirFolder = self._directoryCache[path]
        # Add all the child nodes to the known cache
        for child in dirinfo:
            #if not child.isDirectory():
            self.addFileToDirectoryCache(child)
        self._directoryCache[path].setChildren(dirinfo)

    def _removeFromCache(self, path, removeChildNodes=0):
        if self._directoryCache.has_key(path):
            del self._directoryCache[path]
        if removeChildNodes:
            # Remove all cached paths that are under this directory
            dirPath = addslash(path)
            for keypath in self._directoryCache.keys():
                if keypath.startswith(dirPath):
                    del self._directoryCache[keypath]

    def authenticateWithPassword(self):
        # if we don't have it, pop up a dialog asking for it!
        self.authAttempt = 0
        self._lasterror = ""
        while self.authAttempt < 3:
            #do_login
            self.authAttempt = self.authAttempt + 1
            if not self.username or not self.password:
                dialogproxy = components.classes['@activestate.com/asDialogProxy;1'].\
                            getService(components.interfaces.asIDialogProxy)
                title = "Authenticate"
                if self._lasterror:
                    title = "Authentication Failed, try again..."
                prompt = "Please enter the username and password for "+self.server
                authinfo = dialogproxy.authenticate(title, prompt, self.username, 1, 0)
                if authinfo:
                    self.log.debug("authinfo: got back '%s'", authinfo)
                    self.username, self.password = string.split(authinfo, ":",1)
                else:
                    break
            return self.do_authenticateWithPassword()
        self._lasterror = "Authentication information required"
        self.log.error("FTP AUTHENTICATION ERROR: %s", self._lasterror)
        raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def authenticateWithAgent(self):
        """Use an agent to provide remote connection authorization"""
        return self.do_authenticateWithAgent()

    def openSocket(self):
        """Open the remote socket connection.
        Uses server name {self.server} and port {self.port}"""
        self.do_openSocket()

    #
    # XPcom methods, available through koIFTPConnection interface
    #
    def open(self, server, port, username, password):
        self.server = server
        if port > 0: self.port = port
        else: self.port = koRFProtocolDefaultPort[self.protocol]
        self.openSocket()

        self.username = username
        self.password = password
        self.key = "%s:%s:%s" % (self.server, self.port, self.username)
        self.log.debug("session data %s:%s:%s",self.server,self.port,self.username)
        if not self.username or not self.password:
            sess = RFService.getSession(self.key)
            if sess:
                self.username = sess[0]
                self.password = sess[1]

        if not self.authenticateWithAgent():
            if not self.authenticateWithPassword():
                self._lasterror = "Password authentication failed."
                self.log.error("AUTHENTICATION ERROR: %s", self._lasterror)
                raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
            self.log.debug("Password authentication successful.")
        else:
            self.log.debug("Agent authentication successful.")

        self.log.info("%s connection opened on %s:%s for user %s", self.protocol,self.server,self.port,self.username)
        #save the session
        RFService.saveSession(self.key,[self.username,self.password])
        return 1

    def verifyConnected(self):
        return self.do_verifyConnected()

    def getPathInfo(self, path):
        return self.do_getPathInfo(path)

    def getDirectoryList(self, path):
        return self.do_getDirectoryList(path)
        
    # Note: We add all files and folders that are found to the cache. The
    # sub-directories that get added will not initially have contents, to find
    # the contents, a check is done when the directory gets selected.
    # We return a list of everything found in the directory, including sub
    # directories.
    def list(self, path, refresh):
        #self.log.debug("list: path: '%s'",path)
        self._lasterror = ""
        verifyConnected = True
        # Return the cached copy if not asking for a refresh
        cached_info = None
        if self._directoryCache.has_key(path):
            cached_info = self._directoryCache[path]
        # If told to refresh or the file path is not known (not cached)
        if refresh or not cached_info:
            self.verifyConnected()
            verifyConnected = False
            cached_info = self.getPathInfo(path)
            self._directoryCache[path] = cached_info
            self.log.debug("Returned new list path for: '%s'", path)
        #else:
        #    self.log.debug("Returning cached path for: '%s'",path)
        if cached_info.isDirectory():
            if refresh or cached_info.needsDirectoryListing():
                if verifyConnected:
                    self.verifyConnected()
                # we have a directory that has needs refresh
                # or has not had a directory listing yet
                dirlist = self.getDirectoryList(path)
                self.addFolderToDirectoryCache(path, dirlist)
        return cached_info

    def rename(self, oldName, newName):
        self.verifyConnected()
        self.log.debug("rename %s to %s" ,oldName, newName)
        result = self.do_rename(oldName, newName)
        self._removeFromCache(oldName, removeChildNodes=1)
        return result

    def removeFile(self, name):
        self.verifyConnected()
        self.log.debug("remove file %s",name)
        result = self.do_removeFile(name)
        self._removeFromCache(name, removeChildNodes=0)
        return result

    def changeDirectory(self, path):
        self.verifyConnected()
        self.log.debug("changeDirectory cwd %s",path)
        return self.do_changeDirectory(path)

    def currentDirectory(self):
        self.verifyConnected()
        return self.do_currentDirectory()

    def getHomeDirectory(self):
        self.verifyConnected()
        return self.do_getHomeDirectory()

    def removeDirectory(self, name):
        self.verifyConnected()
        self.log.debug("remove directory %s",name)
        result = self.do_removeDirectory(name)
        self._removeFromCache(name, removeChildNodes=1)
        return result

    def createDirectory(self, name, permissions):
        self.verifyConnected()
        self.log.debug("createDirectory %s, %r" ,name, permissions)
        return self.do_createDirectory(name, permissions)

    def close(self):
        self.log.debug("closing connection")
        return self.do_close()

    def readFile(self, filename):
        self.verifyConnected()
        self.log.debug("reading file: %s", filename)
        return self.do_readFile(filename)

    def writeFile(self, filename, data):
        self.verifyConnected()
        self.log.debug("writing file: %s (%d chars)", filename, len(data))
        return self.do_writeFile(filename, data)

    #
    # These XPcom methods should NOT be implemented/overwritten in child classes
    #
    def refresh(self, path):
        self.verifyConnected()
        # remove path and all children from cache
        if not self._directoryCache.has_key(path): return
        for key in self._directoryCache.keys():
            if key[:len(path)] == path:
                del self._directoryCache[key]
        self.list(path, 1)

    def havePath(self, path):
        # self.log.debug('havePath: checking %s',path)
        haveit = self._directoryCache.has_key(path)
        if not haveit and self._connection:
            self.log.debug('havePath: not cached, list %s',path)
            debug_dirCache(self.log, self._directoryCache)
            info = self.list(path, 0)
            haveit = (info != None)
        self.log.debug('havePath: %s for %s', haveit, path)
        return haveit

    # These are the readonly interface attributes used through pyXPcom
    def get_protocolType(self):
        return self.protocol

    def get_server(self):
        return self.server

    def get_port(self):
        return self.port

    def get_username(self):
        return self.username

    def get_password(self):
        return self.password

    def get_lastError(self):
        return self._lasterror


# Debug the directory cache
def debug_dirCache(log, cache):
    # XXX - Find out to check this
    #if log.getLevel() == logging.DEBUG:
        log.debug("*" * 40)
        log.debug("Directory cache:")
        for ckey in cache.keys():
            value = cache[ckey]
            if isinstance(value, list):
                s = ["%s:" % ckey]
                for dkey in value:
                    s.append("    %s" % dkey)
                log.debug('\n'.join(s))
            else:
                log.debug("%s: %s", ckey, value)
        log.debug("*" * 40)

